#include <math.h>
#include <values.h>
#include <nan.h>
#include <useful.h>
#include <simulator.h>
#include <commands.h>
#include <xerion/minimize.h>
#include "utilsDCP.h"

#define DB(x)	x

#define WLEN	32
#define RLEN	1024	/* maximum response length */
#define	BLANK	'_'

#define MoldError(unit)	((unit)->inputDeriv)

/* used in command_states to specify part of word */
#define ONSET		1
#define VOWEL		2
#define CODA		3
#define BODY		4
#define ONCLEUS		5

#define MAX_ONSET 4	/* schnook, used for VOWELCEN orth code */

#define	NONE		0
#define	SYLLABLE	1
#define	NSYLLABLE	2
#define	LETTERS		3
#define	SEQUENTIAL	4
#define	NSEQUENTIAL	5
#define ATTENTION	6
#define NATTENTION	7
#define GRAPHEMES	8
#define	VOWELCEN	9

static	int	_offset		 = 3 ;
static	int	_recurrent 	 = FALSE ;
static	int	_wta	 	 = FALSE ;
static	int	_nopr	 	 = FALSE ;
static	int	_norm		 = FALSE ;
static	Real	_overlap	 = 0.3 ;
static	Real	_on_threshold 	 = 0.5 ;
static	Real	_print_tolerance = 0.0 ;
static	Real	_print_threshold = MAXFLOAT ;
static	int	_impStart = 0 ;
static	int	_impNum = 0 ;

typedef struct SMentry {
  String	code ;
  String	graphemes ;
  String	phonemes ;
  Real		frequency ;
} SMentry ;

typedef struct SMlist {
  int		size ;
  int		numEntries ;
  SMentry	*entries ;
} SMlist ;

/*****************************************************************************
 * utilities
 *****************************************************************************/
static String	REcompile(re)
  String re ;
{
  char		buffer[WLEN] ;
  /* use regex(compiled, str); */
  sprintf(buffer, "^%s$", re) ;
  return strdup((String)regcmp(buffer, NULL)) ;
}

static	Real	sq(x)
  Real	x ;
{
  return x*x ;
}

static	int	_SSEslice ;

static void	incSumSqErr(unit, data)
  Unit		unit ;
  void		*data ;
{
  Boolean  rec = (unit->net->type&RECURRENT) ;
  Real	target = (rec ? unit->targetHistory[_SSEslice] : unit->target) ;
  Real  output = (rec ? unit->outputHistory[_SSEslice] : unit->output) ;

  if (!isNaN(target)) 
    *(Real *)data += sq(target - output) ;
}

static	Real	groupSumSquaredError(group)
  Group	group ;
{
  Real	error = 0.0 ;
  if (group->net->type&RECURRENT && group->net->currentTime > 0) 
    for (_SSEslice = 0 ; _SSEslice < group->net->currentTime ; _SSEslice++)
      groupForAllUnits(group, incSumSqErr, (void *)&error) ;
  else {
    _SSEslice = 0 ;
    groupForAllUnits(group, incSumSqErr, (void *)&error) ;
  }
  return error ;
}

static	Real	sumSquaredError(net)
  Net	net ;
{
  Real	error = 0.0 ;
  if (net->type&RECURRENT && net->currentTime > 0) 
    for (_SSEslice = 0 ; _SSEslice < net->currentTime ; _SSEslice++)
      netForAllUnits(net, OUTPUT, incSumSqErr, (void *)&error) ;
  else
    netForAllUnits(net, OUTPUT, incSumSqErr, (void *)&error) ;
  return error ;
}

static	void	groupSubtractOldStates(group, oldStates)
  Group	group ;
  Real*	oldStates ;
{
  int unitIdx ;
  Unit *unit ;

  for (unitIdx = 0, unit = group->unit ; unitIdx < group->numUnits ; unitIdx++, unit++) {
    (*unit)->output -= oldStates[unitIdx] ;
  }
}


static	void	printUnitState (unit, data)
  Unit	unit ;
  void	*data ;
{
  if (_print_threshold == MAXFLOAT)
    fprintf(dout, "%f ", unit->output) ;
  else
    fprintf(dout, "%d ", (unit->output >= _print_threshold)) ;
}

static	void	setStateToContribution (unit, data)
  Unit	unit ;
  void	*data ;
{
  unit->output *= -unit->outputDeriv ;
}

static	void	saveError (unit, data)
  Unit	unit ;
  void	*data ;
{
  MoldError(unit) = 0.0 ;
  incSumSqErr(unit, (void *)&MoldError(unit)) ;
}

static	void	impIncChange (group, data)
  Group group ;
  void	*data ;
{
  int	idx ;
  for (idx = _impStart ; idx < _impStart+_impNum ; ++idx) {
    Unit	unit = group->unit[idx] ;
    *(Real *)data -= MoldError(unit) ;
    incSumSqErr(unit, data) ;
  }
}

static	void	printSSEIncreaseAfterLesion (unit, data)
  Unit	unit ;
  void	*data ;
{
  Net	net = unit->net ;
  Real change = 0.0 ; 

  lesionUnit(unit) ;
  MupdateNetActivities(net) ;
  if (_impNum == 0) {
    change = sumSquaredError(net) - (*(Real *)data) ;
  } else {
    netForAllGroups(net, OUTPUT, impIncChange, (void *)&change) ;
  }    
  if (_print_threshold == MAXFLOAT) fprintf(dout, "%f ", change) ;
  else fprintf(dout, "%d ", (change >= _print_threshold)) ;
  unlesionUnit(unit) ;
}

static	void	printErrorIncreaseAfterLesion (unit, data)
  Unit	unit ;
  void	*data ;
{
  Net	net = unit->net ;
  Real change = 0.0 ; 

  lesionUnit(unit) ;
  MupdateNetActivities(net) ;
  change = currentNet->error - (*(Real *)data) ;
  if (_print_threshold == MAXFLOAT) fprintf(dout, "%f ", change) ;
  else fprintf(dout, "%d ", (change >= _print_threshold)) ;
  unlesionUnit(unit) ;
}

static	void	copyOutputToTarget(unit, data)
  Unit	unit ;
  void	*data ;
{
  unit->target = unit->output ;
}

static	void	copyTargetToOutput(unit, data)
  Unit	unit ;
  void	*data ;
{
  unit->output = unit->target ;
}

static	void	zeroTotalInput(unit, data)
  Unit	unit ;
  void	*data ;
{
  unit->totalInput = 0.0 ;
}

static	void	 incrementSimilarity(unit, data)
  Unit	unit ;
  void	*data ;
{
  /* test pattern's output has been copied to target */
  *(Real *)data += unit->output * unit->target ;
}

static	void	 incrementTotalInput(unit, data)
  Unit	unit ;
  void	*data ;
{
  unit->totalInput += *(Real *)data * (2.0*unit->target - 1.0) ;
}

static	void	 computeOutputFromTotalInput(unit, data)
  Unit	unit ;
  void	*data ;
{
  unit->totalInput *= *(Real *)data ;
  unit->output = sigmoid(unit->totalInput) ;
}

static int contains(s1, s2)
  String s1,s2;
{
  int pos = 0, len = strlen(s2);
  while (*s1) 
    if (!strncmp(s1,s2,len)) return pos+len ;
    else { s1++; pos++; }
  return FALSE;
}

static	int	representation(str)
  String	str ;
{
  if   (!str || !strncmp(str, "syllable",   strlen(str))) {
    return SYLLABLE ;
  } else if    (!strncmp(str, "nsyllable",  strlen(str))) {
    return NSYLLABLE ;
  } else if    (!strncmp(str, "sequential", strlen(str))) {
    _recurrent = TRUE ;
    return SEQUENTIAL ;
  } else if    (!strncmp(str, "nsequential", strlen(str))) {
    _recurrent = TRUE ;
    return NSEQUENTIAL ;
  } else if    (!strncmp(str, "attention",  strlen(str))) {
    return ATTENTION ;
  } else if    (!strncmp(str, "nattention",  strlen(str))) {
    return NATTENTION ;
  } else if    (!strncmp(str, "letters",    strlen(str))) {
    return LETTERS ;
  } else if    (!strncmp(str, "vowelcen",    strlen(str))) {
    return VOWELCEN ;
  } else if    (!strncmp(str, "graphemes",    strlen(str))) {
    return GRAPHEMES ;
  } else if    (!strncmp(str, "none",    strlen(str))) {
    return NONE ;
  } else IErrorAbort("unknown representation \"%s\"", str) ;
}

/*****************************************************************************
 * orthography (letters)
 *****************************************************************************/

#define	MAX_WORD_LENGTH		8
#define N_ATTENTION_POSITIONS	MAX_WORD_LENGTH
#define N_LEFT_OF_FIXATION	2
#define N_LETTERS		26
#define MARK_LETTER_SPACE	1
static	String letters 		= "abcdefghijklmnopqrstuvwxyz" ;

#define N_GRAPHEMES		77

String graphemes[N_GRAPHEMES] = {
"eau", "iou", "tch",
"ch", "gh", "gn", "gu", "qu", "ph", "rh", "sh", "th", "wh", "wr", "kn",
"ai", "au", "aw", "ay", "ea", "ee", "ei", "eu", "ew", "ey",
"ia", "ie", "io", "oa", "oe", "oi", "oo", "ou", "ow", "oy", 
"ue", "ui", "uy",
"bb", "ck", "dd", "ff", "gg", "ll", "ng", "nn", "pp", "rr", 
"ss", "tt", "zz", 
"a",  "b",  "c",  "d",  "e",  "f",  "g",  "h",  "i",  "j",  
"k",  "l",  "m",  "n",  "o",  "p",  "q",  "r",  "s",  "t",  
"u",  "v",  "w",  "x",  "y",  "z"
};

static	int	numGraphemesPerPhoneme(gstr, pstr)
  String gstr, pstr ;
{
  int	i, gmove = 0, gsize ;
  String g ;

  if (!*pstr || !*gstr) return -1 ;

  /* silent initial graphemes */
  if ((*gstr == 'h' && *pstr != 'h') ||		/* herb */
      (*gstr == 'p' && *pstr == 's') ||		/* psalm */
      (*gstr == 't' && *pstr == 'z') ||		/* tzar */
      (*gstr == 'c' && *pstr == 'z')) {		/* czar */
    gstr++ ; gmove++ ;
  }
  for (i = 0 ; i < N_GRAPHEMES ; i++) {
    if (!strncmp(gstr, g=graphemes[i], gsize=strlen(graphemes[i]))) {

      /* exceptions to graphemes: reject match */
      if ((!strcmp(g, "ng") && (!strncmp(pstr, "ndZ", 3) ||
				!strncmp(pstr, "nj",  2))) ||	/* strange */
	  (!strcmp(g, "ui") && *pstr == 'w') ||			/* suite */
	  (!strcmp(g, "ue") && *pstr == 'w') ||			/* suede */
	  (!strcmp(g, "gu") && (strlen(gstr)<3 || !strchr("aeiy", *(gstr+2))))
	  ) {
	continue ;

      /* one-to-many GPC: stay put for first (this) phoneme */
      } else if ((!strcmp(g, "ch")  && !strncmp(pstr, "tS", 2)) ||
		 (!strcmp(g, "ch")  && !strncmp(pstr, "kw", 2)) ||	/* choir */
		 (!strcmp(g, "tch") && !strncmp(pstr, "tS", 2)) ||
		 (!strcmp(g, "th")  && !strncmp(pstr, "tT", 2)) ||
		 (!strcmp(g, "qu")  && !strncmp(pstr, "kw", 2)) ||
		 (!strcmp(g, "x")   && !strncmp(pstr, "ks", 2)) ||
		 (!strcmp(g, "j")   && !strncmp(pstr, "dZ", 2)) ||
		 (!strcmp(g, "g")   && !strncmp(pstr, "dZ", 2)) ||
		 (!strcmp(g, "o")   && !strncmp(pstr, "w^", 2)) ||
		 ( strcmp(g, "y")   && !strncmp(pstr, "yU", 2)) ||
		 ( strcmp(g, "y")   && !strncmp(pstr, "yu", 2))   ) {
	return 0 ;

      /* many-to-one GPC: move over additional grapheme */
      } else if (((!strncmp(gstr, "sci", 3) || !strncmp(gstr, "sce", 3)) && *pstr == 's') ||	/* scene */
		  (!strncmp(gstr, "sw", 2)  &&  strncmp(pstr, "sw", 2))			  ||	/* sword */
		  (!strncmp(gstr, "dg", 2)  && !strcmp(pstr, "j"))	  
		 ) {
	gsize++ ;
      } else if (!strncmp(gstr, "sch", 3) && *pstr == 'S') {	/* schnook */
	gsize += 2 ;
      }
      
      /* default case: one grapheme to one phoneme */
      gstr  += gsize ;
      gmove += gsize ;
      pstr++ ;
      break ;
    }
  }
  /* trailing silent graphemes */
  if ((!strncmp(gstr, "gh", 2)  && *pstr != 'g' && *pstr != 'f') ||
      (!strncmp(gstr, "cht", 3) && *pstr == 't')) {	/* yacht */
    gstr  += 2 ;     gmove += 2 ;
  } else if ((*gstr == 'l' && *pstr != 'l')) {	/* walk would */
    gstr++ ;     gmove++ ;
  }
  if (  strlen(gstr)==2 &&
      (!strcmp(gstr, "es") && (*pstr == 's' || *pstr == 'z')) ||
      (!strcmp(gstr, "ed") && (*pstr == 't' || *pstr == 'd'))) {
    gstr++ ;  gmove++ ;
  }
  if (! *pstr) gmove += strlen(gstr) ;
  return gmove ;
}

static	void	printGPCs(graphemes, phonemes) 
  String graphemes, phonemes ;
{
  int	i, gmove ;
  String gstr, pstr ;

  fprintf(dout, "%s\t", graphemes) ;
  for (gstr = graphemes, pstr = phonemes ; *pstr ; pstr++) {
    gmove = numGraphemesPerPhoneme(gstr, pstr) ;
    fprintf(dout, "%c", *pstr) ;
    if (gmove) {
      fprintf(dout, "/") ;
      for (i = 0 ; i < gmove ; i++) 
	fprintf(dout, "%c", gstr[i]) ;
      fprintf(dout, " ") ;
    }
    gstr += gmove ;
  }
  if (*gstr) fprintf(dout, " [%s]", gstr) ;
  fprintf(dout, "\n") ;
}

static void	shiftClampingHistory(group, time, shift, width, mark_space)
  Group	group ;
  int	time, shift, width, mark_space ;
{
  int	here, there ;
  int	max_idx = group->numUnits ;
  Unit	*units = group->unit ;

  for (here = 0, there = shift*width ; 
       here < max_idx ; 
       here++, there++) 
    if (there < max_idx) 
      units[here]->clampingHistory[time] = units[there]->clampingHistory[time] ;
    else if (here%width == 0 && mark_space)
      units[here]->clampingHistory[time] = 1.0 ;
    else
      units[here]->clampingHistory[time] = 0.0 ;
}    

static void	shiftTargetHistory(group, time, shift, width, mark_space)
  Group	group ;
  int	time, shift, width, mark_space ;
{
  int	here, there ;
  int	max_idx = group->numUnits ;
  Unit	*units = group->unit ;

  for (here = 0, there = shift*width ; 
       here < max_idx ; 
       here++, there++) 
    if (there < max_idx) 
      units[here]->targetHistory[time] = units[there]->targetHistory[time] ;
    else if (here%width == 0 && mark_space)
      units[here]->targetHistory[time] = 1.0 ;
    else
      units[here]->targetHistory[time] = 0.0 ;
}

void	Fixate(Letters, Attention, saccade, attention, time)
  Group	Letters, Attention ;
  int	saccade, attention, time ;
{
  int idx ;
  Unit	*units = Attention->unit ;

  if (saccade > 0) {
    shiftClampingHistory(Letters, time, saccade, 
			 MARK_LETTER_SPACE+N_LETTERS, 
			 MARK_LETTER_SPACE) ;
    shiftTargetHistory(Attention, time, saccade, 1, FALSE) ;
  }
  if (attention == 0) {
    units[0]->output = 1.0 ;
    for (idx = 1 ; idx < Attention->numUnits ; idx++) 
      units[idx]->output = 0.0 ;
  }
/*
  for (idx = 0 ; idx < Attention->numUnits ; idx++) 
    units[idx]->output = (idx == attention ? 1.0 : 0.0) ;
*/
}

/*****************************************************************************
 * orthography (syllable)
 *****************************************************************************/
#define SYL_N_GRAPHEMES			108
#define SYL_N_ONSET_GRAPHEMES		33
#define SYL_N_VOWEL_GRAPHEMES		27
#define SYL_N_FINAL_GRAPHEMES		48

static String onset_graphemes[SYL_N_ONSET_GRAPHEMES] = {
"y",  "s",  "p",  "t",  "k",  "q",  "c",  "b",  "d",  "g",  
"f",  "v",  "j",  "z",  "l",  "m",  "n",  "r",  "w",  "h",  "u",
"ch", "gh", "gn", "gu", "ph", "ps", "qu", "rh", "sh", "th", "ts", "wh"
};

static String vowel_graphemes[SYL_N_VOWEL_GRAPHEMES] = {
"e",  "i",  "o",  "u",  "a",  "y",
"ai", "au", "aw", "ay", "ea", "ee", "ei", "eu", "ew", "ey", "ie", 
"oa", "oe", "oi", "oo", "ou", "ow", "oy", "ue", "ui", "uy"
};

static String final_graphemes[SYL_N_FINAL_GRAPHEMES] = {
"h", "r",  "l",  "m",  "n",  "b",  "d",  "g",  "c",  "x",  "f",  "v",  "j",
"s", "z",  "p",  "t",  "k",   
"bb", "ch", "ck", "dd", "dg", "ff", "gg", "gh", "gn", "gu", "ks", "ll", "ng",
"nn", "ph", "pp", "ps", "qu", "rr", "sh", "sl", "ss", "tch", "th", "ts", "tt",
"zz",
"e", "es", "ed"
};

#define NSYL_N_GRAPHEMES		105
#define NSYL_N_ONSET_GRAPHEMES		30
#define NSYL_N_VOWEL_GRAPHEMES		27
#define NSYL_N_FINAL_GRAPHEMES		48

static String nsyl_onset_graphemes[NSYL_N_ONSET_GRAPHEMES] = {
"y",  "s",  "p",  "t",  "k",  "q",  "c",  "b",  "d",  "g",  
"f",  "v",  "j",  "z",  "l",  "m",  "n",  "r",  "w",  "h",  
"ch", "gh", "gn", "ph", "ps", "rh", "sh", "th", "ts", "wh"
};

static String nsyl_vowel_graphemes[NSYL_N_VOWEL_GRAPHEMES] = {
"e",  "i",  "o",  "u",  "a",  "y",  "ai", "au", "aw", "ay", 
"ea", "ee", "ei", "eu", "ew", "ey", "ie", "oa", "oe", "oi", 
"oo", "ou", "ow", "oy", "ue", "ui", "uy"
};

static String nsyl_final_graphemes[NSYL_N_FINAL_GRAPHEMES] = {
"h",  "r",  "l",  "m",  "n",  "b",  "d",  "g",  "c",  "x",  
"f",  "v",  "j",  "s",  "z",  "p",  "t",  "k",  "q",  "bb", 
"ch", "ck", "dd", "dg", "ff", "gg", "gh", "gn", "ks", "ll", 
"ng", "nn", "ph", "pp", "ps", "rr", "sh", "sl", "ss", "tch", 
"th", "ts", "tt", "zz", "u",  "e",  "es", "ed"
};


int onset_end(str)
  String str;
{
  int pos = 0;

  if (strchr("y",*str)) {str++; pos++;}  
  while (!strchr("aeiouy",*str) && *str != BLANK) {
    if ( !strncmp(str, "qu",  2) ||
	 !strncmp(str, "sua", 3) ||
	(!strncmp(str, "gu",  2) && 
	 (strlen(str)==2 || (strlen(str)>2 && strchr("aeiy",*(str+2)))))) {
      str++; pos++;
    }
    str++; pos++;
  }
  return pos;
}

int nsyl_onset_end(str)
  String str;
{
  int pos = 0;
  if (strchr("y",*str)) {str++; pos++;}  
  return pos+strcspn(str, "aeiouy") ;
}

static int digraph_non_final_s ;

int vowel_end(str)
  String str;
{
  int pos = 0;
  digraph_non_final_s = FALSE ;      
  if (*str == BLANK) {  str++; pos++;  }
  while (*str && strchr("aeiouyw", *str)) {
    if (strchr("yw", *str)) {
      str++; pos++; break;
    } else if (strlen(str)>1 && strchr("s",*(str+1)))
      digraph_non_final_s = TRUE ;      
    str++; pos++;
  }
  return pos;
}

static	int	code_exclusive(prev, this, next, mark_null, alts, unit)
  char  prev, this, next, *alts;
  Unit	*unit ;
{
  Real	state; 
  int 	found = FALSE;

  if (mark_null) {
    if (this) state = 0 ;
    else      state = 1 ;
    if (unit) (*(unit++))->output = state ;
    else      fprintf(dout, "%d ", (int)state) ;
  }    
  while (*alts) {
    if (this == *alts) {
      state = 1.0 ;
      found = TRUE;
    } else if ((prev == *alts || next == *alts) && this != NULL) 
      state = _overlap ;
    else
      state = 0.0 ;
    if (unit) 
      (*(unit++))->output = state ;
    else if (_overlap == 0.0)
      fprintf(dout, "%1d ", (int)state) ;
    else
      fprintf(dout, "%.1f ", state) ;
    alts++ ;  
  }
  if (!unit) {
    if (found) fprintf(dout, "#%c", this);
    fprintf(dout, "\n") ;
    fflush(dout);
  }
  return found;
}

static	int	code_exclusive_multi(str, alts, n_alts, unit)
  String str, alts[];
  int n_alts ;
  Unit	*unit ;
{
  int 	i, pos, state, found = FALSE;

  for (i = 0; i < n_alts; i++) {
    if (!found && !strncmp(str,alts[i],strlen(alts[i]))) {
      found = TRUE ;
      pos = i ;
      if (unit) (*(unit++))->output = 1;
      else      fprintf(dout, "1 #%s\n", alts[i]) ;
    } else {
      if (unit) (*(unit++))->output = 0;
      else      fprintf(dout, "0 ") ;
    }
  }
  return (found ? strlen(alts[pos]) : 0) ;
}


static int code_inclusive(str, max_str_n, alts, n_alts, unit)
  String str, alts[] ;
  int max_str_n, n_alts ;
  Unit	*unit ;
{
  /* if unit==NULL then print rather than store values  */
  int i, pos;
  for (i = 0; i < n_alts; i++) {
    pos = contains(str, alts[i]);
    if (pos && pos <= max_str_n) {
      if (unit) (*(unit++))->output = 1;
      else      fprintf(dout, "1 #%s\n", alts[i]) ;
    } else {
      if (unit) (*(unit++))->output = 0;
      else      fprintf(dout, "0 ") ;
    }
  }
  return MIN(strlen(str), max_str_n) ;
}

/*****************************************************************************
 * orthography (general)
 *****************************************************************************/
static	String	default_orthography_group_name(rep) 
  int	rep ;
{
  switch (rep) {
  case GRAPHEMES:
  case NSYLLABLE:
  case SYLLABLE:	return "Graphemes" ;
  case VOWELCEN:
  case LETTERS:		return "Letters" ;
  case NONE:		return NULL ;
  default:
    IErrorAbort("default_orthography_group_name: unknown orthographic encoding") ;
  }
}

static	int	orthography_size(rep) 
  int	rep ;
{
  switch (rep) {
  case GRAPHEMES:
  case SYLLABLE:
    return SYL_N_GRAPHEMES ;
  case NSYLLABLE:
    return NSYL_N_GRAPHEMES ;
  case VOWELCEN:
  case LETTERS:
    return (MARK_LETTER_SPACE + N_LETTERS) * (MAX_WORD_LENGTH + N_LEFT_OF_FIXATION) ;
  default:
    IErrorAbort("orthography_size: unknown orthographic encoding") ;
  }
}

static	int	copyOutputsToClampingHistory(unit, nUnits)
  Unit	*unit ;
  int	nUnits ;
{
  int idx, slice ;
  Net net = (*unit)->net ;
  int nSlices = net->timeSlices ;

  if (net->type&RECURRENT) 
    for (idx = 0 ; idx < nUnits ; idx++, unit++)
      for (slice = 0; slice < nSlices ; slice++)
	(*unit)->clampingHistory[slice] = (*unit)->output ;
  return nUnits ;
}

static	void	encode_orthography(rep, entry, unit) 
  int		rep ;
  SMentry	*entry ;
  Unit		*unit ;
{
  int	idx, onset ;
  String	str	= entry->graphemes ;
  char	prev = NULL, this = NULL ;
  /*
   * if unit==NULL then print rather than store values
   */
  switch (rep) {
  case GRAPHEMES:
  case SYLLABLE:
    if (_recurrent && !unit) fprintf(dout,"[all]\n") ;
    str  += code_inclusive(str, onset_end(str), onset_graphemes, SYL_N_ONSET_GRAPHEMES, unit) ;
    if (unit) unit += copyOutputsToClampingHistory(unit,SYL_N_ONSET_GRAPHEMES) ;
    str  += code_inclusive(str, vowel_end(str), vowel_graphemes, SYL_N_VOWEL_GRAPHEMES, unit) ;
    if (unit) unit += copyOutputsToClampingHistory(unit,SYL_N_VOWEL_GRAPHEMES) ;
    str  += code_inclusive(str, strlen(str),    final_graphemes, SYL_N_FINAL_GRAPHEMES, unit) ;
    if (unit) unit += copyOutputsToClampingHistory(unit,SYL_N_FINAL_GRAPHEMES) ;
    else      fprintf(dout, "\n") ;
    if (*str) fprintf(dout,"\n#WARNING: \"%s\" uncoded\n",str); 
    break ;
  case NSYLLABLE:
    if (_recurrent && !unit) fprintf(dout,"[all]\n") ;
    str  += code_inclusive(str, nsyl_onset_end(str), nsyl_onset_graphemes, NSYL_N_ONSET_GRAPHEMES, unit) ;
    if (unit) unit += copyOutputsToClampingHistory(unit, NSYL_N_ONSET_GRAPHEMES) ;
    str  += code_inclusive(str, vowel_end(str),      nsyl_vowel_graphemes, NSYL_N_VOWEL_GRAPHEMES, unit) ;
    if (unit) unit += copyOutputsToClampingHistory(unit, NSYL_N_VOWEL_GRAPHEMES) ;
    str  += code_inclusive(str, strlen(str),         nsyl_final_graphemes, NSYL_N_FINAL_GRAPHEMES, unit) ;
    if (unit) unit += copyOutputsToClampingHistory(unit, NSYL_N_FINAL_GRAPHEMES) ;
    else      fprintf(dout, "\n") ;
    if (*str) fprintf(dout,"\n#WARNING: \"%s\" uncoded\n",str); 
    break ;
  case LETTERS:
    if (_recurrent && !unit) fprintf(dout,"[all]\n") ;
    for (idx = 0 ; idx < N_LEFT_OF_FIXATION ; idx++) {
      code_exclusive(NULL, NULL, NULL, MARK_LETTER_SPACE, letters, unit) ;
      if (unit) unit += copyOutputsToClampingHistory(unit,MARK_LETTER_SPACE+N_LETTERS) ;
    }
    for (idx = 0 ; idx < MAX_WORD_LENGTH ; idx++) {
      this = *str ;
      str += code_exclusive(prev, *str, *(str+1), MARK_LETTER_SPACE, letters, unit) ;
      if (unit) unit += copyOutputsToClampingHistory(unit,MARK_LETTER_SPACE+N_LETTERS) ;
      prev = this ;
    }
    break ;
  case VOWELCEN:
    onset = onset_end(str) ;
    if (_recurrent && !unit) fprintf(dout,"[all]\n") ;
    for (idx = 0 ; idx < MAX_ONSET-onset ; idx++) {
      code_exclusive(NULL, NULL, NULL, MARK_LETTER_SPACE, letters, unit) ;
      if (unit) unit += copyOutputsToClampingHistory(unit,MARK_LETTER_SPACE+N_LETTERS) ;
    }
    for (idx = 0 ; idx < MAX_WORD_LENGTH+N_LEFT_OF_FIXATION+onset-MAX_ONSET ; idx++) {
      this = *str ;
      str += code_exclusive(prev, *str, *(str+1), MARK_LETTER_SPACE, letters, unit) ;
      if (unit) unit += copyOutputsToClampingHistory(unit,MARK_LETTER_SPACE+N_LETTERS) ;
      prev = this ;
    }
    break ;
  default:
    IErrorAbort("encode_orthography: unknown orthographic encoding") ;
  }
}

static	void	input_orthography(rep, str, group) 
  int	rep ;
  String	str ;
  Group	group ;
{
  static SMentry	entry ;

  if (orthography_size(rep) != group->numUnits)
    IErrorAbort("mismatch in number of orthography units") ;
  
  entry.graphemes = str ;
  encode_orthography(rep, &entry, group->unit) ;
}

/*****************************************************************************
 * phonology
 *****************************************************************************/

#define	SYL_N_PHONEMES		57
#define SYL_N_PHON_ONSETS	21	
#define	SYL_N_PHON_VOWELS	14
#define	SYL_N_PHON_CODAS	22

static String	syllable_phonemes[SYL_N_PHONEMES] = {
"s",
"b", "p", "d", "t", "g", "k",
"f", "v", "z", "T", "D", "S",  "Z", 
"l", "r", "w", "m", "n", "h", 
"y",
"a", "@", "e", "i", "o", "u", "A", "E", "I", "O", "U", "W", "Y", "^", 
"r", "l", "m", "n", "N",  
"b", "g", "d",  
"ps", "ks", "ts",
"s", "f", "v", 
"p", "k", "t",  
"z", "S", "Z",  "T",  "D"
};

/*
 * "nsyllable" differs from "syllable" in the presence of 
 *	/C/ for /tS/
 *	/j/ for /dZ/
 * and also in the ordering of phonemes (for mutual-exclusivity)
 */
#define	NSYL_N_PHONEMES		61
#define NSYL_PHON_VOWEL_SLOT	3	/* will test in loops starting at 0 */
#define NSYL_N_PHON_SLOTS	13
static String	nsyllable_phon_slots[NSYL_N_PHON_SLOTS] = { 
"sSC", "zZjfvTDpbtdkgmnh", "lrwy",
"aeiou@^AEIOUWY",
"r", "l", "mnN", "bgd", "PK+", "sz", "fvpk", "t", "SZTDCj"
};

#define SEQ_N_PHONEMES	36
static String	sequential_phonemes[SEQ_N_PHONEMES] = {
"s", "t", "d", "p", "b", "k", "g", "f", "v", "z", "T", "D", "S", "Z", "r", "l", "m", "n", "N", "w", "h", "y", 
"a", "e", "i", "o", "u", "@", "^", "A", "E", "I", "O", "U", "W", "Y"
};

#define NSEQ_N_PHONEMES	38
static String	nsequential_phonemes[NSEQ_N_PHONEMES] = {
"s", "z", "S", "Z", "f", "v", "T", "D", "t", "d", "p", "b", "k", "g", "C", "j", "N", "m", "n", "h", "l", "r", "w", "y", 
"a", "e", "i", "o", "u", "@", "^", "A", "E", "I", "O", "U", "W", "Y"
};

#define phoneme_onset_end(str)	strcspn(str, "a@eiouAEIOUWY^")

#define default_phonology_group_name(rep) "Phonemes"

/***************************************************************************/

static	int	phonology_size(rep) 
  int	rep ;
{
  switch (rep) {
  case GRAPHEMES:	return orthography_size(rep) ;
  case SYLLABLE:	return SYL_N_PHONEMES ;
  case NSYLLABLE:	return NSYL_N_PHONEMES ;
  case SEQUENTIAL:	return SEQ_N_PHONEMES ;
  case NSEQUENTIAL:	return NSEQ_N_PHONEMES ;
  case ATTENTION:	return SEQ_N_PHONEMES + N_ATTENTION_POSITIONS;
  case NATTENTION:	return NSEQ_N_PHONEMES + N_ATTENTION_POSITIONS;
  default: IErrorAbort("phonology_size: unknown phonological encoding") ;
  }
}

/* ASSUMES NSYLLABLE PHONOLOGICAL CODING */
void	SMnormalizeOutput(group)
  Group	group ;
{
  int	sl, unitIdx ;
  char	*s ;
  Real	sum ;
  Unit *unit, *slotUnit ;

  /* 
   * y_i = exp(x_i)/(r+sum[exp(x_i)])
   *     where r=0 for 1-of-N and r=1 for 0-or-1-of-N
   */
  for (sl=0, unit=group->unit ; 
       sl<NSYL_N_PHON_SLOTS ;
       unit += strlen(nsyllable_phon_slots[sl]), sl++) {
    sum = (sl==NSYL_PHON_VOWEL_SLOT ? 0.0 : 1.0) ;
    for (s=nsyllable_phon_slots[sl], slotUnit=unit ; *s ; s++, slotUnit++) {
      (*slotUnit)->output = exp((*slotUnit)->totalInput) ;
      sum += (*slotUnit)->output ;
    }
    for (s=nsyllable_phon_slots[sl], slotUnit=unit ; *s ; s++, slotUnit++) 
      (*slotUnit)->output /= sum ;
  }
}

/* ASSUMES NSYLLABLE PHONOLOGICAL CODING */
void	SMnormalizeInputDeriv(group)
  Group	group ;
{
  int	sl, unitIdx ;
  char	*s ;
  Real	sum ;
  Unit *unit, *slotUnit ;

  /*
   * dE/dx_k = y_k * [dEdy_k - sum_j(y_j * dEdy_j)]
   *    where j runs over units in slot containing k
   */
  for (sl=0, unit=group->unit ; 
       sl<NSYL_N_PHON_SLOTS ;
       unit += strlen(nsyllable_phon_slots[sl]), sl++) {
    sum = 0.0 ;
    for (s=nsyllable_phon_slots[sl], slotUnit=unit ; *s ; s++, slotUnit++)
      sum += (*slotUnit)->output * (*slotUnit)->outputDeriv ;
    for (s=nsyllable_phon_slots[sl], slotUnit=unit ; *s ; s++, slotUnit++) 
      (*slotUnit)->inputDeriv = (*slotUnit)->output * ((*slotUnit)->outputDeriv - sum) ;
  }
}

/* ASSUMES (N)SYLLABLE PHONOLOGICAL CODING */
Real	SMnormalizedRespProb(group, criterion, sd)
  Group	group ;
  double criterion, sd ;
{
  
  int	idx, sl, unitIdx ;
  char	*s ;
  Unit *unit, *slotUnit ;
  Real	sum, max, none, prob = 1.0 ;

  switch (group->numUnits) {
  case SYL_N_PHONEMES:
    sum = 0.0 ;
    max = -1.0 ;
    for (idx = 0, unit = group->unit ; idx < group->numUnits ; idx++, unit++) {
      Real	output = (*unit)->output ;
      if (idx >= SYL_N_PHON_ONSETS && idx < SYL_N_PHON_ONSETS+SYL_N_PHON_VOWELS) {
	/* vowel */
	sum += output ;
	if (output > max) max = output ;
	if (idx == SYL_N_PHON_ONSETS+SYL_N_PHON_VOWELS-1) 
	  /* after last vowel */
	  prob *= (sd > 0.0 ? poz((max/sum - criterion)/sd) : max/sum) ;
      } else {
	/* consonant */
	prob *= (sd > 0.0 ? 
		 poz((MAX(output, 1.0-output) - criterion)/sd) :
		 MAX(output, 1.0-output)) ;
      }
    }
    return (sd > 0.0 ?
	    /* geometric mean across slots */
	    pow(prob, 1.0/(double)(SYL_N_PHON_ONSETS+1+SYL_N_PHON_CODAS)) :
	    (prob > criterion ? 1.0 : 0.0)) ;
  case NSYL_N_PHONEMES:
    for (sl=0, unit=group->unit ; 
	 sl<NSYL_N_PHON_SLOTS ;
	 unit += strlen(nsyllable_phon_slots[sl]), sl++) {
      sum = 0.0 ;
      max = -1.0 ;
      for (s=nsyllable_phon_slots[sl], slotUnit=unit ; *s ; s++, slotUnit++) {
	Real output = (*slotUnit)->output ;
	sum += output ;
	if (output > max) max = output ;
      }
      none = 1.0 - sum ;
      if (sl!=NSYL_PHON_VOWEL_SLOT) {
	if (none > max) max = none ;
	if (sum < 1.0)  sum = 1.0 ;
      }
      /* z-score of normalized maximum output (possibly null) */
      prob *= (sd > 0 ? poz((max/sum - criterion)/sd) : max/sum) ;
    }
    return (sd > 0.0 ?
	    /* geometric mean across slots */
	    pow(prob, 1.0/(double)NSYL_N_PHON_SLOTS) :
	    (prob > criterion ? 1.0 : 0.0)) ;
  }
}

Boolean	SMoutputClean(group, tolerance)
  Group	group ;
  double tolerance ;	/* roughly, min difference from 0.5 */
{
  
  int	idx, sl, unitIdx ;
  char	*s ;
  Unit *unit, *slotUnit ;
  Real	max, next, none, prob = 1.0 ;

  switch (group->numUnits) {
  case SYL_N_PHONEMES:
    max = -1.0 ;
    for (idx = 0, unit = group->unit ; idx < group->numUnits ; idx++, unit++) {
      Real	output = (*unit)->output ;
      if (idx >= SYL_N_PHON_ONSETS && idx < SYL_N_PHON_ONSETS+SYL_N_PHON_VOWELS) {
	/* vowel */
	if (output > max) {
	  next = max ;
	  max = output ;
	} else if (output > next) {
	  next = output ;
	}
	if (idx == SYL_N_PHON_ONSETS+SYL_N_PHON_VOWELS-1) 
	  /* after last vowel */
	  if (max-next < tolerance) return FALSE ;
      } else {
	/* consonant */
	if (fabs(output - 0.5) < tolerance) return FALSE ;
      }
    }
    return TRUE ;
  case NSYL_N_PHONEMES:
    for (sl=0, unit=group->unit ; 
	 sl<NSYL_N_PHON_SLOTS ;
	 unit += strlen(nsyllable_phon_slots[sl]), sl++) {
      max = -1.0 ; next = -1.0 ;
      for (s=nsyllable_phon_slots[sl], slotUnit=unit ; *s ; s++, slotUnit++) {
	Real output = (*slotUnit)->output ;
	if (output > max) {
	  next = max ;
	  max = output ;
	} else if (output > next) {
	  next = output ;
	}
      }
      if (sl!=NSYL_PHON_VOWEL_SLOT) next = MAX(next, 0.5) ;
      if (fabs(max-next) < tolerance) return FALSE ;
    }
    return TRUE ;
  default:
    for (unitIdx=0, unit=group->unit; 
	 unitIdx < group->numUnits;
	 unitIdx++, unit++) 
      if (fabs((*unit)->output - 0.5) < tolerance) return FALSE ;
    return TRUE ;
  }
}

static void code_slots(str, slots, n_slots)
  String str, slots[] ;
  int n_slots ;
{
  int	i, j, pos ;
  char *s, c, affr=NULL ;

  for (i = 0 ; i < n_slots ; i++) {
    if (*str == BLANK && i == NSYL_PHON_VOWEL_SLOT) str++ ;
    for (j = 0, c = NULL, s = slots[i] ; *s ; j++, s++) {
      if ((*s == 'P' && !strncmp(str,"ps",2)) ||
	  (*s == 'K' && !strncmp(str,"ks",2)) ||
	  (*s == '+' && !strncmp(str,"ts",2))) {
	fprintf(dout, "1 ") ;
	if (c) fprintf(dout, "\n#WARNING: multiple hits in slot %d\n", i) ;
	c    = *s ;
	affr = *str ;
	str++ ;
      } else if (*s == affr) {
	fprintf(dout, "1 ") ;
	if (c) fprintf(dout, "\n#WARNING: multiple hits in slot %d\n", i) ;
	c    = affr ;
      } else if (*str == *s) {
	fprintf(dout, "1 ") ;
	if (c) fprintf(dout, "\n#WARNING: multiple hits in slot %d\n", i) ;
	c = *s ;
	str++ ;
      } else 
	fprintf(dout, "0 ") ;
    }
    if (c) fprintf(dout, "#%c", c) ;
    fprintf(dout, "\n") ;
  }
  if (*str) fprintf(dout, "\n#WARNING: string not empty: \"%s\"\n", str) ;
}

static	void	encode_phonology(rep, entry)
  int		rep ;
  SMentry	*entry ;
{
  String	str = entry->phonemes, gstr = entry->graphemes ;
  String	*seq_phons = sequential_phonemes ;
  int		n_seq_phons = SEQ_N_PHONEMES ;
  int		idx, nalts, attn_shift, attn_pos = 0, done = FALSE, minTime ;

  minTime = _offset ;
  switch (rep) {
  case GRAPHEMES:
    encode_orthography(rep, entry, NULL) ;
    break ;
  case SYLLABLE:
    if (_recurrent) fprintf(dout,"[last %d]\n", _offset) ;
    idx   = 0 ;     nalts = SYL_N_PHON_ONSETS ;
    str  += code_inclusive(str, phoneme_onset_end(str), &(syllable_phonemes[idx]), nalts, NULL) ;
    idx  += nalts ; nalts = SYL_N_PHON_VOWELS ;
    str  += code_exclusive_multi(str, &(syllable_phonemes[idx]), nalts, NULL) ;
    idx  += nalts ; nalts = SYL_N_PHON_CODAS ;
    code_inclusive(str, strlen(str), &(syllable_phonemes[idx]), nalts, NULL) ;
    fprintf(dout, "\n") ;
    break ;
  case NSYLLABLE:
    if (_recurrent) fprintf(dout,"[last %d]\n", _offset) ;
    code_slots(str, nsyllable_phon_slots, NSYL_N_PHON_SLOTS) ;
    break ;
  case NSEQUENTIAL:
    seq_phons = nsequential_phonemes ;
    n_seq_phons = NSEQ_N_PHONEMES ;
  case SEQUENTIAL:
    if (*str == BLANK) str++ ;
    do {
      if (! *str) done = TRUE ;
      fprintf(dout, "[%d]\n", minTime++) ; 
      str += code_exclusive_multi(str, seq_phons, n_seq_phons, NULL) ;
    } while (!done) ;
    break ;
  case NATTENTION:
    seq_phons = nsequential_phonemes ;
    n_seq_phons = NSEQ_N_PHONEMES ;
  case ATTENTION:
    if (*str == BLANK) str++ ;
    do {
      if (! *str) done = TRUE ;
      fprintf(dout, "[%d]\n", minTime++) ; 
      if ((attn_shift=numGraphemesPerPhoneme(gstr, str)) >= 0) {
	attn_pos += attn_shift ;
	gstr     += attn_shift ;
      } else
	attn_pos = -1 ;
      str += code_exclusive_multi(str, seq_phons, n_seq_phons, NULL) ;
      fprintf(dout, "\n") ;
      for (idx = 0 ; idx < N_ATTENTION_POSITIONS ; idx++) 
	fprintf(dout, "%d ", (!done && idx==attn_pos)) ;
      if (done) fprintf(dout, "1\n") ;
      else      fprintf(dout, "0\n") ;
    } while (!done) ;
    break ;
  default:
    IErrorAbort("encode_phonology: unknown phonological encoding") ;
  }
}

static String 	decode_phonology(rep, group, responseWithStates)
  int	rep ;
  Group	group ;
     String responseWithStates ;
{
  int 		idx, wtaIdx, time, maxTime, numOn, sl ;
  Real		wtaOutput = -MAXFLOAT, wtaInput = -MAXFLOAT, affr_state, s_state = MAXFLOAT ;
  Boolean	done = FALSE;
  int		numUnits = group->numUnits ;
  int		vowel_idx = SYL_N_PHON_ONSETS ;
  String*	seq_phons = sequential_phonemes ;
  Unit		*unit ;
  String	s, vowel ;
  char		c, affr = NULL, *affricate = NULL ;
  char		thisResp[RLEN], thisRespWithStates[RLEN], response[RLEN] ;
  Group		Attention, Done ;

  response[0] = NULL;
  if (responseWithStates) responseWithStates[0] = NULL ;
  switch (rep) {
  case GRAPHEMES:
    /* this depends on renameGroup */
    for (idx = 0, unit = group->unit ; idx < numUnits ; idx++, unit++) {
      Real	state = (*unit)->output ;
      if (state >= _on_threshold) {
	sprintf(response, "%s%s ", response, (*unit)->name) ;
	if (responseWithStates) {
	  sprintf(responseWithStates, "%s%s ", responseWithStates, (*unit)->name) ;
	  if (state < _on_threshold+_print_tolerance)
	    sprintf(responseWithStates, "%s(%.2f) ", responseWithStates, state);
	}
      } else if (responseWithStates && state >= _on_threshold-_print_tolerance) {
	sprintf(responseWithStates, "%s(%s%.2f) ", responseWithStates, (*unit)->name, state);
      }
    }
    return response ;
  case NSYLLABLE:
    for (sl = 0, unit = group->unit ; sl < NSYL_N_PHON_SLOTS ; unit += strlen(nsyllable_phon_slots[sl]), sl++) {
      Unit	*slotUnit ;
      Boolean	paren = FALSE ;
      wtaOutput = -MAXFLOAT ;
      for (s = nsyllable_phon_slots[sl], slotUnit = unit ; *s ; s++, slotUnit++) {
	Real output = (*slotUnit)->output ;
	if (output > wtaOutput) { 
	  wtaOutput = output ;  
	  if (_norm) wtaInput = (*slotUnit)->totalInput ;
	  c = *s ;
	}
      }
      if (sl == NSYL_PHON_VOWEL_SLOT ||
	  (!_norm && wtaOutput >= _on_threshold) ||
	  (_norm  && wtaInput  >= 0.0)) {
	if (strchr("PK+", c)) {
	  affr = (c=='P'?'p':(c=='K'?'k':'t')) ;
	  affr_state = wtaOutput ;
	} else if (c == 's' && affr) {
	  s_state = wtaOutput ;
	} else {
	  sprintf(response, "%s%c", response, c) ;
	  if (responseWithStates) {
	    sprintf(responseWithStates, "%s%c", responseWithStates, c) ;
	    if (wtaOutput < _on_threshold+_print_tolerance) {
	      sprintf(responseWithStates, "%s(%.2f", responseWithStates, wtaOutput) ;
	      paren = TRUE ;
	    }
	    for (s = nsyllable_phon_slots[sl], slotUnit = unit ; *s ; s++, slotUnit++) {
	      Real output = (*slotUnit)->output ;
	      if (*s != c && output >= _on_threshold-_print_tolerance) {
		if (!paren) {
		  sprintf(responseWithStates, "%s(%.2f", responseWithStates, wtaOutput) ;
		  paren = TRUE ;
		}
		sprintf(responseWithStates, "%s%c%.2f", responseWithStates, *s, output) ;
	      }
	    } /* slot */
	    if (paren) sprintf(responseWithStates, "%s)", responseWithStates) ;
	  } /* responseWithStates */
	  if (c == affr && s_state != MAXFLOAT) {
	    sprintf(response, "%ss", response) ;
	    if (responseWithStates) {
	      sprintf(responseWithStates, "%ss", responseWithStates) ;
	      if ((affr_state < _on_threshold+_print_tolerance) ||
		  (s_state    < _on_threshold+_print_tolerance)) 
		sprintf(responseWithStates, "%s(%.2f%cs%.2f)",
			responseWithStates, s_state, affr, affr_state) ;
	    } /* responseWithStates */
	  } /* affricates */
	}
      } /* wtaOutput */
    } /* slot */
    return response ;
  case SYLLABLE:
    for (idx = 0, unit = group->unit ; idx < numUnits ; idx++, unit++) {
      Real	state = (*unit)->output ;
      String	phon = syllable_phonemes[idx] ;
      if (_wta && idx >= vowel_idx && idx < vowel_idx+SYL_N_PHON_VOWELS) {
	/* keep track of most active vowel */
	if (state > wtaOutput) { vowel = phon ; wtaOutput = state ; }
	/* print after last vowel */
	if (idx == vowel_idx+SYL_N_PHON_VOWELS-1) {
	  strcat(response, vowel) ;
	  if (responseWithStates) {
	    strcat(responseWithStates, vowel) ;
	    if (fabs(wtaOutput-_on_threshold) < _print_tolerance)
	      sprintf(responseWithStates, "%s(%.2f)", responseWithStates, wtaOutput) ;
	  }
	}
      } else if (state >= _on_threshold) {
	if (strlen(phon) > 1) affricate = phon ;
	else if (affricate && strstr(affricate, phon)) continue ;
	strcat(response, phon) ;
	if (responseWithStates) {
	  strcat(responseWithStates, phon) ;
	  if (state < _on_threshold+_print_tolerance)
	    sprintf(responseWithStates, "%s(%.2f)", responseWithStates, state);
	}
      } else if (responseWithStates && state >= _on_threshold-_print_tolerance) {
	sprintf(responseWithStates, "%s(%s%.2f)", responseWithStates, phon, state);
      }
    }
    return response ;
  case NSEQUENTIAL:
  case NATTENTION:
    seq_phons = nsequential_phonemes ;
  case SEQUENTIAL:
  case ATTENTION:
    response[0] = NULL;
    if (rep==ATTENTION || rep==NATTENTION) {
      if ((Done=groupFromName(group->net, "Done")) == NULL) IErrorAbort("no \"Done\" group") ;
      if ((Attention=groupFromName(group->net, "Attention")) == NULL) IErrorAbort("no \"Attention\" group") ;
    }
/*
    maxTime = group->net->currentTime ;
*/
    maxTime = group->net->timeSlices ;
    for (time = _offset ; time < maxTime ; time++) {
      thisResp[0] = NULL ;      thisRespWithStates[0] = NULL ;
      numOn = 0 ;
      wtaOutput = -MAXFLOAT ;

      /* get active phoneme(s) */
      for (idx = 0, unit = group->unit ; idx < numUnits ; idx++, unit++) {
	Real	state = (*unit)->outputHistory[time] ;
	if (_wta) {
	  if (state > wtaOutput) { wtaIdx = idx ;  wtaOutput = state ; }
	} else if (state >= _on_threshold) {
	  numOn++ ;
	  strcat(thisResp, seq_phons[idx]) ;
	  if (responseWithStates) {
	    strcat(thisRespWithStates, seq_phons[idx]) ;
	    if (fabs(wtaOutput-_on_threshold) < _print_tolerance)
	      sprintf(thisRespWithStates, "%s(%.2f)", thisRespWithStates, wtaOutput) ;
	  }
	} else if (responseWithStates && state >= _on_threshold-_print_tolerance) {
	  sprintf(thisRespWithStates, "%s(%s%.2f)", thisRespWithStates, seq_phons[idx], state);
	}
      }	/* unit */
      if (rep==ATTENTION || rep==NATTENTION)
	done = maxOutputAt(Done,time) > maxOutputAt(Attention,time) ;
      if (_wta) {
	if (done) numOn = 0 ;		/* thisResp == NULL */
	else {
	  strcat(thisResp, seq_phons[wtaIdx]) ;
	  if (responseWithStates) {
	    strcat(thisRespWithStates, seq_phons[wtaIdx]) ;
	    if (fabs(wtaOutput-_on_threshold) < _print_tolerance)
	      sprintf(thisRespWithStates, "%s(%.2f)", thisRespWithStates, wtaOutput) ;
	  }
	  numOn = 1 ;
	}
      }
      /* check if done */
      if (rep==ATTENTION || rep==NATTENTION) {
	if (numOn == 1 || (numOn == 0 && done)) {
	  strcat(response, thisResp) ;
	  if (responseWithStates) strcat(responseWithStates, thisRespWithStates) ;
	} else {
	  sprintf(response, "%s[%s]", response, thisResp) ;
	  if (responseWithStates) sprintf(responseWithStates, "%s[%s]", responseWithStates, thisRespWithStates) ;
	}
	if (done) break ;
      } else {
	if (numOn > 1) {
	  sprintf(response, "%s[%s]", response, thisResp) ;
	  if (responseWithStates) sprintf(responseWithStates, "%s[%s]", responseWithStates, thisRespWithStates) ;
	} else if (numOn == 1) {
	  strcat(response, thisResp) ;
	  if (responseWithStates) strcat(responseWithStates, thisRespWithStates) ;
	} else		     break ;
      }
    }
    return response ;
  default:
    IErrorAbort("decode_phonology: unknown phonological encoding") ;
  }
}

/**********************************************************************/
static	void	renameLink(link, data)
  Link	link ;
  void	*data ;
{
  char	buffer[WLEN] ;
  sprintf(buffer, "%s %s %s", 
	  link->preUnit->name,
	  (link->type&BIDIRECTIONAL ? "<->" : "->"),
	  link->postUnit->name) ;
  link->name = strdup(buffer) ;
}

static	void	renameUnitLinks (unit, data)
  Unit	unit ;
  void	*data ;
{
  unitForAllLinks (unit, ALL, renameLink, NULL) ;
}

static	void	renameGroup(group, data)
  Group	group ;
  void	*data ;
{
  int		sl, unitIdx, letIdx ;
  char		c, *s, buffer[WLEN] ;
  Unit		*unit ;
  String*	seq_phons = sequential_phonemes ;
  Boolean	seq_rep = (groupFromName(currentNet, "Attention")!=NULL) ;

  if (!strcmp(group->name,"Letters")) {
    for (unitIdx = 0, unit = group->unit ; unitIdx < group->numUnits ; unitIdx++, unit++) {
      letIdx = unitIdx%(MARK_LETTER_SPACE+N_LETTERS) ;
      c = (MARK_LETTER_SPACE ? (letIdx ? letters[letIdx-1] : BLANK) : letters[letIdx]) ;
      sprintf(buffer, "%c/%d", c, unitIdx/(MARK_LETTER_SPACE+N_LETTERS)-N_LEFT_OF_FIXATION) ;
      (*unit)->name = strdup(buffer) ;
    }
  } else if (!strcmp(group->name,"Graphemes")) {
    if (group->numUnits == SYL_N_GRAPHEMES) {
      for (unitIdx = 0, unit = group->unit ; unitIdx < SYL_N_ONSET_GRAPHEMES ; unitIdx++, unit++)
	(*unit)->name = strdup(onset_graphemes[unitIdx]) ;
      for (unitIdx = 0 ; unitIdx < SYL_N_VOWEL_GRAPHEMES ; unitIdx++, unit++) 
	(*unit)->name = strdup(vowel_graphemes[unitIdx]) ;
      for (unitIdx = 0 ; unitIdx < SYL_N_FINAL_GRAPHEMES ; unitIdx++, unit++) 
	(*unit)->name = strdup(final_graphemes[unitIdx]) ;
    } else if (group->numUnits == NSYL_N_GRAPHEMES) {
      for (unitIdx = 0, unit = group->unit ; unitIdx < NSYL_N_ONSET_GRAPHEMES ; unitIdx++, unit++)
	(*unit)->name = strdup(nsyl_onset_graphemes[unitIdx]) ;
      for (unitIdx = 0 ; unitIdx < NSYL_N_VOWEL_GRAPHEMES ; unitIdx++, unit++) 
	(*unit)->name = strdup(nsyl_vowel_graphemes[unitIdx]) ;
      for (unitIdx = 0 ; unitIdx < NSYL_N_FINAL_GRAPHEMES ; unitIdx++, unit++) 
	(*unit)->name = strdup(nsyl_final_graphemes[unitIdx]) ;
    }
  } else if (!strcmp(group->name,"Phonemes")) {
    if (!seq_rep && group->numUnits == NSYL_N_PHONEMES) {
      for (sl = 0, unit = group->unit ; sl < NSYL_N_PHON_SLOTS ; sl++) {
	for (s = nsyllable_phon_slots[sl] ; *s ; s++, unit++) {
	  sprintf(buffer, "/%c/", *s) ;
	  (*unit)->name = strdup(buffer) ;
	}
      }
    } else {
      if (seq_rep && group->numUnits == NSEQ_N_PHONEMES)
	seq_phons = nsequential_phonemes ;
      for (unitIdx = 0, unit = group->unit ; unitIdx < group->numUnits ; unitIdx++, unit++) {
	if (seq_rep) sprintf(buffer, "/%s/", seq_phons[unitIdx]) ;
	else         sprintf(buffer, "/%s/", syllable_phonemes[unitIdx]) ;
	(*unit)->name = strdup(buffer) ;
      }  
    }
  } else return ;
  groupForAllUnits(group, renameUnitLinks, NULL) ;
}
/*****************************************************************************/
static Boolean	outputGroupCorrect(group)
  Group	group ;
{
  int	unitIdx, groupIdx ;
  Unit	*unit ;

  for (unitIdx = 0, unit = group->unit ;
       unitIdx < group->numUnits ; 
       unitIdx++, unit++) {
    Real	output = (*unit)->output ;
    Real	target = (*unit)->target ;
    if ((target == 1.0 && output < _on_threshold) ||
	(target == 0.0 && output > _on_threshold))
      return FALSE ;
  }
  return TRUE ;
}
/*****************************************************************************
 * loadSMlist / writeSMlist / writeSMexamples
 *****************************************************************************/

static	SMlist	*defaultSMlist ;

#define MIN_SM_ENTRIES	128

/*****************************************************************************/
int command_loadSMlist (tokc, tokv)
  int	tokc ;
  String	tokv[] ;
{
  int		clear = FALSE ;
  int		exampleType = TRAINING;
  ExampleSet	exampleSet = NULL ;
  Real 		freq;
  FILE		*fp ;
  String	type = NULL, file_name = NULL;
  char		code[WLEN], gstr[WLEN], pstr[WLEN], name[RLEN], *tmp_pstr;
  SMlist	*list ;
  struct ARG_DESC *argd ;

  argd = StartArgs(tokv[0]);
  Args(argd, "[-Ptype <type>%s%?]",&type,      "type of example set (one of TRAINING TESTING VALIDATION)");
  Args(argd, "[-Pclear %P%?]", 	   &clear,     "clear SMlist before loading");
  Args(argd, "<filename>%s", 	   &file_name, "file to read from");
  EndArgs(argd);
  if (GiveHelp(tokc)) {
    ISynopsis("loads SM list") ;
    IHelp
      (tokc, tokv[0], NULL, synopsis,
       NULL);
    return 0;
  }
  (void) ParseArgs(argd, tokc, tokv, 0);
  if (currentNet) {
    netForAllGroups(currentNet, INPUT|OUTPUT, renameGroup, NULL) ;
    if (type) exampleType = findMask(type) ;
    exampleSet = netGetExampleSet(currentNet, exampleType) ;
    if (exampleSet == NULL)
      IErrorAbort("No example set specified for net \"%s\".", currentNet->name) ;
    list = (SMlist *)getExampleSetData(exampleSet) ;
  } else
    list = defaultSMlist ;
  if (list == NULL) {
    list 		= (SMlist *)malloc(sizeof(SMlist)) ;
    list->size 		= 0 ;
    list->numEntries 	= 0 ;
    list->entries 	= NULL ;
    if (currentNet) putExampleSetData(exampleSet, (void *)list) ;
    else            defaultSMlist = list ;
  }
  if (list->size == 0) 
    list->entries = (SMentry *)calloc(list->size=MIN_SM_ENTRIES, sizeof(SMentry)) ;
  if (clear) list->numEntries = 0 ;

  fp = IOpenFileOrAbort(file_name, "r", NULL) ;
  while (fscanf(fp, "%s%s%s%f", code, gstr, pstr, &freq) != EOF) {
    if (list->numEntries >= list->size) 
      extendList((void *)&(list->entries), sizeof(SMentry), list->size, list->size=2*list->size);
    list->entries[list->numEntries].code      = strdup(code) ;
    list->entries[list->numEntries].graphemes = strdup(gstr) ;
    list->entries[list->numEntries].phonemes  = strdup(pstr) ;
    list->entries[list->numEntries].frequency = freq ;
    if (exampleSet) {
      sprintf(name, "%-9s/%s/", gstr, pstr) ;
      exampleSet->example[list->numEntries].name = strdup(name) ;
      exampleSet->example[list->numEntries].frequency = freq ;
    }
    list->numEntries++ ;
  }
  ICloseFile(fp, NULL) ;
  return 1;
}
/*****************************************************************************/
static	void	writeGPCs(list)
  SMlist *list ;
{
  int idx ;

  for (idx = 0 ; idx < list->numEntries ; idx++) {
    SMentry	*entry = &(list->entries[idx]) ;
    printGPCs(entry->graphemes, entry->phonemes) ;
  }
}

static	void	writeSMlist(list)
  SMlist *list ;
{
  int idx ;

  for (idx = 0 ; idx < list->numEntries ; idx++) {
    SMentry	*entry = &(list->entries[idx]) ;
    fprintf(dout, "%s\t%s\t%s\t%f\n",
	    entry->code, entry->graphemes, entry->phonemes, entry->frequency);
  }
}

int command_writeSMlist (tokc, tokv)
  int	tokc ;
  String	tokv[] ;
{
  int		exampleType = TRAINING, gpc = FALSE ;
  ExampleSet	exampleSet ;
  FILE		*fp ;
  String	type = NULL ;
  SMlist	*list ;
  struct ARG_DESC *argd ;

  argd = StartArgs(tokv[0]);
  Args(argd, "[-Ptype <type>%s%?]", &type,	"type of example set (one of TRAINING TESTING VALIDATION)");
  Args(argd, "[-Pgpc %P%?]",	    &gpc,	"parse graphemes and phonemes") ;
  EndArgs(argd);
  if (GiveHelp(tokc)) {
    ISynopsis("writes out SM list") ;
    IHelp
      (tokc, tokv[0], NULL, synopsis,
       NULL);
    return 0;
  }
  (void) ParseArgs(argd, tokc, tokv, 0);
  if (currentNet) {
    if (type) exampleType = findMask(type) ;
    exampleSet = netGetExampleSet(currentNet, exampleType) ;
    if (exampleSet == NULL)
      IErrorAbort("No example set specified for net \"%s\".", currentNet->name) ;
    list = (SMlist *)getExampleSetData(exampleSet) ;
  } else
    list = defaultSMlist ;
  if (list == NULL) 
    IErrorAbort("No SM list found") ;

  if (gpc)
    writeGPCs(list) ;
  else
    writeSMlist(list) ;
  return 1;
}

/*****************************************************************************/
int command_writeSMexamples (tokc, tokv)
  int	tokc ;
  String	tokv[] ;
{
  int		idx, orth_rep, phon_rep, exampleType = TRAINING ;
  ExampleSet	exampleSet ;
  Real 		freq;
  String	type=NULL, orth=NULL, phon=NULL, pstr=NULL;
  SMlist	*list ;
  struct ARG_DESC *argd ;

  _recurrent 	= FALSE ;
  _offset	= 3 ;
  _overlap	= 0.3 ;
  argd = StartArgs(tokv[0]);
  Args(argd, "[-Ptype <type>%s%?]",	  &type,	"type of example set [TRAINING TESTING VALIDATION]");
  Args(argd, "[-Precurrent %P%?]",	  &_recurrent,	"example list is for recurrent net");
  Args(argd, "[-Poffset <n>%i%?]",	  &_offset,	"phonology timeSlice offset") ;
  Args(argd, "[-Poverlap <real>%r%?]", 	  &_overlap,	"amount of letter overlap in position-based representation");
  Args(argd, "[-Porthography <str>%s%?]", &orth,	"orthography coding scheme [syllable letters none]") ;
  Args(argd, "[-Pphonology <str>%s%?]",	  &phon,	"phonology coding scheme [(n)syllable (n)sequential (n)attention]") ;
  EndArgs(argd);
  if (GiveHelp(tokc)) {
    ISynopsis("writes examples from SM list") ;
    IHelp
      (tokc, tokv[0], NULL, synopsis,
       NULL);
    return 0;
  }
  (void) ParseArgs(argd, tokc, tokv, 0);
  orth_rep = representation(orth) ;
  phon_rep = representation(phon) ;
  
  fprintf(dout, "# ") ;
  for (idx = 0 ; idx < tokc ; idx++) fprintf(dout, "%s ", tokv[idx]) ;
  fprintf(dout, "\n\n") ;

  /* SM LIST */
  if (currentNet) {
    if (type) exampleType = findMask(type) ;
    exampleSet = netGetExampleSet(currentNet, exampleType) ;
    if (exampleSet == NULL)
      IErrorAbort("No example set specified for net \"%s\".", currentNet->name) ;
    list = (SMlist *)getExampleSetData(exampleSet) ;
  } else
    list = defaultSMlist ;
  if (list == NULL) 
    IErrorAbort("No SM list found") ;

  for (idx = 0 ; idx < list->numEntries ; idx++) {
    SMentry	*entry = &(list->entries[idx]) ;

    /* HEADER */
    fprintf(dout, "# %d %s %s %f %s\n",
	    idx, entry->graphemes, entry->phonemes, entry->frequency, entry->code) ;
    fflush(dout) ;

    /* ORTHOGRAPHY */
    if (orth_rep == NONE) 
      /* phon => phon mapping */
      encode_phonology(phon_rep, entry) ;
    else
      encode_orthography(orth_rep, entry, NULL) ;
    fprintf(dout,",\n") ;  
    fflush(dout) ;
     
    /* PHONOLOGY */
    encode_phonology(phon_rep, entry) ;
    fprintf(dout,";\n\n") ; 
    fflush(dout) ;
  }
  return 1;
}

/*****************************************************************************
 * pronounce
 *****************************************************************************/
int command_pronounce (tokc, tokv)
  int	tokc ;
  String	tokv[] ;
{
  int		rep, i, e, numUnits ;
  Real		sse, sd = 0.0, lesion_prob = -1.0 ;
  ExampleSet	exampleSet ;
  SMlist	*list ;
  Group		input_group, output_group, hidden_group;
  Boolean	error ;
  String	input_group_name  = NULL, output_group_name = NULL ;
  String	hidden_group_name = NULL ;
  String	type=NULL, code=NULL, word=NULL ;
  String	nonword=NULL, resp=NULL, orth=NULL, phon=NULL, re=NULL, reCompiled ;
  String	responseWithStates ;
  int		orth_rep, phon_rep, reps = 1 ;
  int		bidir = FALSE ;
  int		ex, example_n = -1, exampleType = TRAINING;
  int		errors = FALSE, tolerance_correct = FALSE ;
  String 	preGroupRE = NULL, postGroupRE = NULL ;
  Group		Phonemes ;
  Group		*preGroupArray,  *preGroup, *postGroupArray, *postGroup ;
  struct ARG_DESC *argd ;

  if (currentNet == NULL) IErrorAbort("There is no current net.") ;
  if (groupFromName(currentNet, "Letters") != NULL)   orth_rep = LETTERS ;
  if (groupFromName(currentNet, "Attention") != NULL) {
    if ((Phonemes=groupFromName(currentNet, "Phonemes"))==NULL) 
      IErrorAbort("no \"Phonemes\" group") ;
    else if (Phonemes->numUnits==NSYL_N_PHONEMES) 
      phon_rep = NATTENTION ;
    else
      phon_rep = ATTENTION ;
  }
  _wta		= FALSE ;
  _nopr		= FALSE ;
  _norm		= FALSE ;
  _offset	= 0 ;
  _on_threshold = 0.5 ;
  _print_tolerance = 0.0 ;
  _print_threshold = MAXFLOAT ;
  _impStart = 0 ; _impNum = 0 ;
  argd = StartArgs(tokv[0]);
  Args(argd, "[-Ptype <type>%s%?]", 	    &type, 		"type of example set (one of TRAINING TESTING VALIDATION)");
  Args(argd, "[-PinputGroup <name>%s%?]",   &input_group_name,	"name of input group");
  Args(argd, "[-PoutputGroup <name>%s%?]",  &output_group_name, "name of output group");
  Args(argd, "[-PhiddenGroup <name>%s%?]",  &hidden_group_name, "name of hidden group (for states/correlations analysis)");
  Args(argd, "[-PpreGroup <regexp>%s%?]",   &preGroupRE,	"regular expression specifying source groups.");
  Args(argd, "[-PpostGroup <regexp>%s%?]",  &postGroupRE,	"regular expression specifying destination groups.");
  Args(argd, "[-Pexample <n>%i%?]", 	    &example_n, 	"example to run");
  Args(argd, "[-Pcode <str>%s%?]", 	    &code, 		"only run examples with code <string>") ;
  Args(argd, "[-Pword <str>%s%?]", 	    &word, 		"only run word <string>") ;
  Args(argd, "[-Pre <regexp>%s%?]", 	    &re, 		"only run words matching <regexp>") ;
  Args(argd, "[-Pnonword <string>%s%?]",    &nonword, 		"use nonword <string> as input") ;
  Args(argd, "[-Perrors %P%?]", 	    &errors, 		"only print errors");
  Args(argd, "[-PuseToleranceCorrect %P%?]",&tolerance_correct,	"use tolerance correctness of output as test");
  Args(argd, "[-Plesion <real>%r%?]",	    &lesion_prob,	"proportion of connections (specified by preGroup and postGroup) to remove");
  Args(argd, "[-Pbidirectional %P%?]", 	    &bidir, 		"lesion/add noise bidirectionally");
  Args(argd, "[-Porthography <str>%s%?]",   &orth,		"orthography coding scheme [(n)syllable letters none]") ;
  Args(argd, "[-Pphonology <str>%s%?]",	    &phon,		"phonology coding scheme [(n)syllable (n)sequential (n)attention]") ;
  Args(argd, "[-PonThreshold <real>%r%?]",  &_on_threshold,	"threshold for active phonemes");
  Args(argd, "[-PtolerancePrint <real>%r%?]",&_print_tolerance,	"tolerance around onThreshold for printing states of phonemes");
  Args(argd, "[-Psd <real>%r%?]", 	    &sd,		"standard deviation for adding noise to weights");
  Args(argd, "[-Pwta %P%?]",		    &_wta,		"winner-take-all (most active phoneme) decoding") ;
  Args(argd, "[-Pnopr %P%?]",		    &_nopr,		"do not print activity levels of units") ;
  Args(argd, "[-Pnorm %P%?]",		    &_norm,		"normalized output units") ;
  Args(argd, "[-Poffset <n>%i%?]",	    &_offset,		"phonology timeSlice offset") ;
  Args(argd, "[-Preps <n>%i%?]",	    &reps,		"number of replications to run") ;
  EndArgs(argd);
  if (GiveHelp(tokc)) {
    ISynopsis("pronounce words") ;
    IHelp
      (tokc, tokv[0], NULL, synopsis,
       NULL);
    return 0;
  }
  ParseArgs(argd, tokc, tokv, 0);
  if (_nopr) responseWithStates = NULL ;
  else responseWithStates = ((String)calloc(RLEN, sizeof(char))) ;
  orth_rep = representation(orth) ;
  phon_rep = representation(phon) ;
  if (!input_group_name)  input_group_name  = default_orthography_group_name(orth_rep) ;
  if (!output_group_name) output_group_name = default_phonology_group_name(phon_rep) ;
  if (!hidden_group_name) hidden_group_name = "Hidden" ;
  output_group = groupFromName(currentNet, output_group_name) ;
  if (output_group == NULL) IErrorAbort("pronounce: Group \"%s\" does not exists.", output_group_name) ;
  if (nonword) {
    input_group = groupFromName(currentNet, input_group_name) ;
    if (input_group == NULL) IErrorAbort("pronounce: Group \"%s\" does not exists.", input_group_name) ;
  }
  if (sd > 0.0 || lesion_prob > 0.0) {
    preGroupArray = groupsFromRegex(currentNet, preGroupRE) ;
    if (preGroupArray == NULL) IErrorAbort("Bad string: \"%s\"", preGroupRE) ;
    if (preGroupArray[0] == NULL) IErrorAbort("No match: \"%s\"", preGroupRE) ;
    postGroupArray = groupsFromRegex(currentNet, postGroupRE) ;
    if (postGroupArray == NULL) IErrorAbort("Bad string: \"%s\"", postGroupRE) ;
    if (postGroupArray[0] == NULL) IErrorAbort("No match: \"%s\"", postGroupRE) ;
  }
  if (re) reCompiled = REcompile(re) ;
  if (type) exampleType = findMask(type) ;
  exampleSet = netGetExampleSet(currentNet, exampleType) ;
  if (exampleSet == NULL) IErrorAbort("No example set specified for net \"%s\".", currentNet->name) ;
  list = (SMlist *)getExampleSetData(exampleSet) ;
  if (list == NULL) {
    fprintf(dout,"SM list is NULL - using default list.\n") ;
    list = defaultSMlist ;
  }

  for (rep = 0; rep < reps ; rep++) {
    if (sd > 0.0) 	/* adding noise to weights */
      for (preGroup = preGroupArray ; *preGroup != NULL ; ++preGroup)
	for (postGroup = postGroupArray ; *postGroup != NULL ; ++postGroup) {
	  corruptWeights(*preGroup, *postGroup, sd) ;
	  if (bidir) corruptWeights(*postGroup, *preGroup, sd) ;
	}
    else if (lesion_prob > 0.0) 	/* lesioning connections */
      for (preGroup = preGroupArray ; *preGroup != NULL ; ++preGroup)
	for (postGroup = postGroupArray ; *postGroup != NULL ; ++postGroup)
	  lesion(*preGroup, *postGroup, lesion_prob, MAXFLOAT, TRUE) ;

    fprintf(dout,"input\ttarget\toutput\tcode\tCEerr\t\tSSerr\n") ;
    fprintf(dout,"-----\t------\t------\t----\t-----\t\t-----\n") ;
    Mreset(exampleSet) ;

    for (ex = 0; ex < exampleSet->numExamples; ex++) {
      SMentry	*entry = &(list->entries[ex]) ;

      if ((example_n >= 0 && ex != example_n)		||
	  (code           && !strstr(entry->code, code))	||
	  (re		  && !regex(reCompiled, entry->graphemes))||
	  (word           && strcmp(entry->graphemes, word)))
	continue;

      if (nonword) input_orthography(orth_rep, nonword, input_group) ;
      else      MgetNumber(exampleSet, ex) ;
      currentNet->error = 0.0 ;

      MupdateNetActivities(currentNet) ;
      /* for recurrent networks, need net->currentTime to contain last slice */

      if (nonword) {
	resp = decode_phonology(phon_rep, output_group, responseWithStates) ;
	fprintf(dout,"%s\t/   /\t/%s/\n", nonword, (responseWithStates?responseWithStates:resp)) ;
	sse = 0.0 ;
      } else {
	if (errors && tolerance_correct && outputGroupCorrect(output_group)) continue ;
	resp = decode_phonology(phon_rep, output_group, responseWithStates) ;
	error = strcmp(entry->phonemes, resp) ;
	if (errors && !error) continue ;
	sse = groupSumSquaredError(output_group) ;
	fprintf(dout,"%s%s\t/%s/\t/%s/\t%s\t%f\t%f\n", 
		(error ? "*" : ""), entry->graphemes, entry->phonemes, 
		(responseWithStates?responseWithStates:resp), entry->code,
		currentNet->error, sse) ;
      }
      fflush(dout);
      if (word || nonword) break ;
    }
    if (sd > 0.0)	/* added noise to weights */
      for (preGroup = preGroupArray ; *preGroup != NULL ; ++preGroup)
	for (postGroup = postGroupArray ; *postGroup != NULL ; ++postGroup) {
	  restoreWeights(*preGroup, *postGroup) ;
	  if (bidir) restoreWeights(*postGroup, *preGroup) ;
	}
  } /* reps */
  if (lesion_prob > 0.0)	/* lesioned connections */
    for (preGroup = preGroupArray ; *preGroup != NULL ; ++preGroup)
      for (postGroup = postGroupArray ; *postGroup != NULL ; ++postGroup)
	lesion(*preGroup, *postGroup, (Real)0.0, MAXFLOAT, TRUE) ;
  return 1 ;
}
/*****************************************************************************/
static void	distortGraphemes(Graphemes, inCluster, d) 
  Group Graphemes ;
  int	inCluster ;
  Real	d ;
{
  int		t,unitIdx, minUnitIdx, maxUnitIdx ;
  int		final_e_idx = SYL_N_ONSET_GRAPHEMES+SYL_N_VOWEL_GRAPHEMES+SYL_N_FINAL_GRAPHEMES-3 ;

  switch (inCluster) {
  case 0:	/* onset */
    minUnitIdx = 0 ;
    maxUnitIdx = SYL_N_ONSET_GRAPHEMES ;
    break ;
  case 1: 	/* vowel */
    minUnitIdx = SYL_N_ONSET_GRAPHEMES ;
    maxUnitIdx = SYL_N_ONSET_GRAPHEMES+SYL_N_VOWEL_GRAPHEMES ;
    if (Graphemes->net->type&RECURRENT) 
      for (t = 0 ; t < Graphemes->net->timeSlices ; t++) 
	Graphemes->unit[final_e_idx]->clampingHistory[t] *= d ;
    else
      Graphemes->unit[final_e_idx]->output *= d ;
    break ;
  case 2: 	/* coda */
    minUnitIdx = SYL_N_ONSET_GRAPHEMES+SYL_N_VOWEL_GRAPHEMES ;
    maxUnitIdx = SYL_N_ONSET_GRAPHEMES+SYL_N_VOWEL_GRAPHEMES+SYL_N_FINAL_GRAPHEMES ;
    break ;
  }
  for (unitIdx = minUnitIdx ; unitIdx < maxUnitIdx ; unitIdx++) {
    Unit unit = Graphemes->unit[unitIdx] ;
    if (unitIdx == final_e_idx) continue ;
    if (unit->net->type&RECURRENT) 
      for (t = 0 ; t < unit->timeSlices ; t++) 
	unit->clampingHistory[t] *= d ;
    else
      unit->output *= d ;
  }
}

static String	decode_syllable_onset(Phonemes)
  Group	Phonemes ;
{
  int	idx ;
  char	response[RLEN] ;

  response[0] = NULL ;
  for (idx = 0; idx < SYL_N_PHON_ONSETS ; idx++) {
    Unit	unit = Phonemes->unit[idx] ;
    if (unit->output >= _on_threshold)
      strcat(response, syllable_phonemes[idx]) ;
  }
  return response ;
}

static String	decode_syllable_vowel(Phonemes)
  Group	Phonemes ;
{
  int 		idx, wtaIdx ;
  Real		wtaOutput = -MAXFLOAT  ;
  char		response[RLEN], *vowel ;

  response[0] = NULL ;
  for (idx = SYL_N_PHON_ONSETS; 
       idx < SYL_N_PHON_ONSETS+SYL_N_PHON_VOWELS ; idx++) {
    Unit	unit = Phonemes->unit[idx] ;
    if (_wta) {
      if (unit->output > wtaOutput) {
	vowel = syllable_phonemes[idx] ;
	wtaOutput = unit->output ;
      }
      if (idx == SYL_N_PHON_ONSETS+SYL_N_PHON_VOWELS-1) {
	strcat(response, vowel) ;
      }
    } else {
      if (unit->output >= _on_threshold)
	strcat(response, syllable_phonemes[idx]) ;
    }
  }
  return response ;
}

static String	decode_syllable_coda(Phonemes)
  Group	Phonemes ;
{
  int	idx ;
  char	response[RLEN], *affricate = NULL ;

  response[0] = NULL ;
  for (idx = SYL_N_PHON_ONSETS+SYL_N_PHON_VOWELS; 
       idx < SYL_N_PHONEMES; idx++) {
    Unit	unit = Phonemes->unit[idx] ;
    String	phon = syllable_phonemes[idx] ;
    if (unit->output >= _on_threshold) {
      if      (strlen(phon) > 1)                     affricate = phon ;
      else if (affricate && strstr(affricate, phon)) continue ;
      strcat(response, phon) ;
    }
  }
  return response ;
}

static Boolean	clusterCorrect(Phonemes, pron, outCluster)
  Group	Phonemes ;
  String pron ;
  int 	outCluster ;
{
  int end ;
  String cluster ;
  String vowels = "a@eiouAEIOUWY^" ;

  _wta = FALSE ;
  switch (outCluster) {
  case 0: /* onset */
    end = strcspn(pron, vowels) ;
    cluster = decode_syllable_onset(Phonemes) ;
    if (end != strlen(cluster)) return FALSE ;
    if (end == 0) return TRUE ;
    if (!strncmp(pron, cluster, end)) return TRUE ;
    break ;
  case 1: /* vowel (WTA) */
  _wta = TRUE ;
  case 2: /* vowel (output > 0.5) */
    pron   += strcspn(pron, vowels) ;
    cluster = decode_syllable_vowel(Phonemes) ;
    if (!strncmp(pron, cluster, 1)) return TRUE ;
    break ;
  case 3: /* coda */
    pron   += strcspn(pron, vowels) ;
    pron   += strspn(pron, vowels) ;
    cluster = decode_syllable_coda(Phonemes) ;
    if (!strcmp(pron, cluster)) return TRUE ;
    break ;
  }
  return FALSE ;
}

/*****************************************************************************
 * distortSyllable
 *****************************************************************************/
int command_distortSyllable (tokc, tokv)
  int	tokc ;
  String	tokv[] ;
{
  int		i, e, numUnits ;
  int		inCluster, outCluster, nClusters = 3 ;
  Real		d, dLo = 0.0, dHi = 1.0, dTol = 0.001 ;
  ExampleSet	exampleSet ;
  SMlist	*list ;
  Group		Graphemes, Phonemes;
  Boolean	corr ;
  String	type=NULL, code=NULL, word=NULL, nonword=NULL, pron, re=NULL, reCompiled ;
  int		ex, example_n = -1, exampleType = TRAINING;
  struct ARG_DESC *argd ;

  _offset	= 0 ;
  _on_threshold = 0.5 ;
  _print_tolerance = 0.0 ;
  argd = StartArgs(tokv[0]);
  Args(argd, "[-Ptype <type>%s%?]", 	    &type, 		"type of example set (one of TRAINING TESTING VALIDATION)");
  Args(argd, "[-Pexample <n>%i%?]", 	    &example_n, 	"example to run");
  Args(argd, "[-Pcode <str>%s%?]", 	    &code, 		"only run examples with code <string>") ;
  Args(argd, "[-Pword <str>%s%?]", 	    &word, 		"only run word <string>") ;
  Args(argd, "[-Pre <regexp>%s%?]", 	    &re, 		"only run words matching <regexp>") ;
  Args(argd, "[-Pnonword <string>%s%?]",    &nonword, 		"use nonword <string> as input") ;
  Args(argd, "[-PonThreshold <real>%r%?]",  &_on_threshold,	"threshold for active phonemes");
  Args(argd, "[-Poffset <n>%i%?]",	    &_offset,		"phonology timeSlice offset") ;
  Args(argd, "[-PtoleranceDistortion <real>%r%?]",  &dTol, "tolerance for determining distortion threshold");
  EndArgs(argd);
  if (GiveHelp(tokc)) {
    ISynopsis("effect on output clusters of distorting input clusters") ;
    IHelp
      (tokc, tokv[0], NULL, synopsis,
       "Only works with SYLLABLE orthographic and phonological encodings.",
       NULL);
    return 0;
  }
  ParseArgs(argd, tokc, tokv, 0);
  if (currentNet == NULL) IErrorAbort("There is no current net.") ;
  /* only words with syllable representations */
  Graphemes = groupFromName(currentNet, "Graphemes") ;
  if (Graphemes == NULL) IErrorAbort("distortSyllable: Group \"Graphemes\" does not exists.") ;
  Phonemes = groupFromName(currentNet, "Phonemes") ;
  if (Phonemes == NULL) IErrorAbort("distortSyllable: Group \"Phonemes\" does not exists.") ;
  if (re) reCompiled = REcompile(re) ;
  if (type) exampleType = findMask(type) ;
  exampleSet = netGetExampleSet(currentNet, exampleType) ;
  if (exampleSet == NULL) IErrorAbort("No example set specified for net \"%s\".", currentNet->name) ;
  list = (SMlist *)getExampleSetData(exampleSet) ;
  if (list == NULL) {
    fprintf(dout,"SM list is NULL - using default list.\n") ;
    list = defaultSMlist ;
  }

  for (ex = 0; ex < exampleSet->numExamples; ex++) {
    SMentry	*entry = &(list->entries[ex]) ;
    if ((example_n >= 0 && ex != example_n)		||
	(code           && !strstr(entry->code, code))	||
	(re		&& !regex(reCompiled, entry->graphemes))||
	(word           && strcmp(entry->graphemes, word)))
      continue;

    if (nonword) {
      input_orthography(SYLLABLE, nonword, Graphemes) ;
      MupdateNetActivities(currentNet) ;
      pron = decode_phonology(SYLLABLE, Phonemes, NULL) ;
      fprintf(dout, "NW\t%s\t%s\t", nonword, pron) ;
    } else {
      pron = entry->phonemes ;
      fprintf(dout, "%s\t%s\t%s\t", entry->code, entry->graphemes, pron) ;
    }

    for (inCluster = 0; inCluster < nClusters; inCluster++) {
      for (outCluster = 0; outCluster < nClusters+1; outCluster++) {
	/* extra output cluster is for two vowel criteria */

	if (nonword) input_orthography(SYLLABLE, nonword, Graphemes) ;
	else      MgetNumber(exampleSet, ex) ;
	MupdateNetActivities(currentNet) ;
	corr = clusterCorrect(Phonemes, pron, outCluster) ;
	if (!corr) { fprintf(dout, "%.4f ", 1.0) ; continue ; }

	if (nonword) input_orthography(SYLLABLE, nonword, Graphemes) ;
	else      MgetNumber(exampleSet, ex) ;
	distortGraphemes(Graphemes, inCluster, 0.0) ;
	MupdateNetActivities(currentNet) ;
	corr = clusterCorrect(Phonemes, pron, outCluster);
	if (corr) { fprintf(dout, "%.4f ", 0.0) ; continue ; }

	dLo = 0.0 ;  dHi = 1.0 ; d = 0.5 ;  
	do {
	  if (nonword) input_orthography(SYLLABLE, nonword, Graphemes) ;
	  else      MgetNumber(exampleSet, ex) ;
	  distortGraphemes(Graphemes, inCluster, d) ;
	  MupdateNetActivities(currentNet) ;
	  corr = clusterCorrect(Phonemes, pron, outCluster) ;
	  if (corr) dHi = d ;
	  else      dLo = d ;
	  d = (dHi+dLo)/2.0 ;
	} while (dHi-dLo > dTol) ;
	fprintf(dout, "%.4f ", d) ;
      } /* outCluster */
    } /* inCluster */
    fprintf(dout, "\n") ;
    fflush(dout) ;
    if (word || nonword) break ;
  } /* example */
  return 1 ;
}
/*****************************************************************************
 * states
 *****************************************************************************/
#define ENTRIES_INC	128

static int	partFromName(name)
  String name ;
{
  if      (!strcmp(name,"ONSET")) return ONSET ;
  else if (!strcmp(name,"VOWEL")) return VOWEL ;
  else if (!strcmp(name,"CODA"))  return CODA  ;
  else if (!strcmp(name,"BODY"))  return BODY  ;
  else if (!strcmp(name,"ONCLEUS"))  return ONCLEUS  ;
  else
    IErrorAbort("Unknown part name \"%s\"", name) ;
  return 0 ;
}

static String	partOfItem(item, part)
  String item ;
  int	 part ;
{
  char	*ch, string[WLEN] ;

  string[0] = NULL ;
  switch(part) {
  case ONSET:
    strncat(string, item, onset_end(item)) ;
    break ;
  case VOWEL:
    strcat(string, "_") ;
    item += onset_end(item) ;
    strncat(string, item, vowel_end(item)) ;
    item += vowel_end(item) ;
    if (ch = strchr(item, 'e')) strcat(string, "_e") ;
    break ;
  case CODA:
    strcat(string, "_") ;
    item += onset_end(item) ;
    item += vowel_end(item) ;
    strcat(string, item) ;
    if (ch = strchr(string, 'e')) ch[0] = NULL ;
    break ;
  case BODY:
    strcat(string, "_") ;
    item += onset_end(item) ;
    strcat(string, item) ;
    break ;
  case ONCLEUS:
    strncat(string, item, onset_end(item)) ;
    item += onset_end(item) ;
    strncat(string, item, vowel_end(item)) ;
    item += vowel_end(item) ;
    strcat(string, "_") ;
    if (ch = strchr(item, 'e')) strcat(string, ch) ;
    break ;
  }
  return string ;
}

static String	itemWithoutPart(item, part)
  String item ;
  int	 part ;
{
  char	*ch, string[WLEN] ;

  string[0] = NULL ;
  switch(part) {
  case ONSET:
    strcat(string, "_") ;
    item += onset_end(item) ;
    strcat(string, item) ;
    break ;
  case VOWEL:
    strncat(string, item, onset_end(item)) ;
    item += onset_end(item) ;
    strcat(string, "_") ;
    item += vowel_end(item) ;
    strcat(string, item) ;
    if (ch = strchr(string, 'e')) ch[0] = NULL ;
    break ;
  case CODA:
    strncat(string, item, onset_end(item)) ;
    item += onset_end(item) ;
    strncat(string, item, vowel_end(item)) ;
    item += vowel_end(item) ;
    strcat(string, "_") ;
    if (ch = strchr(item, 'e')) strcat(string, ch) ;
    break ;
  case BODY:
    strncat(string, item, onset_end(item)) ;
    break ;
  case ONCLEUS:
    strcat(string, "_") ;
    item += onset_end(item) ;
    item += vowel_end(item) ;
    strcat(string, item) ;
    if (ch = strchr(string, 'e')) ch[0] = NULL ;
    break ;
  }
  return string ;
}

int command_states (tokc, tokv)
  int	tokc ;
  String	tokv[] ;
{
  int		i, e, d, numUnits, numEntries = 0, maxEntries = 0 ;
  Boolean	found ;
  Real*		*state=NULL; 
  Real		sse, netErr, sd=0.1, *itemOutput=NULL, *partOutput=NULL ;
  ExampleSet	exampleSet, consTrainingSet ;
  SMlist	*list ;
  Group		input_group, output_group, hidden_group;
  String	input_group_name=NULL, output_group_name=NULL ;
  String	hidden_group_name=NULL, part_name=NULL ;
  String	type=NULL, code=NULL, word=NULL, re=NULL, reCompiled ;
  String	nonword=NULL, resp=NULL, orth=NULL, phon=NULL, item ;
  String	*distrProns=NULL ;
  int		distribution = 0, maxCount, maxIdx ;
  int		*distrCounts=NULL, nDistrProns, maxDistrProns ;
  Real		distrCriterion = 0.05 ;
  char		string[WLEN] ;
  int		orth_rep, phon_rep, itemPart ;
  int		states = FALSE, contributions = FALSE, correlations = FALSE ;
  int		importance = FALSE, diff = FALSE, error=FALSE ;
  Real		consistency = 0.0 ;
  int		cex, consTrainingType = TRAINING;
  Real		similarity ;
  String	consType = FALSE ;
  int		ex, example_n = -1, exampleType = TRAINING;
  String 	preGroupRE = NULL, postGroupRE = NULL ;
  struct ARG_DESC *argd ;

  if (currentNet == NULL) IErrorAbort("There is no current net.") ;
  if (groupFromName(currentNet, "Letters") != NULL)   orth_rep = LETTERS ;
  if (groupFromName(currentNet, "Attention") != NULL) phon_rep = ATTENTION ;
  _offset	= 0 ;
  _impStart = 0 ; _impNum = 0 ;
  argd = StartArgs(tokv[0]);
  Args(argd, "[-Ptype <type>%s%?]", 	    &type, 		"type of example set (one of TRAINING TESTING VALIDATION)");
  Args(argd, "[-PinputGroup <name>%s%?]",   &input_group_name,	"name of input group");
  Args(argd, "[-PoutputGroup <name>%s%?]",  &output_group_name, "name of output group");
  Args(argd, "[-PhiddenGroup <name>%s%?]",  &hidden_group_name, "name of hidden group (for states/correlations analysis)");
  Args(argd, "[-Pexample <n>%i%?]", 	    &example_n, 	"example to run");
  Args(argd, "[-Pcode <str>%s%?]", 	    &code, 		"only run examples with code <string>") ;
  Args(argd, "[-Pword <str>%s%?]", 	    &word, 		"only run word <string>") ;
  Args(argd, "[-Pre <regexp>%s%?]", 	    &re, 		"only run words matching <regexp>") ;
  Args(argd, "[-Pnonword <string>%s%?]",    &nonword, 		"use nonword <string> as input") ;
  Args(argd, "[-Pstates %P%?]", 	    &states, 		"print states of hidden group");
  Args(argd, "[-Pimportance %P%?]", 	    &importance, 	"successively lesion hidden units and print SSE");
  Args(argd, "[-Pconsistency <real>%r%?]",  &consistency, 	"compute error based on frequency-consistency equation (epsilon <real>)") ;
  Args(argd, "[-PtrainingType <type>%s%?]", &consType, 		"training set for frequency-consistency equation (one of TRAINING TESTING VALIDATION)");
  Args(argd, "[-PimpStart <n>%i%?]",	    &_impStart,		"first output unit number of subset over which to compute importance") ;
  Args(argd, "[-PimpNum <n>%i%?]",	    &_impNum,		"number of output units over which to compute importance") ;
  Args(argd, "[-Pcontributions %P%?]", 	    &contributions,	"print contributions (y*dEdy) of hidden group");
  Args(argd, "[-PthreshPrint <real>%r%?]",  &_print_threshold,	"threshold for printing unit contributions/importances");
  Args(argd, "[-Pcorrelations %P%?]", 	    &correlations,	"print correlations of states of hidden group");
  Args(argd, "[-Perror %P%?]",	 	    &error,		"print error of output group");
  /* diff: (word - wordWithoutPart) (cor) (part - null) */
  Args(argd, "[-Pdiff <part>%s%?]",  	    &part_name, 	"compute hidden activity differences for word and <part> [ONSET VOWEL CODA BODY ONCLEUS]");
  Args(argd, "[-Porthography <str>%s%?]",   &orth,		"orthography coding scheme [syllable letters none]") ;
  Args(argd, "[-Pphonology <str>%s%?]",	    &phon,		"phonology coding scheme [(n)syllable (n)sequential (n)attention]") ;
  Args(argd, "[-Poffset <n>%i%?]",	    &_offset,		"phonology timeSlice offset") ;
  Args(argd, "[-Pdistribution <n>%i%?]",    &distribution, 	"generate a distribution of responses (sample size <n>, noise <sd>)");
  Args(argd, "[-Psd <real>%r%?]",	    &sd,		"standard deviation for noise");
  EndArgs(argd);
  if (GiveHelp(tokc)) {
    ISynopsis("print states/correlations of units") ;
    IHelp
      (tokc, tokv[0], NULL, synopsis,
       NULL);
    return 0;
  }
  ParseArgs(argd, tokc, tokv, 0);
  if (!correlations && !error && !importance) states = TRUE ;
  orth_rep = representation(orth) ;
  phon_rep = representation(phon) ;
  if (!input_group_name)  input_group_name  = default_orthography_group_name(orth_rep) ;
  if (!output_group_name) output_group_name = default_phonology_group_name(phon_rep) ;
  if (!hidden_group_name) hidden_group_name = "Hidden" ;
  output_group = groupFromName(currentNet, output_group_name) ;
  if (output_group == NULL) IErrorAbort("states: Group \"%s\" does not exists.", output_group_name) ;
  hidden_group = groupFromName(currentNet, hidden_group_name) ;
  if (hidden_group == NULL) IErrorAbort("states: Group \"%s\" does not exists.", hidden_group_name) ;
  numUnits = hidden_group->numUnits ;
  if (part_name != NULL) {
    itemPart = partFromName(part_name) ;
    itemOutput = ((Real *)calloc(numUnits, sizeof(Real))) ;
    partOutput = ((Real *)calloc(numUnits, sizeof(Real))) ;
    diff = TRUE ;
  }
  if (distribution) {
    maxDistrProns = WLEN ;
    distrProns = ((String *)calloc(maxDistrProns, sizeof(String))) ;
    distrCounts = ((int *)calloc(maxDistrProns, sizeof(int))) ;
    for (i=0 ; i<maxDistrProns ; i++) 
      distrProns[i] = ((String)calloc(WLEN, sizeof(char))) ;
  }
  if (nonword || diff) {
    input_group = groupFromName(currentNet, input_group_name) ;
    if (input_group == NULL) IErrorAbort("states: Group \"%s\" does not exists.", input_group_name) ;
  }
  if (re) reCompiled = REcompile(re) ;
  if (type) exampleType = findMask(type) ;
  exampleSet = netGetExampleSet(currentNet, exampleType) ;
  if (exampleSet == NULL) IErrorAbort("No example set specified for net \"%s\".", currentNet->name) ;
  list = (SMlist *)getExampleSetData(exampleSet) ;
  if (list == NULL) {
    fprintf(dout,"SM list is NULL - using default list.\n") ;
    list = defaultSMlist ;
  }

  Mreset(exampleSet) ;
  for (ex = 0; ex < exampleSet->numExamples; ex++) {
    SMentry	*entry = &(list->entries[ex]) ;

    if ((example_n >= 0 && ex != example_n)			||
	(code           && !strstr(entry->code, code))		||
	(re		&& !regex(reCompiled, entry->graphemes))||
	(word           && strcmp(entry->graphemes, word)))
      continue;

    if (nonword)
      fprintf(dout, "%s\t", nonword) ;
    else 
      fprintf(dout, "%s\t%s\t/%s/\t", entry->code, entry->graphemes, entry->phonemes) ;

    if (diff) {
      item = (nonword ? nonword : entry->graphemes) ;
      input_orthography(orth_rep, itemWithoutPart(item, itemPart), input_group) ;
      MupdateNetActivities(currentNet) ;
      saveOutputs(hidden_group, itemOutput) ;
    }

    if (nonword) input_orthography(orth_rep, nonword, input_group) ;
    else      MgetNumber(exampleSet, ex) ;
    currentNet->error = 0.0 ;
    MupdateNetActivities(currentNet) ;
    if (contributions) {
      MupdateNetGradients(currentNet) ;
      groupForAllUnits(hidden_group, setStateToContribution, NULL) ;
    }
    /* for recurrent networks, need net->currentTime to contain last slice */

    if (diff) {
      groupSubtractOldStates(hidden_group, itemOutput) ;
      saveOutputs(hidden_group, itemOutput) ;	/* item - itemWithoutPart */
      if (correlations) {
	string[0] = BLANK ; string[1] = NULL ;
	input_orthography(orth_rep, string, input_group) ;
	MupdateNetActivities(currentNet) ;
	saveOutputs(hidden_group, partOutput) ;
	input_orthography(orth_rep, partOfItem(item, itemPart), input_group) ;
	MupdateNetActivities(currentNet) ;
	groupSubtractOldStates(hidden_group, partOutput) ;
	saveOutputs(hidden_group, partOutput) ;
	fprintf(dout, "%f", correlation(itemOutput, partOutput, numUnits)) ;
      }
    }

    if (nonword) sse = 0.0 ;
    else	 sse = groupSumSquaredError(output_group) ;

    if (error)
      fprintf(dout, "%f\t%f\t", currentNet->error, sse) ;

    if (correlations && !diff) {
      if (numEntries >= maxEntries) {
	extendList((void *)&state, sizeof(Real*), maxEntries, maxEntries+ENTRIES_INC) ;
	for (i = maxEntries; i < maxEntries+ENTRIES_INC; i++)
	  state[i] = ((Real *)calloc(numUnits, sizeof(Real))) ;
	maxEntries += ENTRIES_INC ;
      }
      saveOutputs(hidden_group, state[numEntries++]) ;
    }
    if (importance) {
      if (_impNum > 0) {	/* use SSE */
	netForAllUnits(currentNet, OUTPUT, saveError, NULL) ;
	groupForAllUnits(hidden_group, printSSEIncreaseAfterLesion, (void *)&sse) ;
      } else {			/* use whatever network uses */
	netErr = currentNet->error ;
	groupForAllUnits(hidden_group, printErrorIncreaseAfterLesion, (void *)&netErr) ;
      }
    }

    if (consistency != 0.0) {
      netForAllUnits(currentNet, INPUT,  copyOutputToTarget, NULL) ;
      netForAllUnits(currentNet, OUTPUT, copyTargetToOutput, NULL) ;
      netForAllUnits(currentNet, OUTPUT, zeroTotalInput,     NULL) ;
      if (consType) consTrainingType = findMask(consType) ;
      consTrainingSet = netGetExampleSet(currentNet, consTrainingType) ;
      if (consTrainingSet == NULL) 
	IErrorAbort("No training set specified for net \"%s\".", currentNet->name) ;
      Mreset(consTrainingSet) ;
      for (cex = 0; cex < consTrainingSet->numExamples; cex++) {
	MgetNumber(consTrainingSet, cex) ;
	similarity = 0.0 ;
	netForAllUnits(currentNet, INPUT, incrementSimilarity, (void *)&similarity) ;
	similarity *= consTrainingSet->example[cex].frequency ;
	netForAllUnits(currentNet, OUTPUT, incrementTotalInput, (void *)&similarity) ;
      }
      netForAllUnits(currentNet, OUTPUT, copyOutputToTarget, NULL) ;
      netForAllUnits(currentNet, OUTPUT, computeOutputFromTotalInput, (void *)&consistency) ;
      fprintf(dout, "%f\t", groupError(output_group)) ;
    }

    if (distribution) {
      groupForAllUnits(output_group, invertOutput, NULL) ;
      for (d=0, nDistrProns=0 ; d<distribution ; d++) {
	groupForAllUnits(output_group, corruptOutput, (void *)&sd) ;
	resp = decode_phonology(phon_rep, output_group, NULL) ;	
	for (i=0, found=FALSE ; i<nDistrProns ; i++)
	  if (!strcmp(distrProns[i], resp)) { distrCounts[i]++ ; found = TRUE ;}
	if (!found) {
	  if (nDistrProns == maxDistrProns) {
	    maxDistrProns *= 2 ;
	    extendList((void *)&distrProns, sizeof(String *), nDistrProns, maxDistrProns) ;
	    extendList((void *)&distrCounts, sizeof(int *), nDistrProns, maxDistrProns) ;
	    for (i=nDistrProns ; i<maxDistrProns ; i++) 
	      distrProns[i] = ((String)calloc(WLEN, sizeof(char))) ;
	  } /* extend storage */
	  sprintf(distrProns[nDistrProns], "%s", resp) ;
	  distrCounts[nDistrProns++] = 1 ;
	} /* not found */
      }
      do {		/* print in order of decending frequency */
	for (i=0, maxCount=0 ; i<nDistrProns ; i++)
	  if (distrCounts[i] > maxCount) { maxCount = distrCounts[i] ; maxIdx = i ; }
	distrCounts[maxIdx] = 0 ;
	if ((Real)maxCount/(Real)distribution >= distrCriterion)
	  fprintf(dout, "%s\t%.3f\t", distrProns[maxIdx], (Real)maxCount/(Real)distribution) ;
	else maxCount = 0 ;
      } while (maxCount) ;
      fprintf(dout, "END") ;
    } else if (states) 
      groupForAllUnits(hidden_group, printUnitState, NULL) ;
    fprintf(dout, "\n") ;
    fflush(dout);
    if (word || nonword) break ;
  }
  if (correlations && !diff) {
    for (ex = 0 ; ex < numEntries ; ex++) {
      for (e = 0 ; e < numEntries ; e++) 
	fprintf(dout, "%f ", correlation(state[ex], state[e], numUnits)) ;
      fprintf(dout, "\n") ;
    }
    for (ex = 0 ; ex < maxEntries ; ex++) free(state[ex]) ;
    free(state) ;
  }
  if (diff) {
    free(itemOutput) ;
    free(partOutput) ;
  }
  return 1 ;
}
/*****************************************************************************/
