/**********************************************************************
 * Copyright 1990,1991 by University of Toronto, Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee, 
 * provided that the above copyright notice appears in all copies and that 
 * both the copyright notice and this permission notice appear in 
 * supporting documentation, and that the name of University of Toronto 
 * not be used in advertising or publicity pertaining to distribution 
 * of the software without specific, written prior permission.  
 * University of Toronto makes no representations about the suitability 
 * of this software for any purpose. It is provided "as is" without 
 * express or implied warranty. 
 *
 * UNIVERSITY OF TORONTO DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS 
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS, IN NO EVENT SHALL UNIVERSITY OF TORONTO BE LIABLE FOR ANY 
 * SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER 
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF 
 * CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 **********************************************************************/

#include <math.h>
#include <xerion/itf.h>
#include <xerion/minimize.h>
#include <commands.h>

/*
extern Minimize	(*getMinimizeRecord) ARGS((void)) ;
*/

extern Minimize	initMinimize ARGS((Minimize mz, int n)) ;
extern int	getNVars ARGS((Minimize mz)) ;

int command_checkpoint(tokc,tokv)
int tokc;
char *tokv[];
{
  FILE	*fp ;
  char	command[BUFSIZ] ;
  int	max_interval, freq, epoch ;

  IUsage("<filename root> [<max interval>]");
  if (GiveHelp(tokc)) {
    ISynopsis("Checkpoint weights of network into a file");
    IHelp
      (IHelpArgs,
       "Save the weights of the current network into a specified file.",
       "The actual file name will appended with the current epoch number",
       "and the extension \".wts\".  This is best used as a trace.",
       "EXAMPLE",
       "\txerion-> checkpoint encoder",
       "Intervals between checkpointing will be no greater than",
       "<max interval>.  if <max interval> is 0, standard logarithmic",
       "intervals are used.",
       NULL);
    return 0;
  }
  if (tokc < 2 || tokc > 3) IErrorAbort(IPrintUsage(tokv[0], usage)) ;
  if (currentNet == NULL) IErrorAbort("No current network") ;

  epoch = currentNet->mz->nIterations ;

  freq = 1 ;
  while (floor((double)(epoch/freq)) > 10) freq *= 10 ; 
  if (freq < 10) freq = 10 ;
  if (tokc == 3) {
    max_interval = atoi(tokv[2]) ;
    if (max_interval > 0 && freq > max_interval) freq = max_interval ;
  }

  /* read in general parameter changes */
  sprintf(command,"%s.params", tokv[1]) ;
  if (fp=fopen(command, "r")) {	/* just a check to see if it exists */
    ICloseFile(fp,NULL) ;
    sprintf(command,"read -quiet %s.params", tokv[1]) ;
    IDoCommandLine(command) ;
  }

  /* read in epoch-specific parameter changes */
  sprintf(command,"%s.%d.params", tokv[1], epoch) ;
  if (fp=fopen(command, "r")) {	/* just a check to see if it exists */
    ICloseFile(fp,NULL) ;
    sprintf(command,"read %s.%d.params", tokv[1], epoch) ;
    IDoCommandLine(command) ;
  }

  /* possibly checkpoint weights */
  if (epoch%freq==0) {
    fprintf(dout,"Checkpointing weights \"%s.%d.wts\".\n", tokv[1], epoch) ;
    sprintf(command, "saveWeights -nonamed -binary %s.%d.wts", tokv[1], epoch, tokv[1], epoch) ;
    IDoCommandLine(command) ;
  }
  return 1 ;
}
/****************************************************************************
 * PDP weight files (savePDPweights, loadPDPweights)
 ****************************************************************************/
static	void	readWt(link, data)
  Link	link ;
  void	*data ;
{
  FILE	*fp = (FILE *)data ;
  double wt ;

  if (! (link->preUnit->group->type & BIAS)) {
    if (fscanf(fp, "%lf", &wt)==EOF) {
      ICloseFile(fp,NULL) ;
      IErrorAbort("Error reading PDP weight for link %s", link->name) ;
    } else 
      link->weight = (Real)wt ;
  }
}
static	void	writeWt(link, data)
  Link	link ;
  void	*data ;
{
  FILE	*fp = (FILE *)data ;

  if (! (link->preUnit->group->type & BIAS))
    fprintf(fp, "%f\n", link->weight) ;
}

static	void	readWeights(unit, data)
  Unit	unit ;
  void	*data ;
{
  unitForAllLinks(unit, INCOMING, (LinkDataProc)readWt, data) ;
}

static	void	writeWeights(unit, data)
  Unit	unit ;
  void	*data ;
{
  unitForAllLinks(unit, INCOMING, (LinkDataProc)writeWt, data) ;
}

static	void	readBias(unit, data)
  Unit	unit ;
  void	*data ;
{
  FILE	*fp = (FILE *)data ;
  double wt ;

  if (fscanf(fp, "%lf", &wt)==EOF) 
    IErrorAbort("Error reading PDP bias for unit %s", unit->name) ;
  else if (unit->numIncoming &&
	   unit->incomingLink[0]->preUnit->group->type & BIAS) {
    /* units w/o biases (i.e. input units) will have 0.0 in file */
    unit->incomingLink[0]->weight = (Real)wt ;
  }
}

static	void	writeBias(unit, data)
  Unit	unit ;
  void	*data ;
{
  FILE	*fp = (FILE *)data ;
  double wt ;

  if (unit->numIncoming &&
      unit->incomingLink[0]->preUnit->group->type & BIAS)
    fprintf(fp, "%f\n", unit->incomingLink[0]->weight) ;
  else
    fprintf(fp, "0.0\n") ;
}
/****************************************************************************/
int command_loadPDPweights (tokc, tokv)
  int	tokc ;
  char	*tokv[] ;
{
  FILE		*fp ;
  String	filename = NULL ;
  int		numUnits = 0 ;
  struct ARG_DESC *argd ;
  static int	*unitNum, staticNumUnits = 0 ;

  argd = StartArgs(tokv[0]);
  Args(argd, "[<file-name>%s]",	   &filename, 
       "file to read from (defaults to stdin)");
  EndArgs(argd);
  if (GiveHelp(tokc)) {
    ISynopsis("read weights from PDP weight file (or stdin).") ;
    IHelp
      (tokc, tokv[0], NULL, synopsis,
       NULL);
    return 0;
  }
  (void) ParseArgs(argd, tokc, tokv, 0);
  if (!currentNet) IErrorAbort("No current network") ;

  if (filename) fp = IOpenFileOrAbort(filename, "r", NULL) ;
  else          fp = din ;
  
  netForAllOtherUnits(currentNet, BIAS, (UnitDataProc)readWeights, (void *)fp) ;
  netForAllOtherUnits(currentNet, BIAS, (UnitDataProc)readBias,    (void *)fp) ;
  markToRebuildDisplay(CONNECTION_DISPLAY) ;

  if (filename) ICloseFile(fp,NULL) ;
  return 1 ;
}
/****************************************************************************/
int command_savePDPweights (tokc, tokv)
  int	tokc ;
  char	*tokv[] ;
{
  FILE		*fp ;
  String	filename = NULL ;
  int		numUnits = 0 ;
  struct ARG_DESC *argd ;
  static int	*unitNum, staticNumUnits = 0 ;

  argd = StartArgs(tokv[0]);
  Args(argd, "[<file-name>%s]",	   &filename, 
       "file to write to (defaults to stdout)");
  EndArgs(argd);
  if (GiveHelp(tokc)) {
    ISynopsis("write weights to PDP weight file (or stdout).") ;
    IHelp
      (tokc, tokv[0], NULL, synopsis,
       NULL);
    return 0;
  }
  (void) ParseArgs(argd, tokc, tokv, 0);
  if (!currentNet) IErrorAbort("No current network") ;

  if (filename) fp = IOpenFileOrAbort(filename, "w", NULL) ;
  else          fp = dout ;
  
  netForAllOtherUnits(currentNet, BIAS, (UnitDataProc)writeWeights, (void *)fp) ;
  netForAllOtherUnits(currentNet, BIAS, (UnitDataProc)writeBias,    (void *)fp) ;

  if (filename) ICloseFile(fp,NULL) ;
  return 1 ;
}
/****************************************************************************/
static	void	readState(unit, data)
  Unit  unit ;
  void	*data ;
{
  FILE	*fp = (FILE *)data ;
  double st ;

  if (fscanf(fp, "%lf", &st)==EOF) {
    ICloseFile(fp,NULL) ;
    IErrorAbort("Error reading state for unit %s", unit->name) ;
  } else 
    unit->output = (Real)st ;
}
static	void	readStates(group, data)
  Group group ;
  void	*data ;
{
  groupForAllUnits(group, (UnitDataProc)readState, data) ;
}
static	void	writeState(unit, data)
  Unit	unit ;
  void	*data ;
{
  FILE	*fp = (FILE *)data ;
  fprintf(fp, "%f ", unit->output) ;
}
static	void	writeStates(group, data)
  Group group ;
  void	*data ;
{
  FILE	*fp = (FILE *)data ;
  groupForAllUnits(group, (UnitDataProc)writeState, data) ;
  fprintf(fp, "\n") ;
}
/****************************************************************************/
int	command_saveStates (tokc, tokv)
  int	tokc ;
  char	*tokv[] ;
{
  FILE		*fp ;
  String	filename = NULL ;
  String groupRE = NULL ;
  Group	*groupArray, *group ;
  struct ARG_DESC *argd ;

  if (currentNet == NULL) IErrorAbort("There is no current net.") ;

  argd = StartArgs(tokv[0]);
  Args(argd, "[-Pgroup <regexp>%s%?]",	&groupRE,
       "regular expression specifying destination groups.");
  Args(argd, "[<file-name>%s]",	   &filename, 
       "file to read from (defaults to stdin)");
  EndArgs(argd);
  if (GiveHelp(tokc)) {
    ISynopsis("save states of some units") ;
    IHelp
      (tokc, tokv[0], NULL, synopsis,
      NULL);
    return 0;
  }
  ParseArgs(argd, tokc, tokv, 0);
  if (filename) fp = IOpenFileOrAbort(filename, "w", NULL) ;
  else          fp = dout ;

  if (!groupRE) groupRE = ".*" ;
  groupArray = groupsFromRegex(currentNet, groupRE) ;
  if (groupArray == NULL) IErrorAbort("Bad string: \"%s\"", groupRE) ;
  if (groupArray[0] == NULL) IErrorAbort("No match: \"%s\"", groupRE) ;
  for (group = groupArray ; *group != NULL ; ++group) {
    fprintf(fp, "%s\n", (*group)->name) ;
    writeStates(*group, (void *)fp) ;
  }

  if (filename) ICloseFile(fp,NULL) ;
  free(groupArray) ;
  return 1 ;
}
/**********************************************************************/
int command_loadStates (tokc, tokv)
  int	tokc ;
  char	*tokv[] ;
{
  FILE		*fp ;
  String	filename = NULL ;
  char		groupname[128] ;
  Group		group ;
  struct ARG_DESC *argd ;

  if (currentNet == NULL) IErrorAbort("There is no current net.") ;

  argd = StartArgs(tokv[0]);
  Args(argd, "[<file-name>%s]",	   &filename, 
       "file to read from (defaults to stdin)");
  EndArgs(argd);
  if (GiveHelp(tokc)) {
    ISynopsis("load states of units") ;
    IHelp
      (tokc, tokv[0], NULL, synopsis,
      NULL);
    return 0;
  }
  ParseArgs(argd, tokc, tokv, 0);

  if (filename) fp = IOpenFileOrAbort(filename, "r", NULL) ;
  else          fp = din ;
  
  while (fscanf(fp, "%s", groupname) != EOF) {
    if ((group=groupFromName(currentNet, groupname)) == NULL)
      IErrorAbort("unknown group \"%s\".", groupname) ;
    readStates(group, (void *)fp) ;
  }
  markToRebuildDisplay(ACTIVITY_DISPLAY) ;

  if (filename) ICloseFile(fp,NULL) ;
  return 1 ;
}
/****************************************************************************/

static double dotProduct (n, x, y)
  int	n ;
  Real	*x ;
  Real	*y ;
{
  double result = 0.0;
  int i;
  if (x==y)
    for (i=0; i<n; i++)
      result += x[i] * x[i];
  else
    for (i=0; i<n; i++)
      result += x[i] * y[i];
  return result;
}

int command_saveMinimize(tokc,tokv)
int tokc;
char *tokv[];
{
  FILE		*fp ;
  int		nvars, status ;
  intFunc	getNVars ;
  char		command[BUFSIZ], filename[BUFSIZ], *fileroot = NULL ;
  Minimize 	mz ;
  struct ARG_DESC *argd ;
  argd = StartArgs(tokv[0]);
  Args(argd, "<file-root>%s",	   &fileroot, 
       "root of files to save to (appended with .mzp and .mzv)") ;
  EndArgs(argd);
  if (GiveHelp(tokc)) {
    ISynopsis("save minimize record to files");
    IHelp
      (tokc, tokv[0], NULL, synopsis,
       NULL);
    return 0;
  }
  (void) ParseArgs(argd, tokc, tokv, 0);

  if (!currentNet) IErrorAbort("No current network") ;

  mz = currentNet->mz ;
  getNVars = mz->getNVars;
  nvars = getNVars(mz) ;
  
  sprintf(command, "sh rm -f %s.mzp %s.mzv", fileroot, fileroot) ;
  IDoCommandLine(command) ;

  sprintf(filename, "%s.mzv", fileroot) ;

  if (!(fp=fopen(filename, "w")))
    IErrorAbort("unable to open \"%s\" for writing") ;
  status = write(fileno(fp), &nvars,     sizeof(nvars)) ;
  if (status == EOF || status == 0) IErrorAbort("Error writing nvars to %s", filename) ;
  status = write(fileno(fp), mz->x,      nvars*sizeof(*mz->x)) ;
  if (status == EOF || status == 0) IErrorAbort("Error writing mz->x to %s", filename) ;
  status = write(fileno(fp), mz->search, nvars*sizeof(*mz->search)) ;
  if (status == EOF || status == 0) IErrorAbort("Error writing mz->search to %s", filename) ;
  status = write(fileno(fp), mz->rstd,   nvars*sizeof(*mz->rstd)) ;
  if (status == EOF || status == 0) IErrorAbort("Error writing mz->rstd to %s", filename) ;
  status = write(fileno(fp), mz->rsty,   nvars*sizeof(*mz->rsty)) ;
  if (status == EOF || status == 0) IErrorAbort("Error writing mz->rsty to %s", filename) ;
  fclose(fp) ;

  sprintf(command, "show -set currentNet.mz > %s.mzp", fileroot) ;
  IDoCommandLine(command) ;

  return 1 ;
}

int command_restoreMinimize(tokc,tokv)
int tokc;
char *tokv[];
{
  FILE		*fp ;
  int		status, nvars ;
  intFunc	getNVars ;
  char		command[BUFSIZ], filename[BUFSIZ], *fileroot = NULL ;
  Minimize 	mz = NULL;
  struct ARG_DESC *argd ;
  argd = StartArgs(tokv[0]);
  Args(argd, "<file-name>%s",	   &fileroot, 
       "file to read from (defaults to stdin)");
  EndArgs(argd);
  if (GiveHelp(tokc)) {
    ISynopsis("restore minimize record from file");
    IHelp
      (tokc, tokv[0], NULL, synopsis,
       NULL);
    return 0;
  }
  (void) ParseArgs(argd, tokc, tokv, 0);
  if (!currentNet) IErrorAbort("No current network") ;
  mz = currentNet->mz ;
  getNVars = mz->getNVars;
  checkParameters(mz);
  mz = initMinimize(mz, getNVars(mz)) ;
  sprintf(command, "read %s.mzp", fileroot) ;
  IDoCommandLine(command) ;
  
  sprintf(filename, "%s.mzv", fileroot) ;
  fp = IOpenFileOrAbort(filename, "r", NULL) ;
  status = read(fileno(fp), &nvars,     sizeof(nvars)) ;
  if (status == EOF || status == 0) IErrorAbort("Error reading nvars from %s", filename) ;
  if (nvars != getNVars(mz)) IErrorAbort("Inconsistent number of variables (%d %d)", nvars, getNVars(mz)) ;
  status = read(fileno(fp), mz->x,      nvars*sizeof(*mz->x)) ;
  if (status == EOF || status == 0) IErrorAbort("Error reading mz->x from %s", filename) ;
  status = read(fileno(fp), mz->search, nvars*sizeof(*mz->search)) ;
  if (status == EOF || status == 0) IErrorAbort("Error reading mz->search from %s", filename) ;
  status = read(fileno(fp), mz->rstd,   nvars*sizeof(*mz->rstd)) ;
  if (status == EOF || status == 0) IErrorAbort("Error reading mz->rstd from %s", filename) ;
  status = read(fileno(fp), mz->rsty,   nvars*sizeof(*mz->rsty)) ;
  if (status == EOF || status == 0) IErrorAbort("Error reading mz->rsty from %s", filename) ;
  ICloseFile(fp,NULL) ;

  mz->xLen = -1.0 ;	/* minimize.c is modified to look for this */
  mz->chksum = (  (int)mz->x ^ (int)mz->bestx ^ (int)mz->gradient
		^ (int)mz->bestGradientF ^ (int)mz->bestGradientD
		^ (int)mz->prevGradient ^ (int)mz->update ^ (int)mz->search
		^ (int)mz->start ^ (int)mz->rsty ^ (int)mz->rstd
		^ (int)mz->lsStep ^ (int)mz->lsFuncValue ^(int)mz->lsSlope
		^ (int)mz->lsOrder ^ (int)mz->lsHaveSlope) ;
  mz->prevDirectionMethod = mz->directionMethod ;
  mz->prevStepMethod = mz->directionMethod;
  mz->lsResultCode = MZSUCCEED ;
  mz->mzResultCode = MZMAXITER ;
  return 1 ;
}
/****************************************************************************/
