#include <useful.h>
#include <simulator.h>
#include <commands.h>

#include "utilsDCP.h"

static	Real	_linkLength_sd ;
static	Real	_linkLength_min_r ;

static	void	linkSetLength(link, data)
  Link	link ;
  void	*data ;
{
  Real x1 = unitXPosition(link->preUnit),
       y1 = unitYPosition(link->preUnit),
       x2 = unitXPosition(link->postUnit),
       y2 = unitYPosition(link->postUnit),

       /* 1.0 is minimal length */
       length = (_linkLength_sd > 0.0 ?
		 GAUSS_DIST(x1,y1,x2,y2,_linkLength_sd) :
		 (_linkLength_min_r == 2.0 ?
		  1.0+DISTANCE(x1,y1,x2,y2) :
		  (_linkLength_min_r > 0.0 ?
		   1.0+MINR_DIST(x1,y1,x2,y2,_linkLength_min_r) :
		   1.0))) ;
  setLinkLength(link, length) ;
  if (length > *(Real *)data) *(Real *)data = length ;
}

static	void	linkNormLength(link, data)
  Link	link ;
  void	*data ;
{
  setLinkLength(link, getLinkLength(link)/(*(Real *)data)) ;
}

static	void	linkInitLength(link, data)
  Link	link ;
  void	*data ;
{
  setLinkLength(link, 1.0) ;
}


static	void	unitSetLengths(unit, data)
  Unit	unit ;
  void	*data ;
{
  /*
   * OUTGOING prevents lengths of bias weights from being set 
   * (this function is not called for the BIAS unit
   */
  unitForAllLinks(unit, OUTGOING, (LinkDataProc)linkSetLength, data) ;
}

static	void	unitNormLengths(unit, data)
  Unit	unit ;
  void	*data ;
{
  unitForAllLinks(unit, INCOMING, (LinkDataProc)linkNormLength, data) ;
}

static	void	unitInitLengths(unit, data)
  Unit	unit ;
  void	*data ;
{
  unitForAllLinks(unit, INCOMING, (LinkDataProc)linkInitLength, NULL) ;
}

static	void	unitInitOutgoingLengths(unit, data)
  Unit	unit ;
  void	*data ;
{
  unitForAllLinks(unit, OUTGOING, (LinkDataProc)linkInitLength, NULL) ;
}


/*****************************************************************************/
/*****************************************************************************/
int	command_setLinkLengths(tokc, tokv)
  int	tokc ;
  char	**tokv ;
{
  Real	maxLength = 0.0 ;
  String preGroupRE = NULL ;
  Group	*preGroupArray,  *preGroup ;
  int	init = 0, norm = 0 ;
  struct ARG_DESC *argd ;

  if (currentNet == NULL)
    IErrorAbort("There is no current net.") ;

  _linkLength_sd    = 0.0 ;
  _linkLength_min_r = 2.0 ;
  argd = StartArgs(tokv[0]);
  Args(argd, "[-Psd <real>%r%?]",		&_linkLength_sd,
       "SD of gaussian distance metric.");
  Args(argd, "[-Pminkowski-r <real>%r%?]",	&_linkLength_min_r,
       "R of Minkowski-r distance metric; 2.0 is euclidean (default).");
  Args(argd, "[-Pinit %P%?]",			&init,
       "initialize length lengths to 1.0.");
  Args(argd, "[-Pnorm %P%?]",			&norm,
       "normalize length lengths to [0,1].");
  Args(argd, "[-PpreGroup <regexp>%s%?]",	&preGroupRE,	"regular expression specifying source groups.");
  EndArgs(argd);
  if (GiveHelp(tokc)) {
    ISynopsis("Sets lengths of links");
    IHelp
      (tokc, tokv[0], NULL, synopsis,
       "Sets lengths of links based on positions in activity display." ,
       "EXAMPLES",
       "SEE ALSO",
       NULL) ;
    return 1 ;
  }
  ParseArgs(argd, tokc, tokv, 0);
  if (preGroupRE != NULL) {
    preGroupArray = groupsFromRegex(currentNet, preGroupRE) ;
    if (preGroupArray == NULL)    IErrorAbort("Bad string: \"%s\"", preGroupRE) ;
    if (preGroupArray[0] == NULL) IErrorAbort("No match: \"%s\"", preGroupRE) ;
    for (preGroup = preGroupArray ; *preGroup != NULL ; ++preGroup) {
      if (init) {
	fprintf(dout,"Initializing %s outgoing link lengths to 1.0\n", (*preGroup)->name) ;
	groupForAllUnits(*preGroup, (UnitDataProc)unitInitOutgoingLengths, NULL) ;
      } else {
	if (_linkLength_sd > 0.0) 
	  fprintf(dout,"Setting %s outgoing link lengths to unnormalized Gaussian distance (sd=%g)\n", 
		  (*preGroup)->name, _linkLength_sd) ;
	else
	  fprintf(dout,"Setting %s outgoing link lengths to unnormalized Minkowski distance (r=%g)\n", 
		  (*preGroup)->name, _linkLength_min_r) ;
	groupForAllUnits(*preGroup, (UnitDataProc)unitSetLengths, (void *)&maxLength) ;
      }    
    }
  } else if (init)
    netForAllOtherUnits(currentNet, BIAS, (UnitDataProc)unitInitLengths, NULL) ;
  else {
    if (_linkLength_sd > 0.0)
      fprintf(dout,"Setting link lengths to %snormalized Gaussian distance (sd=%g)\n", 
	     (norm?"":"un"), _linkLength_sd) ;
    else 
      fprintf(dout,"Setting link lengths to %snormalized Minkowski distance (r=%g)\n", 
	     (norm?"":"un"), _linkLength_min_r) ;
    netForAllOtherUnits(currentNet, BIAS, (UnitDataProc)unitSetLengths, (void *)&maxLength) ;
    if (norm)
      netForAllOtherUnits(currentNet, BIAS, (UnitDataProc)unitNormLengths, (void *)&maxLength) ;
  }
  return 1 ;
}
/*****************************************************************************/
/*****************************************************************************/
int	command_copyPositions(tokc, tokv)
  int	tokc ;
  char	**tokv ;
{
  int		idx ;
  Group		fromGroup, toGroup ;
  String	fromGroup_name, toGroup_name ;
  struct ARG_DESC *argd ;

  if (currentNet == NULL)
    IErrorAbort("There is no current net.") ;

  argd = StartArgs(tokv[0]);
  Args(argd, "[-PfromGroup <name>%s%?]", &fromGroup_name, "name of from group");
  Args(argd, "[-PtoGroup <name>%s%?]",   &toGroup_name,	  "name of to group");
  EndArgs(argd);
  if (GiveHelp(tokc)) {
    ISynopsis("Sets positions of <toGroup> to be same as <fromGroup>");
    IHelp
      (tokc, tokv[0], NULL, synopsis,
       "Sets positions of <toGroup> to be same as <fromGroup>." ,
       "EXAMPLES",
       "SEE ALSO",
       NULL) ;
    return 1 ;
  }
  ParseArgs(argd, tokc, tokv, 0);
  fromGroup = groupFromName(currentNet, fromGroup_name) ;
  if (fromGroup == NULL) IErrorAbort("Group \"%s\" does not exists.", fromGroup_name) ;
  toGroup = groupFromName(currentNet, toGroup_name) ;
  if (toGroup == NULL) IErrorAbort("Group \"%s\" does not exists.", toGroup_name); 
  if (fromGroup->numUnits != toGroup->numUnits)
    IErrorAbort("Groups \"%s\" and \"%s\" do not have the same number of units (%d vs. %d).",
		fromGroup_name, toGroup_name, fromGroup->numUnits, toGroup->numUnits) ;

  for (idx = 0 ; idx < fromGroup->numUnits ; idx++) {
    Unit from = fromGroup->unit[idx], to = toGroup->unit[idx] ;
    MxPosition(to) = MxPosition(from) ;
    MyPosition(to) = MyPosition(from) ;
  }
  return 1 ;
}
/*****************************************************************************/
