#include <useful.h>
#include <simulator.h>
#include <commands.h>
#include <values.h>

#ifndef ibm
#ifndef linux
#include <nan.h>
#endif
#endif

#include "utilsDCP.h"

static	Real	probability ;
static	Real	minWeightMag ;
static	int	numLinks ;
static	int	numLesioned ;
static	Real	xPosition ;
static	Real	yPosition ;
static	int	prePosition ;

#define MoldWeight(link)	((link)->deltaWeight)

/***********************************************************************/
static void	restoreWeightsFromGroup(unit, data)
  Unit		unit ;
  void		*data ;
{
  Group	preGroup = (Group)data ;
  int	idx ;
  int	numIncoming = unit->numIncoming ;
  Link	*incoming   = unit->incomingLink ;

  for (idx = 0 ; idx < numIncoming ; ++idx) {
    Link        link    = incoming[idx] ;
    if (link->preUnit->group == preGroup) 
      link->weight = MoldWeight(link) ;
  }
}
/***********************************************************************/
static void	corruptWeightsFromGroup(unit, data)
  Unit		unit ;
  void		*data ;
{
  Group	preGroup = (Group)data ;
  int	idx ;
  int	numIncoming = unit->numIncoming ;
  Link	*incoming   = unit->incomingLink ;

  for (idx = 0 ; idx < numIncoming ; ++idx) {
    Link        link    = incoming[idx] ;
    if (link->preUnit->group == preGroup) {
      MoldWeight(link) = link->weight ;
      link->weight += gaussianNoise(probability) ;
    }
  }
}
/***********************************************************************/
void	corruptWeights (preGroup, postGroup, sd)
  Group	preGroup ;
  Group	postGroup ;
  double sd ;
{
  probability = sd ;
  fprintf(dout,"Corrupting %s ==> %s weights with N(0,%.2g).\n",
	  preGroup->name, postGroup->name, sd) ;
  groupForAllUnits(postGroup, (UnitDataProc)corruptWeightsFromGroup, preGroup) ;
}
/***********************************************************************/
static void	randomizeWeightsFromGroup(unit, data)
  Unit		unit ;
  void		*data ;
{
  Group	preGroup = (Group)data ;
  int	idx ;
  int	numIncoming = unit->numIncoming ;
  Link	*incoming   = unit->incomingLink ;

  for (idx = 0 ; idx < numIncoming ; ++idx) {
    Link        link    = incoming[idx] ;
    if (link->preUnit->group == preGroup) {
      MoldWeight(link) = link->weight ;
      link->weight = uniformNoise(-probability,probability) ;
    }
  }
}
/***********************************************************************/
void	randomizeWeights (preGroup, postGroup, randRange)
  Group	preGroup ;
  Group	postGroup ;
  double randRange ;
{
  probability = randRange ;
  fprintf(dout,"Randomizing %s ==> %s weights within [%.2g,%.2g].\n",
	  preGroup->name, postGroup->name, -probability, probability) ;
  groupForAllUnits(postGroup, (UnitDataProc)randomizeWeightsFromGroup, preGroup) ;
}
/***********************************************************************/
static void	scaleWeightsFromGroup(unit, data)
  Unit		unit ;
  void		*data ;
{
  Group	preGroup = (Group)data ;
  int	idx ;
  int	numIncoming = unit->numIncoming ;
  Link	*incoming   = unit->incomingLink ;

  for (idx = 0 ; idx < numIncoming ; ++idx) {
    Link        link    = incoming[idx] ;
    if (link->preUnit->group == preGroup) {
      MoldWeight(link) = link->weight ;
      link->weight *= probability ;	/* set to scale by scaleWeights */
    }
  }
}
/***********************************************************************/
void	scaleWeights (preGroup, postGroup, scale)
  Group	preGroup ;
  Group	postGroup ;
  double scale ;
{
  probability = scale ;
  fprintf(dout,"Scaling %s ==> %s weights by %.2g.\n",
	  preGroup->name, postGroup->name, probability) ;
  groupForAllUnits(postGroup, (UnitDataProc)scaleWeightsFromGroup, preGroup) ;
}
/***********************************************************************/
void	restoreWeights (preGroup, postGroup)
  Group	preGroup ;
  Group	postGroup ;
{
  fprintf(dout,"Restoring %s ==> %s weights.\n", 
	  preGroup->name, postGroup->name) ;
  groupForAllUnits(postGroup, (UnitDataProc)restoreWeightsFromGroup, preGroup) ;
}
/***********************************************************************
 *	Name:		lesionLinksFromGroup
 *	Description:	lesion links to unit from group
 *	Parameters:	
 *		Unit	unit - the unit receiving the links
 *		void	*data - the group sending the links
 *	Return Value:	NONE
 ***********************************************************************/
static void	lesionLinksFromGroup(unit, data)
  Unit		unit ;
  void		*data ;
{
  Group	preGroup = (Group)data ;
  int	idx ;
  Real	prob = probability ;
  int	numIncoming = unit->numIncoming ;
  Link	*incoming   = unit->incomingLink ;

  if (xPosition>=0.0 && yPosition>=0.0 && !prePosition)
    prob = GAUSS_2D(MxPosition(unit),MyPosition(unit),xPosition,yPosition,probability) ;

  for (idx = 0 ; idx < numIncoming ; ++idx) {
    Link        link    = incoming[idx] ;
    if (link->preUnit->group == preGroup) {
      numLinks++ ;
      if (isLesionedLink(link)) unlesionLink(link) ;
      if (prePosition && xPosition>=0.0 && yPosition>=0.0) 
	prob = GAUSS_2D(MxPosition(link->preUnit),MyPosition(link->preUnit),xPosition,yPosition,probability) ;
      if (WITH_PROBABILITY(prob) &&
	  fabs(link->weight) < minWeightMag) {
	lesionLink(link) ;
	numLesioned++ ;
      }
    }
  }
}
static void	unitLesion(unit, data)
  Unit		unit ;
  void		*data ;
{
  Real	prob = probability ;

  if (xPosition>=0.0 && yPosition>=0.0)
    prob = GAUSS_2D(MxPosition(unit), MyPosition(unit), xPosition, yPosition, probability) ;
  if (WITH_PROBABILITY(prob)) {
    lesionUnit(unit) ;
    numLesioned++ ;
  } else
    unlesionUnit(unit) ;
}

/***********************************************************************
 *	Name:		lesion
 *	Description:	lesion links between two groups
 *	Parameters:	
 *		Group	preGroup   - the group the links are to come from
 *		Group	postGroup  - the group the links are to go to
 *		Real	prob       - proportion of links to lesion
 *	Return Value:	
 ***********************************************************************/
void	lesion (preGroup, postGroup, prob, minWt, print)
  Group	preGroup ;
  Group	postGroup ;
  double prob, minWt ;
  int 	print ;
{
  probability = prob ;
  minWeightMag = minWt ;
  numLinks    = 0 ;
  numLesioned = 0 ;
  if (preGroup) {	/* LINKS */
    groupForAllUnits(postGroup, (UnitDataProc)lesionLinksFromGroup, preGroup) ;
    if (print && numLinks) {
      if (xPosition>=0.0 && yPosition>=0.0) fprintf(dout,"Gaussian [%g,%g] %s ", xPosition, yPosition,(prePosition?"pre":"")) ;
      if (probability == 0.0) fprintf(dout,"Restored ") ;
      else                    fprintf(dout,"Lesioned %.2g (%d) of ", probability, numLesioned) ;
      fprintf(dout,"%s ==> %s links", preGroup->name, postGroup->name) ;
      if (minWeightMag < MAXFLOAT) 
	fprintf(dout, " with |w|<%.4g", minWeightMag) ;
      fprintf(dout, "\n") ;
    }
  } else {		/* UNITS */
    groupForAllUnits (postGroup, (UnitDataProc)unitLesion, NULL) ;
    if (print) {
      if (xPosition>=0.0 && yPosition>=0.0) fprintf(dout,"Gaussian [%g,%g] ", xPosition, yPosition) ;
      if (probability == 0.0) fprintf(dout,"Restored ") ;
      else                    fprintf(dout,"Lesioned %.2g (%d) of ", probability, numLesioned) ;
      fprintf(dout,"%s units\n", postGroup->name) ;
    }
  }
}

/***********************************************************************
 *	Name:		command_lesion
 *	Description:	(un)lesions units/connections
 *	Parameters:	
 *		int	tokc    - the number of command line tokens
 *		char	*tokv[] - the vector of tokens
 *	Return Value:	
 *		int	0 on failure, 1 on success
 ***********************************************************************/
int	command_lesion (tokc, tokv)
  int	tokc ;
  char	*tokv[] ;
{
  Boolean lesioningLinks = FALSE ;
  int	noise = FALSE ;
  Real	sd = 0.0, randRange = 0.0, scale = 0.0 ;
  String preGroupRE = NULL, postGroupRE = NULL ;
  Group	*preGroupArray,  *preGroup ;
  Group	*postGroupArray, *postGroup ;
  struct ARG_DESC *argd ;

  if (currentNet == NULL)
    IErrorAbort("There is no current net.") ;

  probability = 1.0 ;
  minWeightMag = MAXFLOAT ;
  xPosition = yPosition = -1.0 ;
  prePosition = FALSE ;
  argd = StartArgs(tokv[0]);
  Args(argd, "[-Pprobability <real>%r%?]",	&probability,	"probability of removing each specified link.");
  Args(argd, "[-PminWeightMag <real>%r%?]",	&minWeightMag,  "only lesion links with weight magnitudes below this.");
  Args(argd, "[-Pnoise %P%?]", 	    		&noise, 	"add noise to weights.");
  Args(argd, "[-Prandomize <real>%r%?]",	&randRange,    	"randomize weights within +/-<real>.");
  Args(argd, "[-Psd <real>%r%?]",		&sd,       	"standard deviation of noise added to weights.");
  Args(argd, "[-Pscale <real>%r%?]",		&scale,       	"factor by which to multply weights.");
  Args(argd, "[-Px <real>%r%?]",		&xPosition,    	"x position (column) of position-based gaussian lesion (w/ sd).");
  Args(argd, "[-Py <real>%r%?]",		&yPosition,    	"y position (row) of position-based gaussian lesion (w/ sd).");
  Args(argd, "[-PusePrePosition %P%?]", 	&prePosition, 	"[x y] specifies position of PRE unit (post by default).");
  Args(argd, "[-PpreGroup <regexp>%s%?]",	&preGroupRE,	"regular expression specifying source groups.");
  Args(argd, "[-PpostGroup <regexp>%s%?]",	&postGroupRE,   "regular expression specifying destination groups.");
  EndArgs(argd);
  if (GiveHelp(tokc)) {
    ISynopsis("lesion some units or links between groups") ;
    IHelp
      (tokc, tokv[0], NULL, synopsis,
       "Randomly lesion some proportion P of units in <preGroup>, or links",
       "between <preGroup> and <postGroup>, in the current network.  Group", 
       "names may  use regular expressions similar to those used in grep to ",
       "lesion more than one group or set of links." ,
       "Using P = 1.0 (the default) lesions all specified units/links;" ,
       "P = 0.0 restores all such units/links.", 
       "EXAMPLE",
       "\txerion-> lesion -pre \"input\" -post \"hidden\"",
       "",
       "will lesion all of the links between the group \"input\" and the",
       "group \"hidden\".",
       "",
       "\txerion-> lesion -prob 0.3 -post \"hidden*\"",
       "",
       "will lesion 0.3 of the units in all groups whose name begins with",
       "\"hidden\".",
      NULL);
    return 0;
  }
  ParseArgs(argd, tokc, tokv, 0);

  postGroupArray = groupsFromRegex(currentNet, postGroupRE) ;
  if (postGroupArray == NULL)
    IErrorAbort("Bad string: \"%s\"", postGroupRE) ;
  if (postGroupArray[0] == NULL)
    IErrorAbort("No match: \"%s\"", postGroupRE) ;

  if (preGroupRE != NULL) {
    lesioningLinks = TRUE ;
    preGroupArray = groupsFromRegex(currentNet, preGroupRE) ;
    if (preGroupArray == NULL)
      IErrorAbort("Bad string: \"%s\"", preGroupRE) ;
    if (preGroupArray[0] == NULL)
      IErrorAbort("No match: \"%s\"", preGroupRE) ;
  }
  if (xPosition>=0.0 && yPosition>=0.0) {
    if (sd == 0.0) IErrorAbort("Cannot use position-based lesioning with sd=0.0") ;
    else           probability = sd ;
  }
  for (postGroup = postGroupArray ; *postGroup != NULL ; ++postGroup) {
    if (lesioningLinks) 
      for (preGroup = preGroupArray ; *preGroup != NULL ; ++preGroup) {
	if (noise) {
	  if (sd == 0.0) restoreWeights(*preGroup, *postGroup) ;
	  else           corruptWeights(*preGroup, *postGroup, sd) ;
	} else if (scale != 0.0) {
	  if (scale == 1.0) restoreWeights(*preGroup, *postGroup) ;
	  else		    scaleWeights(*preGroup, *postGroup, scale) ;
	} else if (randRange) {
	  if (randRange == 0.0) restoreWeights(*preGroup, *postGroup) ;
	  else                  randomizeWeights(*preGroup, *postGroup, randRange) ;
	} else
	  lesion(*preGroup, *postGroup, probability, minWeightMag, TRUE) ;
      }
    else
      lesion(NULL, *postGroup, probability, minWeightMag, TRUE) ;
  }
  free(preGroupArray) ;
  free(postGroupArray) ;
  
  return 1 ;
}
/***********************************************************************/
static void	countLesionedLinks(unit, data)
  Unit		unit ;
  void		*data ;
{
  int	idx, numLesioned ;
  int	numIncoming = unit->numIncoming ;
  Link	*incoming   = unit->incomingLink ;

  for (idx = 0, numLesioned = 0 ; idx < numIncoming ; ++idx)
    if (isLesionedLink(incoming[idx])) numLesioned++ ;

  fprintf(dout, "%s %d %d %d of %d\n", unit->name, unit->activityColumn, unit->activityRow, numLesioned, numIncoming) ;
  if (numIncoming > 0) unit->output = (Real)numLesioned/(Real)numIncoming ;
}
/**********************************************************************/
int command_countLesionedLinks(tokc,tokv)
int tokc;
char *tokv[];
{
  if (tokc==SHORT_HELP || tokc==LONG_HELP) {
    help_string = "count number of lesioned incoming connections for each hidden unit" ;
    return 0;
  }

  if (currentNet == NULL)
    IErrorAbort("There is no current net.") ;

  netForAllOtherUnits(currentNet, BIAS|INPUT|OUTPUT, (UnitDataProc)countLesionedLinks, NULL) ;

  return 1;
}

/**********************************************************************/
/**********************************************************************/
/***********************************************************************/
static Real _mags[32] ;
static int  _cnts[32] ;

static void	incInWtMag(link, data)
  Link link ;
  void *data ;
{
  int i ;
  Net net = link->postUnit->net ;

  for (i = 0 ; i < net->numGroups ; i++) 
    if (link->preUnit->group == net->group[i]) {
      _mags[i] += fabs(link->weight) ;
      _cnts[i]++ ;
      return ;
    }
  IErrorAbort("group '%s' not found", link->preUnit->group->name) ;
}

static void	incOutWtMag(link, data)
  Link link ;
  void *data ;
{
  int i ;
  Net net = link->preUnit->net ;

  for (i = 0 ; i < net->numGroups ; i++) 
    if (link->postUnit->group == net->group[i]) {
      _mags[i] += fabs(link->weight) ;
      _cnts[i]++ ;
      return ;
    }
  IErrorAbort("group '%s' not found", link->postUnit->group->name) ;
}

static void	printWtMags(unit, data)
  Unit unit ;
  void *data ;
{
  int i ;

  fprintf(dout, "%s %d %d ", unit->name, MxPosition(unit), MyPosition(unit)) ;

  for (i = 0 ; i < unit->net->numGroups ; i++) {
    _mags[i] = 0.0 ;  _cnts[i] = 0 ;
  }
  unitForAllLinks(unit, INCOMING, (LinkDataProc)incInWtMag, NULL) ;
  for (i = 0 ; i < unit->net->numGroups ; i++) 
    if (_cnts[i] > 0) {
      fprintf(dout, "%s %f ", unit->net->group[i]->name, 
	      _mags[i]/((Real)_cnts[i])) ;
      _mags[i] = 0.0 ;
      _cnts[i] = 0 ;
    }
  unitForAllLinks(unit, OUTGOING, (LinkDataProc)incOutWtMag, NULL) ;
  for (i = 0 ; i < unit->net->numGroups ; i++) 
    if (_cnts[i] > 0)
      fprintf(dout, "%s %f ", unit->net->group[i]->name, 
	      _mags[i]/((Real)_cnts[i])) ;
  fprintf(dout, "\n") ;
}

int command_printWeightMags(tokc,tokv)
int tokc;
char *tokv[];
{
  if (tokc==SHORT_HELP || tokc==LONG_HELP) {
    help_string = "print average weight magnitudes by group for each hidden unit" ;
    return 0;
  }

  if (currentNet == NULL)
    IErrorAbort("There is no current net.") ;

  netForAllOtherUnits(currentNet, BIAS|INPUT|OUTPUT, (UnitDataProc)printWtMags, NULL) ;

  return 1;
}
