/**********************************************************************
 * Copyright 1990,1991 by University of Toronto, Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee, 
 * provided that the above copyright notice appears in all copies and that 
 * both the copyright notice and this permission notice appear in 
 * supporting documentation, and that the name of University of Toronto 
 * not be used in advertising or publicity pertaining to distribution 
 * of the software without specific, written prior permission.  
 * University of Toronto makes no representations about the suitability 
 * of this software for any purpose. It is provided "as is" without 
 * express or implied warranty. 
 *
 * UNIVERSITY OF TORONTO DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS 
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS, IN NO EVENT SHALL UNIVERSITY OF TORONTO BE LIABLE FOR ANY 
 * SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER 
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF 
 * CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 **********************************************************************/

#include <sys/times.h>
#include <math.h>
#include <values.h>
#include <commands.h>
#include <xerion/minimize.h>
#include "utilsDCP.h"

#ifndef ibm
#ifndef linux
#include <nan.h>
#endif
#endif

void	Pause(msg)
  String	msg ;
{
  fprintf(dout, "\n%s [Return to continue] ", msg) ;
  getc(din) ;
}

Real	sq (x)
  Real x ; 
{ 
  return x*x ; 
}

Real	now()
{
  struct tms	tms;
  times(&tms) ;
  return (tms.tms_utime + tms.tms_stime)/100.0 ;
}

/* if x is normally distributed with mean M and standard deviation S
 * then (x - M)/S follows the standard normal distribution (M=0,S=1)
 */
double poz (z)	/* computes probability of normal z score */
double	z;
{
  double	y, x, w;
  if (z == 0.0)
    x = 0.0;
  else
    {
      y = 0.5 * fabs (z);
      if (y >= 3.0)
	x = 1.0;
      else if (y < 1.0)
	{
	  w = y*y;
	  x = ((((((((0.000124818987 * w
		      -0.001075204047) * w +0.005198775019) * w
		    -0.019198292004) * w +0.059054035642) * w
		  -0.151968751364) * w +0.319152932694) * w
		-0.531923007300) * w +0.797884560593) * y * 2.0;
	}
      else
	{
	  y -= 2.0;
	  x = (((((((((((((-0.000045255659 * y
			   +0.000152529290) * y -0.000019538132) * y
			 -0.000676904986) * y +0.001390604284) * y
		       -0.000794620820) * y -0.002034254874) * y
		     +0.006549791214) * y -0.010557625006) * y
		   +0.011630447319) * y -0.009279453341) * y
		 +0.005353579108) * y -0.002141268741) * y
	       +0.000535310849) * y +0.999936657524;
	}
    }
  return (z > 0.0 ? ((x + 1.0) / 2.0) : ((1.0 - x) / 2.0));
}

Real gaussianNoise(sd)
  double sd ;
{
  static int iset=FALSE;
  static Real gset;
  Real v1,v2,r,fac;

  if (sd == 0.0) return 0.0 ;
  if (iset) {
    iset = FALSE;
    return (gset*sd);
  } else {
    do {
      v1 = uniformNoise(-1.0,1.0) ;
      v2 = uniformNoise(-1.0,1.0) ;
      r  = v1*v1 + v2*v2;
    } while (r >= 1.0);
    fac = sqrt(-2.0*log(r)/r);
    gset = v1*fac;
    iset = TRUE;
    return (v2*fac*sd);
  }
}

static int	_unitIdx ;
static int	maxUnits ;
static int	_time ;

static void	countUnit(unit, data)
  Unit		unit ;
  void		*data ;
{
  _unitIdx++ ;
}
static	void	saveOutput(unit, data)
  Unit	unit ;
  void	*data ;
{
  ((Real *)data)[_unitIdx++] = unit->output ;
}
static	void	saveOutputAt(unit, data)
  Unit	unit ;
  void	*data ;
{
  ((Real *)data)[_unitIdx++] = unit->outputHistory[_time] ;
}
static	void	saveTarget(unit, data)
  Unit	unit ;
  void	*data ;
{
  ((Real *)data)[_unitIdx++] = unit->target ;
}
static	void	saveTargetAt(unit, data)
  Unit	unit ;
  void	*data ;
{
  ((Real *)data)[_unitIdx++] = unit->targetHistory[_time] ;
}
static	void	saveTotalInput(unit, data)
  Unit	unit ;
  void	*data ;
{
  ((Real *)data)[_unitIdx++] = unit->totalInput ;
}

static	void	restoreOutput(unit, data)
  Unit	unit ;
  void	*data ;
{
  unit->output = ((Real *)data)[_unitIdx++] ;
}
static	void	restoreTotalInput(unit, data)
  Unit	unit ;
  void	*data ;
{
  unit->totalInput = ((Real *)data)[_unitIdx++] ;
}


void	invertOutput(unit, data)
  Unit	unit ;
  void	*data ;
{
  if (isLesionedUnit(unit)) return ;
  unit->totalInput = sigmoidInv(unit->output) ;
}

void	corruptOutput(unit, data)
  Unit	unit ;
  void	*data ;
{
  Real input = unit->totalInput + gaussianNoise(*(Real *)data) ;
  unit->output = sigmoid(input) ;
}

int	netCountUnits(net) 
  Net	net ;
{
  _unitIdx = 0 ;
  netForAllUnits(net, ALL, (UnitDataProc)countUnit, NULL) ;
  return _unitIdx ;
}
void	netSaveOutputs(net, v)
  Net	net ;
  Real	*v ;
{
  _unitIdx = 0 ;
  netForAllUnits(net, ALL, (UnitDataProc)saveOutput, (void *)v) ;
}
void	netSaveTotalInputs(net, v)
  Net	net ;
  Real	*v ;
{
  _unitIdx = 0 ;
  netForAllUnits(net, ALL, (UnitDataProc)saveTotalInput, (void *)v) ;
}
void	netRestoreOutputs(net, v)
  Net	net ;
  Real	*v ;
{
  _unitIdx = 0 ;
  netForAllUnits(net, ALL, (UnitDataProc)restoreOutput, (void *)v) ;
}
void	netRestoreTotalInputs(net, v)
  Net	net ;
  Real	*v ;
{
  _unitIdx = 0 ;
  netForAllUnits(net, ALL, (UnitDataProc)restoreTotalInput, (void *)v) ;
}

/**********************************************************************/
int		groupIdxFromName (net, name)
  Net		net ;
  String	name ;
{
  int	groupIdx ;
  Group	*group ;

  for (groupIdx=0, group=net->group;
       groupIdx < net->numGroups; 
       groupIdx++, group++)
    if (!strcmp((*group)->name, name)) return groupIdx ;
  return -1 ;
}
/****************************************************************************/

void		saveOutputs(group, v)
  Group	group ;
  Real	*v;
{
  _unitIdx = 0 ;
  groupForAllUnits(group, (UnitDataProc)saveOutput, (void *)v) ;
}

void		saveTargets(group, v)
  Group	group ;
  Real	*v;
{
  _unitIdx = 0 ;
  groupForAllUnits(group, (UnitDataProc)saveTarget, (void *)v) ;
}

void		saveOutputsAt(t, group, v)
  int	t ;
  Group	group ;
  Real	*v;
{
  _unitIdx = 0 ;
  _time    = t ;
  groupForAllUnits(group, (UnitDataProc)saveOutputAt, (void *)v) ;
}

void		saveTargetsAt(t, group, v)
  int	t ;
  Group	group ;
  Real	*v;
{
  _unitIdx = 0 ;
  _time    = t ;
  groupForAllUnits(group, (UnitDataProc)saveTargetAt, (void *)v) ;
}

void		restoreOutputs(group, v)
  Group	group ;
  Real	*v;
{
  _unitIdx = 0 ;
  groupForAllUnits(group, (UnitDataProc)restoreOutput, (void *)v) ;
}

int	binarizedDifference(v1,v2,n) 
     Real *v1, *v2 ;
     int n ;
{
  int i, diff = 0 ;
  for (i = 0 ; i < n ; i++) 
    if ((v1[i]>0.5 && v2[i]<0.5) ||
	(v1[i]<0.5 && v2[i]>0.5)) diff++ ;
  return diff ;
}


Real		proximity(v1,v2,n)
  Real	*v1, *v2;
  int n;
{
  int i;
  /* normalized dot product, cosine */
  Real l12 = 0.0, l1 = 0.0, l2 = 0.0;
  for (i = 0 ; i < n; i++) {
    l1  += v1[i]*v1[i];    l2  += v2[i]*v2[i];    l12 += v1[i]*v2[i];
  }
  if (l1==0.0 || l2==0.0) return 0.0;
  return (l12/sqrt(l1*l2));
}

Real		correlation(v1,v2,lv)
  Real *v1,*v2;
  int lv;
{
  int i;
  Real d1,d2,flv=((Real) lv);
  Real mean1=0.0, mean2=0.0, var1=0.0, var2=0.0, cov=0.0;
  for (i = 0; i < lv; i++) {
    mean1 += v1[i];   mean2 += v2[i];
  }
  mean1 /= flv;  mean2 /= flv;   
  for (i = 0; i < lv; i++) {
    d1 = (v1[i]-mean1);
    d2 = (v2[i]-mean2);
    var1 += d1*d1;
    var2 += d2*d2;
    cov  += d1*d2;
  }
  if (var1==0.0 || var2 == 0.0) return 0.0;
  return ((cov/flv)/(sqrt(var1/flv)*sqrt(var2/flv)));
}

Real		divergence(act,targ,n)
  Real	*act, *targ;
  int n;
{
  /* t*log(t/a) */
  int i;
  Real	div = 0.0, sum = 0.0 ;
  for (i = 0 ; i < n ; i++) sum += act[i] ;
  if (sum > 0.0) {
    sum = 1.0/sum ;
    for (i = 0 ; i < n ; i++) {
      Real t = targ[i] ;
      if (t > 0.0) 
	div += t*log((t+MIN_UNIT_OUTPUT)/(act[i]*sum+MIN_UNIT_OUTPUT)) ;
    }
    return div ;
  } else 
    return 0.0 ;
}

Real		gpe(act,targ,n)
  Real	*act, *targ;
  int n;
{
  /* 
   * Grammatical Prediction Error (Christiansen) 
   * 1 - hits / ((hits + FAs)(1 + misses))
   */
  int i;
  Real	hits = 0.0, FAs = 0.0, misses = 0.0 ;
  for (i = 0 ; i < n ; i++) {
    Real t = targ[i], a = act[i] ;
    if (t > 0) {
      hits += a ;
      if (t > a) misses += t-a ;
    } else
      FAs += a ;
  }
  return (hits+FAs > 0 ? 1.0-hits/((hits+FAs)*(1+misses)) : 1.0) ;
}

void	printVector(v, n, str)
  Real *v ;
  int n ;
  String str ;
{
  int i ;
  for (i=0 ; i<n ; i++) {
    fprintf(dout, str, v[i]) ;
    fprintf(dout, " ") ;
  }
  fprintf(dout, "\n") ;
}

Real		groupOutputProximity(group, v)
  Group	group ;
  Real	*v ;
{
  static Real	*output ;

  if (group->numUnits > maxUnits) {
    extendList((void *)&output, sizeof(Real*), maxUnits, group->numUnits) ;
    maxUnits = group->numUnits ;
  }
  
  saveOutputs(group, output) ;
  return proximity(output, v, group->numUnits) ;
}


/**********************************************************************/
static void	unitIncSSE(unit, data)
  Unit	unit ;
  void	*data ;
{
  if (!isNaN(unit->target)) {
    Real	error = unit->target - unit->output ;
    *(Real *)data += error*error ;
  }
}
static void	unitIncCE(unit, data)
  Unit	unit ;
  void	*data ;
{
  if (!isNaN(unit->target)) {
    Real	output = CLIP(unit->output, MIN_UNIT_OUTPUT, MAX_UNIT_OUTPUT) ;
    *(Real *)data += INFO(output, unit->target) ;
  }
}
static void	unitIncCityBlock(unit, data)
  Unit	unit ;
  void	*data ;
{
  if (!isNaN(unit->target)) 
    *(Real *)data += fabs(unit->target - unit->output) ;
}


static	Real	*_outputs=NULL, *_targets=NULL ;
static	int	_numOutputs = 0 ;

static void	perfCountUnit(unit, data)
  Unit		unit ;
  void		*data ;
{
  if (!isNaN(unit->target)) _unitIdx++ ;
}
static	void	perfSaveOutput(unit, data)
  Unit	unit ;
  void	*data ;
{
  if (!isNaN(unit->target)) ((Real *)data)[_unitIdx++] = unit->output ;
}
static	void	perfSaveTarget(unit, data)
  Unit	unit ;
  void	*data ;
{
  if (!isNaN(unit->target)) ((Real *)data)[_unitIdx++] = unit->target ;
}
/***********************************************************************/
int	command_printPerformance(tokc, tokv)
  int	tokc ;
  char	**tokv ;
{
  Real	measure ;

  IUsage("");
  if (GiveHelp(tokc)) {
    ISynopsis("Print performance measures for network.");
    IHelp
      (IHelpArgs,
       "Print performance measures (SSE, CE, city-block, cosine, correlation, divergence, GPE)",
       "NOTE: Normalization for divergence assumes only one OUTPUT group.",
       "EXAMPLES",
       "SEE ALSO",
       NULL) ;
    return 1 ;
  }

  if (currentNet == NULL)
    IErrorAbort("There is no current net.") ;

  measure = 0.0 ;
  netForAllUnits(currentNet, OUTPUT, (UnitDataProc)unitIncSSE, (void *)&measure) ;
  fprintf(dout, "%f ", measure) ;
  measure = 0.0 ;
  netForAllUnits(currentNet, OUTPUT, (UnitDataProc)unitIncCE, (void *)&measure) ;
  fprintf(dout, "%f ", measure) ;
  measure = 0.0 ;
  netForAllUnits(currentNet, OUTPUT, (UnitDataProc)unitIncCityBlock, (void *)&measure) ;
  fprintf(dout, "%f ", measure) ;
  _unitIdx = 0 ;
  netForAllUnits(currentNet, OUTPUT, (UnitDataProc)perfCountUnit, NULL) ;
  if (!_outputs || _unitIdx > _numOutputs) {
    _numOutputs = _unitIdx ;
    _outputs = (Real *)calloc(_numOutputs, sizeof(Real)) ;
    _targets = (Real *)calloc(_numOutputs, sizeof(Real)) ;
  }
  _unitIdx = 0 ;
  netForAllUnits(currentNet, OUTPUT, (UnitDataProc)perfSaveOutput, (void *)_outputs) ;
  _unitIdx = 0 ;
  netForAllUnits(currentNet, OUTPUT, (UnitDataProc)perfSaveTarget, (void *)_targets) ;
  fprintf(dout, "%f ", proximity(_outputs, _targets, _numOutputs)) ;
  fprintf(dout, "%f ", correlation(_outputs, _targets, _numOutputs)) ;
  fprintf(dout, "%f ", divergence(_outputs, _targets, _numOutputs)) ;
  fprintf(dout, "%f ", gpe(_outputs, _targets, _numOutputs)) ;
  fprintf(dout, "\n") ;
  return 1 ;
}
/**********************************************************************/
static Real	alpha ;
void	initEpoch (net, onset)
  Net	net ;
  int	onset ;
{
  /* only use alpha after onset epochs */
  if (onset) {
    if (net->mz->nIterations == 0) {
      alpha = net->mz->alpha ;
      net->mz->alpha = 0 ;
      if (alpha > 0.0)
	fprintf(dout,"Setting alpha to 0.0 until iteration %d\n", onset) ;
    }
    if (net->mz->nIterations == onset) {
      net->mz->alpha = alpha ;
      if (alpha != 0.0)
	fprintf(dout,"Resetting alpha from 0.0 to %f\n", alpha) ;
    }
  }

  /* this is really for checkpoint.c */
  fflush(dout) ;
}
/**********************************************************************/
int	lastTargetSlice(net)
  Net net;
{
  int	slice, groupIdx, unitIdx ;
  Group	*group ;
  Unit	*unit ;

  for (slice = net->timeSlices-1; slice >= 0; slice--) 
    for (groupIdx = 0, group = net->group ;
	 groupIdx < net->numGroups ;
	 groupIdx++, group++) 
      if ((*group)->type&OUTPUT)
	for (unitIdx = 0, unit = (*group)->unit ;
	     unitIdx < (*group)->numUnits ;
	     unitIdx++, unit++) 
	  if (!isNaN((*unit)->targetHistory[slice])) return slice+1 ;
  if (slice < 0) return net->timeSlices ;
}
/**********************************************************************/
int	numDiffFromTarget(net, tolerance)
  Net	net ;
  double tolerance ;
{
  int	count = 0, unitIdx, groupIdx ;
  Unit	*unit ;
  Group	*group ;

  for (groupIdx=0, group=net->group;
       groupIdx < net->numGroups; 
       groupIdx++, group++)
    if ((*group)->type&OUTPUT)
      for (unitIdx=0, unit=(*group)->unit; 
	   unitIdx < (*group)->numUnits;
	   unitIdx++, unit++) {
	Real	target = (*unit)->target ;
	if (!isNaN(target) && fabs(target-(*unit)->output) >= tolerance)
	  count++ ;
      }
  return count ;
}
/**********************************************************************/
Real	outputProximity(net)
  Net	net ;
{
  int	unitIdx, groupIdx ;
  Real l12 = 0.0, l1 = 0.0, l2 = 0.0;
  Unit	*unit ;
  Group	*group ;

  for (groupIdx=0, group=net->group;
       groupIdx < net->numGroups; 
       groupIdx++, group++)
    if ((*group)->type&OUTPUT)
      for (unitIdx=0, unit=(*group)->unit; 
	   unitIdx < (*group)->numUnits;
	   unitIdx++, unit++) {
	  Real	output = (*unit)->output ;
	  Real	target = (*unit)->target ;
	  if (!isNaN(target)) {
	    l1  += output*output ;
	    l2  += target*target ;
	    l12 += output*target ;
	  }
	}
  if (l1==0.0 || l2==0.0) return 0.0;
  else return (l12/sqrt(l1*l2));
}
/**********************************************************************/
static void	linkIncEnergy(link, data)
  Link	link ;
  void	*data ;
{
  *(Real *)data += link->weight * (2.0 * link->preUnit->output - 1.0) ;
}
  
static void	unitIncEnergy(unit, data)
  Unit	unit ;
  void	*data ;
{
  Real	unitEnergy = 0.0 ;
  unitForAllLinks(unit, INCOMING, (LinkDataProc)linkIncEnergy, (void *)&unitEnergy) ;
  *(Real *)data -= unitEnergy * (2.0 * unit->output - 1.0) ;
}

void	groupIncEnergy(group, data)
  Group	group ;
  void	*data ;
{
  Real	energy = 0.0 ;
  groupForAllUnits(group, (UnitDataProc)unitIncEnergy, (void *)&energy) ;
  *(Real *)data += energy ;
}

Real	groupEnergy(group)
  Group	group ;
{
  Real	energy = 0.0 ;
  groupForAllUnits(group, (UnitDataProc)unitIncEnergy, (void *)&energy) ;
  return energy ;
}

Real	netEnergy(net)
  Net	net ;
{
  Real	energy = 0.0 ;
  netForAllOtherUnits(net, BIAS, (UnitDataProc)unitIncEnergy, (void *)&energy) ;
  return energy ;
}
/**********************************************************************/
Real	aveOutputActivity(net,thresh)
  Net	net ;
  double thresh ;
{
  int	unitIdx, groupIdx, count = 0 ;
  Real	activity = 0 ;
  Unit	*unit ;
  Group	*group ;

  for (groupIdx=0, group=net->group;
       groupIdx < net->numGroups; 
       groupIdx++, group++)
    if ((*group)->type&OUTPUT)
      for (unitIdx=0, unit=(*group)->unit; 
	   unitIdx < (*group)->numUnits;
	   unitIdx++, unit++) {
	if ((*unit)->output >= thresh)
	  activity += (*unit)->output ;
	count++ ;
      }
  if (count > 0) return activity/(Real)count ;
  else           return 0.0 ;
}
/**********************************************************************/
Real	aveOutputStress(net)
  Net	net ;
{
  int	unitIdx, groupIdx, count = 0 ;
  Real	stress = 0 ;
  Unit	*unit ;
  Group	*group ;

  for (groupIdx=0, group=net->group;
       groupIdx < net->numGroups; 
       groupIdx++, group++)
    if ((*group)->type&OUTPUT)
      for (unitIdx=0, unit=(*group)->unit; 
	   unitIdx < (*group)->numUnits;
	   unitIdx++, unit++) {
	Real	output = (*unit)->output ;
	stress += unitStress(output) ;
	count++ ;
      }
  if (count > 0) return stress/(Real)count ;
  else           return 0.0 ;
}
/**********************************************************************/
Real	groupAveOutputStress(group)
  Group group ;
{
  int	unitIdx, count = 0 ;
  Real	stress = 0.0 ;
  Unit	*unit ;

  for (unitIdx=0, unit=group->unit; 
       unitIdx < group->numUnits;
       unitIdx++, unit++) {
    Real	output = (*unit)->output ;
    stress += unitStress(output) ;
    count++ ;
  }
  if (count > 0) return stress/(Real)count ;
  else           return 0.0 ;
}
/**********************************************************************/
Boolean	groupOutputCorrect(group, tolerance)
  Group  group ;
  double tolerance ;
{
  int	unitIdx ;
  Unit	*unit ;

  if (tolerance <= 0.0)    return FALSE ;

  for (unitIdx=0, unit=group->unit; 
       unitIdx < group->numUnits;
       unitIdx++, unit++) {
    Real	output = (*unit)->output ;
    Real	target = (*unit)->target ;
    if (!isNaN(target) && fabs(target - output) > tolerance) return FALSE ;
  }
  return TRUE ;
}

Boolean	outputCorrect(net, tolerance)
  Net	net ;
  double tolerance ;
{
  int	groupIdx ;
  Group	*group ;

  if (tolerance <= 0.0) return FALSE ;

  for (groupIdx=0, group=net->group;
       groupIdx < net->numGroups; 
       groupIdx++, group++) {
    if ((*group)->type&OUTPUT &&
	!groupOutputCorrect((*group), tolerance)) return FALSE ;
  }
  return TRUE ;
}

/**********************************************************************/
Boolean	groupOutputCorrectWTA(group, tolerance)
  Group  group ;
  double tolerance ;
{
  int	unitIdx ;
  Real	correctOutput = -MAXFLOAT, maxOtherOutput = -MAXFLOAT ;
  Unit	*unit ;

  if (! group->type&OUTPUT) return TRUE ;
  if (tolerance <= 0.0)    return FALSE ;

  for (unitIdx=0, unit=group->unit; 
       unitIdx < group->numUnits;
       unitIdx++, unit++) {
    Real	output = (*unit)->output ;
    if      ((*unit)->target==1.0)    correctOutput = output ;
    else if (output > maxOtherOutput) maxOtherOutput = output ;
  }    
  if (correctOutput-maxOtherOutput < tolerance) return FALSE ;
  return TRUE ;
}

Boolean	outputCorrectWTA(net, tolerance)
  Net	net ;
  double tolerance ;
{
  int	groupIdx ;
  Group	*group ;

  if (tolerance <= 0.0) return FALSE ;

  for (groupIdx=0, group=net->group;
       groupIdx < net->numGroups; 
       groupIdx++, group++) {
    if ((*group)->type&OUTPUT &&
	!groupOutputCorrectWTA((*group), tolerance)) return FALSE ;
  }
  return TRUE ;
}
/**********************************************************************/
Boolean	attentionCorrectWTA(Attention, Done, tolerance)
  Group  Attention, Done ;
  double tolerance ;
{
  int	unitIdx ;
  Real	correctOutput = -MAXFLOAT, maxOtherOutput = -MAXFLOAT ;
  Real	doneOutput = Done->unit[0]->output ;
  Real	doneTarget = Done->unit[0]->target ;
  Unit	*unit ;

  if (tolerance <= 0.0)    return FALSE ;

  for (unitIdx=0, unit=Attention->unit; 
       unitIdx < Attention->numUnits;
       unitIdx++, unit++) {
    Real	output = (*unit)->output ;
    if      ((*unit)->target==1.0)    correctOutput  = output ;
    else if (output > maxOtherOutput) maxOtherOutput = output ;
  }
  if      (doneTarget == 1.0)           correctOutput  = doneOutput ;
  else if (doneOutput > maxOtherOutput) maxOtherOutput = doneOutput ;
  if (correctOutput-maxOtherOutput < tolerance) return FALSE ;
  return TRUE ;
}
/**********************************************************************/
Boolean	outputClean(net, tolerance)
  Net	net ;
  double tolerance ;
{
  int	groupIdx ;
  Group	*group ;

  if (tolerance <= 0.0) return FALSE ;

  for (groupIdx=0, group=net->group;
       groupIdx < net->numGroups; 
       groupIdx++, group++) {
    if ((*group)->type&OUTPUT &&
	!groupOutputClean((*group), tolerance)) return FALSE ;
  }
  return TRUE ;
}
/**********************************************************************/
Boolean	groupOutputClean(group, tolerance)
  Group  group ;
  double tolerance ;
{
  int	unitIdx ;
  Unit	*unit ;

  if (tolerance <= 0.0)    return FALSE ;

  if (!strcmp(group->name, "Phonemes"))
    return SMoutputClean(group, tolerance) ;

  for (unitIdx=0, unit=group->unit; 
       unitIdx < group->numUnits;
       unitIdx++, unit++) {
    Real	output = (*unit)->output ;
    Real	target = floor(output+0.5) ;
    if (fabs(target - output) > tolerance) return FALSE ;
  }
  return TRUE ;
}
/**********************************************************************/
Boolean	groupOutputCleanWTA(group, tolerance)
  Group  group ;
  double tolerance ;
{
  int	unitIdx ;
  Real	bestOutput = -MAXFLOAT, nextBestOutput = -MAXFLOAT ;
  Unit	*unit ;

  if (! group->type&OUTPUT) return TRUE ;
  if (tolerance <= 0.0)    return FALSE ;

  for (unitIdx=0, unit=group->unit; 
       unitIdx < group->numUnits;
       unitIdx++, unit++) {
    Real	output = (*unit)->output ;
    if (output > bestOutput) {
      if (bestOutput > nextBestOutput) nextBestOutput = bestOutput ;
      bestOutput = output ;
    } else if (output > nextBestOutput) nextBestOutput = output ;
  }    
  if (bestOutput-nextBestOutput < tolerance) return FALSE ;
  return TRUE ;
}
/**********************************************************************/
Boolean	attentionCleanWTA(Attention, Done, tolerance)
  Group  Attention, Done ;
  double tolerance ;
{
  int	unitIdx ;
  Real	bestOutput = Done->unit[0]->output, nextBestOutput = -MAXFLOAT ;
  Unit	*unit ;

  if (tolerance <= 0.0)    return FALSE ;

  for (unitIdx=0, unit=Attention->unit; 
       unitIdx < Attention->numUnits;
       unitIdx++, unit++) {
    Real	output = (*unit)->output ;
    if (output > bestOutput) {
      if (bestOutput > nextBestOutput) nextBestOutput = bestOutput ;
      bestOutput = output ;
    } else if (output > nextBestOutput) nextBestOutput = output ;
  }    
  if (bestOutput-nextBestOutput < tolerance) return FALSE ;
  return TRUE ;
}
/**********************************************************************/
Boolean	groupOutputCleanOff(group, tolerance)
  Group	 group ;
  double tolerance ;
{
  int	unitIdx ;
  Unit	*unit ;

  if (tolerance <= 0.0) return FALSE ;

  for (unitIdx=0, unit=group->unit; 
       unitIdx < group->numUnits;
       unitIdx++, unit++) {
    if ((*unit)->output > tolerance) return FALSE ;
  }
  return TRUE ;
}
/**********************************************************************/
Real	sumOutput(group)
  Group	group ;
{
  int	unitIdx ;
  Real	sumOutput = 0.0 ;
  Unit	*unit ;

  for (unitIdx = 0, unit = group->unit ;
       unitIdx < group->numUnits ;
       unitIdx++, unit++)
    sumOutput += (*unit)->output ;
  return sumOutput ;
}
/**********************************************************************/
Real	maxOutput(group)
  Group	group ;
{
  int	unitIdx ;
  int	maxIdx = 0 ;
  Real	maxOutput = -1.0 ;
  Unit	*unit ;

  for (unitIdx = 0, unit = group->unit ;
       unitIdx < group->numUnits ;
       unitIdx++, unit++)
    if ((*unit)->output > maxOutput) {
      maxOutput = (*unit)->output ;
      maxIdx    = unitIdx ;
    }
  return maxOutput ;
}
/**********************************************************************/
Real	maxOutputAt(group, time)
  Group	group ;
  int	time ;
{
  int	unitIdx ;
  int	maxIdx = 0 ;
  Real	maxOutput = -1.0 ;
  Unit	*unit ;

  for (unitIdx = 0, unit = group->unit ;
       unitIdx < group->numUnits ;
       unitIdx++, unit++) {
    Real output = (*unit)->outputHistory[time] ;
    if (output > maxOutput) {
      maxOutput = output ;
      maxIdx    = unitIdx ;
    }
  }
  return maxOutput ;
}
/**********************************************************************/
int	maxOutputPosition(group)
  Group	group ;
{
  int	unitIdx ;
  int	maxIdx = 0 ;
  Real	maxOutput = -1.0 ;
  Unit	*unit ;

  for (unitIdx = 0, unit = group->unit ;
       unitIdx < group->numUnits ;
       unitIdx++, unit++)
    if ((*unit)->output > maxOutput) {
      maxOutput = (*unit)->output ;
      maxIdx    = unitIdx ;
    }
  return maxIdx ;
}
/**********************************************************************/
int	maxTargetPosition(group)
  Group	group ;
{
  int	unitIdx ;
  int	maxIdx = 0 ;
  Real	maxTarget = -1.0 ;
  Unit	*unit ;

  for (unitIdx = 0, unit = group->unit ;
       unitIdx < group->numUnits ;
       unitIdx++, unit++)
    if ((*unit)->target > maxTarget) {
      maxTarget = (*unit)->target ;
      maxIdx    = unitIdx ;
    }
  return maxIdx ;
}
/**********************************************************************/
void	cleanupWTA(group)
  Group	group ;
{
  int	unitIdx, maxIdx ;
  Unit	*unit ;

  maxIdx = maxOutputPosition(group) ;
  for (unitIdx = 0, unit = group->unit ;
       unitIdx < group->numUnits ;
       unitIdx++, unit++)
    (*unit)->output = (unitIdx == maxIdx ? 1.0 : 0.0) ;
}
/**********************************************************************/
void	cleanupThreshold(group, thresh)
  Group	group ;
  double thresh ;
{
  int	unitIdx ;
  Unit	*unit ;

  for (unitIdx = 0, unit = group->unit ;
       unitIdx < group->numUnits ;
       unitIdx++, unit++)
    (*unit)->output = ((*unit)->output >= thresh ? 1.0 : 0.0) ;
}
/**********************************************************************/
void	setBias(unit, data)
  Unit	unit ;
  void	*data ;
{
  if (unit->numIncoming > 0)
    unit->incomingLink[0]->weight = *(Real *)data ;
}

int	command_setBiases (tokc, tokv)
  int	tokc ;
  char	*tokv[] ;
{
  char	*name ;
  Real	value ;
  int	numUnits, idx ;
  Group	group, bias ;

  IUsage("<group-name> <value>") ;
  if (GiveHelp(tokc)) {
    ISynopsis("set the biases of all units in a group") ;
    IHelp
      (IHelpArgs,
       "Sets the bias weights of all units in a group to a specified value.",
       NULL);
    return 0;
  }

  if (tokc != 3)
    IErrorAbort(IPrintUsage(tokv[0], usage)) ;

  name = tokv[1] ;
  value = atof(tokv[2]) ;

  /* Check for all the possible errors */
  if (currentNet == NULL)
    IErrorAbort("There is no current net.") ;
  group = groupFromName(currentNet, name) ;
  if (group == NULL)
    IErrorAbort("Group \"%s\" does not exist.", group->name) ;
  
  groupForAllUnits(group, (UnitDataProc)setBias, (void *)&value) ;

  return 1 ;
}

#define rME_TARGET_VARIANCE 6.12
static Real	_rME_sum ;
/**********************************************************************/
/**********************************************************************/
void	unitRandMaxEntropy(unit, data)
  Unit	unit ;
  void	*data ;
{
  int i ;
  int	numIn = unit->numIncoming ;
  Real sd, sum ;

  if (numIn <= 0) return ;
  sd = rME_TARGET_VARIANCE / sqrt((Real)(numIn-1.0)) ;
  sum = 0.0 ;
  for (i = 1 ; i < numIn ; i++) {	/* skip bias */
    Real w = gaussianNoise(sd) ;
    unit->incomingLink[i]->weight = w ;
    sum += w ;
  }
  sum /= (Real)(numIn-1.0) ;
  for (i = 1 ; i < numIn ; i++)
    unit->incomingLink[i]->weight -= sum ;

  /* zero bias */
  unit->incomingLink[0]->weight = 0.0 ;
}

/**********************************************************************/
int	command_randomizeMaxEntropy (tokc, tokv)
  int	tokc ;
  char	*tokv[] ;
{
  String postGroupRE = NULL ;
  Group	*postGroupArray,  *postGroup ;
  struct ARG_DESC *argd ;

  if (currentNet == NULL)
    IErrorAbort("There is no current net.") ;

  argd = StartArgs(tokv[0]);
  Args(argd, "[-PpostGroup <regexp>%s%?]",	&postGroupRE,	"regular expression specifying destination groups.");
  EndArgs(argd);
  if (GiveHelp(tokc)) {
    ISynopsis("randomize weights to maximize entropy of unit states") ;
    IHelp
      (tokc, tokv[0], NULL, synopsis,
       "Randomize weights to maximize entropy of unit states.",
      NULL);
    return 0;
  }
  ParseArgs(argd, tokc, tokv, 0);

  postGroupArray = groupsFromRegex(currentNet, postGroupRE) ;
  if (postGroupArray == NULL)
    IErrorAbort("Bad string: \"%s\"", postGroupRE) ;
  if (postGroupArray[0] == NULL)
    netForAllUnits(currentNet, ALL, (UnitDataProc)unitRandMaxEntropy, NULL) ;
  else 
    for (postGroup = postGroupArray ; *postGroup != NULL ; ++postGroup)
      groupForAllUnits(*postGroup, (UnitDataProc)unitRandMaxEntropy, NULL) ;
  free(postGroupArray) ;
  return 1 ;
}

/***********************************************************************/
static void printOutputScores(group, data)
  Group group ;
  void  *data ;
{
  Real tolerance = *(Real *)data ;
  Real  stress = 0.0, maxZero = 0.0, minOne = 1.0 ;
  int	unitIdx, countP5 = 0, countTol = 0 ;
  Unit	*unit ;

  if (tolerance > 0.0) 
    for (unitIdx=0, unit=group->unit; 
	 unitIdx < group->numUnits;
	 unitIdx++, unit++) {
      Real	output = (*unit)->output ;
      Real	target = (*unit)->target ;
      if (!isNaN(target)) {
	Real diff = fabs(target - output) ;
	if (diff > 0.5)       countP5++ ;
	if (diff > tolerance) countTol++ ;
	if        (target == 0.0) {
	  if (output > maxZero) maxZero = output ;
	} else if (target == 1.0) {
	  if (output < minOne)  minOne  = output ;
	}
      }
      stress += unitStress(output) ;
    }
  fprintf(dout, "%s %d %d %f %f %f %f ", group->name, countP5, countTol, maxZero, minOne, stress/(Real)group->numUnits, groupError(group)) ;
}

/**********************************************************************/
static void	saveSigInvToTotalInput(unit, data)
  Unit		unit ;
  void		*data ;
{
  unit->totalInput = sigmoidInv(unit->output) ;
}
/**********************************************************************/
static void	distortOutputFromTotalInput(unit, data)
  Unit	unit ;
  void	*data ;
{
  Real input = unit->totalInput + gaussianNoise(*(Real *)data) ;
  unit->output = sigmoid(input) ;
}

static Boolean noNaNs(v, n)
     Real *v ;
     int n ;
{
  int i ;
  for (i = 0 ; i < n ; i++) if (isNaN(v[i])) return FALSE ;
  return TRUE ;
}

/**********************************************************************/
int command_scoreOutputs (tokc, tokv)
  int	tokc ;
  String	tokv[] ;
{
  int		ex, exM, r, reps, example_n = -1, exampleType = TRAINING ;
  int		i, best, bestDiff ;
  Real		tolerance = 0.2, sd = 0.0, thisMatch, bestMatch ;
  Real		**bestMatch_targets, *currentStates ;
  String	type=NULL, name=NULL, bestMatch_name=NULL ;
  Group		bestMatch_group ;
  ExampleSet	exampleSet ;
  struct ARG_DESC *argd ;

  if (currentNet == NULL) IErrorAbort("There is no current net.") ;
  argd = StartArgs(tokv[0]);
  Args(argd, "[-Ptype <type>%s%?]", 	 &type, 	"type of example set (one of TRAINING TESTING VALIDATION)");
  Args(argd, "[-Pname <str>%s%?]", 	 &name, 	"only run examples with name matching <string>") ;
  Args(argd, "[-PbestMatchGroup <name>%s%?]", &bestMatch_name, 	"compute best-matching example over group <name>") ;
  Args(argd, "[-Ptolerance <real>%r%?]", &tolerance, 	"correct tolerance (diff from target; default 0.2)") ;
  Args(argd, "[-Pexample <n>%i%?]", 	 &example_n, 	"example to run");
  Args(argd, "[-Psd <real>%r%?]",	 &sd,		"sd of Gaussian noise to add to output units' inputs") ;
  Args(argd, "[-Preps <n>%i%?]", 	 &reps, 	"number of replications of noise (sd) to run");
  EndArgs(argd);
  if (GiveHelp(tokc)) {
    ISynopsis("print states/correlations of units") ;
    IHelp
      (tokc, tokv[0], NULL, synopsis,
       NULL);
    return 0;
  }
  ParseArgs(argd, tokc, tokv, 0);
  if (type) exampleType = findMask(type) ;
  exampleSet = netGetExampleSet(currentNet, exampleType) ;
  if (exampleSet == NULL) IErrorAbort("No example set specified for net \"%s\".", currentNet->name) ;

  if (bestMatch_name) {
    bestMatch_group = groupFromName(currentNet, bestMatch_name) ;
    if (bestMatch_group == NULL) IErrorAbort("scoreOutputs: Group \"%s\" does not exists.", bestMatch_name) ;
    bestMatch_targets = (Real **)calloc(exampleSet->numExamples, sizeof(Real*)) ;
    Mreset(exampleSet) ;
    for (exM = 0; exM < exampleSet->numExamples; exM++) {
      if ((example_n >= 0 && exM != example_n) ||
	  (name           && !strstr(exampleSet->example[exM].name, name)))
	continue ;
      bestMatch_targets[exM] = (Real *)calloc(bestMatch_group->numUnits, sizeof(Real)) ;
      MgetNumber(exampleSet, exM) ;
      saveTargetsAt(0, bestMatch_group, bestMatch_targets[exM]) ;
    }
    currentStates = (Real *)calloc(bestMatch_group->numUnits, sizeof(Real)) ;
  }

  Mreset(exampleSet) ;
  for (ex = 0; ex < exampleSet->numExamples; ex++) {
    if ((example_n >= 0 && ex != example_n) ||
	(name           && !strstr(exampleSet->example[ex].name, name)))
      continue ;
    MgetNumber(exampleSet, ex) ;
    currentNet->error = 0.0 ;
    MupdateNetActivities(currentNet) ;
    if (bestMatch_name) {
      saveOutputs(bestMatch_group, currentStates) ;
      for (exM = 0, best = -1, bestMatch = -2.0; exM < exampleSet->numExamples; exM++) {
	if ((example_n >= 0 && exM != example_n) ||
	    (name           && !strstr(exampleSet->example[exM].name, name)))
	  continue ;
	if (noNaNs(bestMatch_targets[exM], bestMatch_group->numUnits)) {
	  thisMatch = correlation(currentStates, bestMatch_targets[exM], bestMatch_group->numUnits) ;
	  if (thisMatch > bestMatch) { 
	    best = exM ; 
	    bestMatch = thisMatch ; 
	    bestDiff = binarizedDifference(currentStates, bestMatch_targets[best], bestMatch_group->numUnits) ;
	  }
	}
      }
      fprintf(dout, "best %d %4.3f %d %s ", best, bestMatch, bestDiff, exampleSet->example[best].name) ;
    }
    if (sd > 0.0) {
      netForAllUnits(currentNet, OUTPUT, (UnitDataProc)saveSigInvToTotalInput, NULL) ;
      for (r = 0 ; r < reps ; r++) {
	netForAllUnits(currentNet, OUTPUT, (UnitDataProc)distortOutputFromTotalInput, (void*)&sd) ;
	fprintf(dout, "%s %f ", exampleSet->example[ex].name, currentNet->error) ;
	netForAllGroups(currentNet, OUTPUT, (GroupDataProc)printOutputScores, (void *)&tolerance) ;
	fprintf(dout, "%f\n", sd) ;
      }
    } else {
      fprintf(dout, "%s %f ", exampleSet->example[ex].name, currentNet->error) ;
      netForAllGroups(currentNet, OUTPUT, (GroupDataProc)printOutputScores, (void *)&tolerance) ;
      fprintf(dout, "\n") ;
    }
    fflush(dout) ;
  }
  return 1 ;
}
    
/**********************************************************************/
#define MnonzeroCount(unit) (unit->outputDeriv)

static void	zeroNonzeroCount(unit, data)
  Unit		unit ;
  void		*data ;
{
  MnonzeroCount(unit) = 0.0 ;
}

static void	incNonzeroCount(unit, data)
  Unit		unit ;
  void		*data ;
{
  int s ;

  if (unit->net->type & RECURRENT) {
    for (s = 0 ; s < unit->net->timeSlices; s++)
      if (!isNaN(unit->clampingHistory[s]) && unit->clampingHistory[s] != 0.0) {
	MnonzeroCount(unit)++ ;
	break ;
      }
  } else {
    if (unit->output != 0.0) MnonzeroCount(unit)++ ;
  }
}

static void	pruneZeroCountOutgoingLinks(unit, data)
  Unit		unit ;
  void		*data ;
{
  char	command[BUFSIZ] ;

  if (MnonzeroCount(unit) == 0.0) {
    /* prune */
    fprintf(dout, "Disconnecting %s\n", unit->name) ;
    sprintf(command,"disconnectUnits %s .*", unit->name) ;
    IDoCommandLine(command) ;
  }
}

int command_pruneInputs (tokc, tokv)
  int	tokc ;
  String	tokv[] ;
{
  int		ex, exampleType = TRAINING ;
  String	type=NULL, name=NULL ;
  ExampleSet	exampleSet ;
  struct ARG_DESC *argd ;

  if (currentNet == NULL) IErrorAbort("There is no current net.") ;
  argd = StartArgs(tokv[0]);
  Args(argd, "[-Ptype <type>%s%?]", 	 &type, 	"type of example set (one of TRAINING TESTING VALIDATION)");
  EndArgs(argd);
  if (GiveHelp(tokc)) {
    ISynopsis("prune outgoing weights of unused input units") ;
    IHelp
      (tokc, tokv[0], NULL, synopsis,
       NULL);
    return 0;
  }
  ParseArgs(argd, tokc, tokv, 0);
  if (type) exampleType = findMask(type) ;
  exampleSet = netGetExampleSet(currentNet, exampleType) ;
  if (exampleSet == NULL) IErrorAbort("No example set specified for net \"%s\".", currentNet->name) ;

  netForAllUnits(currentNet, INPUT, (UnitDataProc)zeroNonzeroCount, NULL) ;
  Mreset(exampleSet) ;
  for (ex = 0; ex < exampleSet->numExamples; ex++) {
    MgetNumber(exampleSet, ex) ;
    netForAllUnits(currentNet, INPUT, (UnitDataProc)incNonzeroCount, NULL) ;
  }
  netForAllUnits(currentNet, INPUT, (UnitDataProc)pruneZeroCountOutgoingLinks, NULL) ;
  netForAllUnits(currentNet, INPUT, (UnitDataProc)zeroNonzeroCount, NULL) ;
  return 1 ;
}


/**********************************************************************/
void	unitPrintOutput(unit, data)
  Unit	unit ;
  void	*data ;
{
  fprintf(dout, "%f ", (isLesionedUnit(unit) ? *(Real *)data : unit->output)) ;
}

int	command_printGroupOutputs (tokc, tokv)
  int	tokc ;
  char	*tokv[] ;
{
  char	*name ;
  Real	value ;
  int	numUnits, idx ;
  Group	group ;

  IUsage("<group-name> <value>") ;
  if (GiveHelp(tokc)) {
    ISynopsis("print outputs of <group-name>, substituting <value> for NaN"); 
    IHelp
      (IHelpArgs,
       "print outputs of <group-name>, substituting <value> for NaN" ,
       NULL);
    return 0;
  }

  if (tokc != 3)
    IErrorAbort(IPrintUsage(tokv[0], usage)) ;

  name = tokv[1] ;
  value = atof(tokv[2]) ;

  /* Check for all the possible errors */
  if (currentNet == NULL)
    IErrorAbort("There is no current net.") ;
  group = groupFromName(currentNet, name) ;
  if (group == NULL)
    IErrorAbort("Group \"%s\" does not exist.", group->name) ;
  
  groupForAllUnits(group, (UnitDataProc)unitPrintOutput, (void *)&value) ;
  fprintf(dout, "\n") ;

  return 1 ;
}
