#include <commands.h>
#include <simulator.h>
#include "barak.h"

/*********************************************************************
 * 				Lesioning
 *********************************************************************/
Boolean		isLesionedLink(link)
  Link		link ;
{
  return Mstatus(link)&LESIONED ;
}

void		lesionLink(link)
  Link		link ;
{
  if (isLesionedLink(link)) return ;
  link->deriv = 0.0 ;
  MlesionedWeight(link) = link->weight ;
  link->weight = 0.0 ;
  Mstatus(link) |= LESIONED ;
  Mcompiled(link->postUnit->net) = FALSE ;
}

void		unlesionLink(link)
  Link		link ;
{
  if (!isLesionedLink(link)) return ;
  link->weight = MlesionedWeight(link) ;
  Mstatus(link) &= ~LESIONED ;
  Mcompiled(link->postUnit->net) = FALSE ;
}


Boolean		isLesionedUnit(unit)
  Unit		unit ;
{
  return Mstatus(unit)&LESIONED ;
}

void		lesionUnit(unit)
  Unit		unit ;
{
  if (isLesionedUnit(unit)) return ;
  Mstatus(unit) |= LESIONED ;
  setNaN(unit->output) ;		/*  unit->output = 0.0 ; */
  if (unit->net->type & RECURRENT) {
    int i ;
    for (i = 0 ; i < unit->net->timeSlices ; i++)
      setNaN(unit->outputHistory[i]) ;
  }
  unitForAllLinks(unit, ALL, (LinkDataProc)lesionLink, NULL) ;
  Mcompiled(unit->net) = FALSE ;
}

void		unlesionUnit(unit)
  Unit		unit ;
{
  if (!isLesionedUnit(unit)) return ;
  unitForAllLinks(unit, ALL, (LinkDataProc)unlesionLink, NULL) ;
  unit->output = MinitialUnitOutput(unit->net) ;
  if (unit->net->type & RECURRENT) {
    int i ;
    for (i = 0 ; i < unit->net->timeSlices ; i++)
      unit->outputHistory[i] = MinitialUnitOutput(unit->net) ;
  }
  Mstatus(unit) &= ~LESIONED ;
  Mcompiled(unit->net) = FALSE ;
}

/*****************************************************************************/
Boolean		isClampedLink(link)
  Link		link ;
{
  return link->type&FROZEN ;
}

Boolean		isClampedUnit(unit)
  Unit		unit ;
{
  return Mstatus(unit)&CLAMPED ;
}

void		clampLink(link)
  Link		link ;
{
  link->type |= FROZEN ;
}

void		unclampLink(link)
  Link		link ;
{
  link->type &= !FROZEN ;
}


/*****************************************************************************/
/*****************************************************************************/
static void	clampLinksFromGroup(unit, data)
  Unit		unit ;
  void		*data ;
{
  Group	preGroup = (Group)data ;
  int	idx ;
  int	numIncoming = unit->numIncoming ;
  Link	*incoming   = unit->incomingLink ;

  for (idx = 0 ; idx < numIncoming ; ++idx) {
    Link        link    = incoming[idx] ;
    if (link->preUnit->group == preGroup) clampLink(link) ;
  }
}

static void	unclampLinksFromGroup(unit, data)
  Unit		unit ;
  void		*data ;
{
  Group	preGroup = (Group)data ;
  int	idx ;
  int	numIncoming = unit->numIncoming ;
  Link	*incoming   = unit->incomingLink ;

  for (idx = 0 ; idx < numIncoming ; ++idx) {
    Link        link    = incoming[idx] ;
    if (link->preUnit->group == preGroup) unclampLink(link) ;
  }
}

int	command_clampLinks (tokc, tokv)
  int	tokc ;
  char	*tokv[] ;
{
  int	unclamp = FALSE ;
  String preGroupRE = NULL, postGroupRE = NULL ;
  Group	*preGroupArray,  *preGroup ;
  Group	*postGroupArray, *postGroup ;
  struct ARG_DESC *argd ;

  if (currentNet == NULL)
    IErrorAbort("There is no current net.") ;

  argd = StartArgs(tokv[0]);
  Args(argd, "[-PpreGroup <regexp>%s%?]",	&preGroupRE,	"regular expression specifying source groups.");
  Args(argd, "[-PpostGroup <regexp>%s%?]",	&postGroupRE,   "regular expression specifying destination groups.");
  Args(argd, "[-Punclamp %P%?]",	  	&unclamp,	"unclamp rather than clamp");
  EndArgs(argd);
  if (GiveHelp(tokc)) {
    ISynopsis("clamp (i.e., freeze) the weights on all links between some groups") ;
    IHelp
      (tokc, tokv[0], NULL, synopsis,
       "Freeze the weights on links between groups.  Group",
       "names may  use regular expressions similar to those used in grep to ",
       "lesion more than one group or set of links." ,
       "EXAMPLE",
       "\txerion-> clampLinks -pre \"input\" -post \"hidden\"",
       "",
       "will clamp all of the links between the group \"input\" and the",
       "group \"hidden\".",
       "",
       "\txerion-> clampLinks -unclamp -pre \"input\" -post \"hidden\"",
       "",
       "will unclamp all of the links.",
      NULL);
    return 0;
  }
  ParseArgs(argd, tokc, tokv, 0);

  preGroupArray = groupsFromRegex(currentNet, preGroupRE) ;
  if (preGroupArray == NULL)
    IErrorAbort("Bad string: \"%s\"", preGroupRE) ;
  if (preGroupArray[0] == NULL)
    IErrorAbort("No match: \"%s\"", preGroupRE) ;

  if (preGroupRE != NULL) {
    postGroupArray = groupsFromRegex(currentNet, postGroupRE) ;
    if (postGroupArray == NULL)
      IErrorAbort("Bad string: \"%s\"", postGroupRE) ;
    if (postGroupArray[0] == NULL)
      IErrorAbort("No match: \"%s\"", postGroupRE) ;
  }
  for (preGroup = preGroupArray ; *preGroup != NULL ; ++preGroup) 
    for (postGroup = postGroupArray ; *postGroup != NULL ; ++postGroup) {
      fprintf(dout,"%slamping %s ==> %s links\n", 
	      (unclamp?"Unc":"C"),(*preGroup)->name, (*postGroup)->name) ;
      if (unclamp)
	groupForAllUnits(*postGroup, (UnitDataProc)unclampLinksFromGroup, *preGroup) ;
      else
	groupForAllUnits(*postGroup, (UnitDataProc)clampLinksFromGroup, *preGroup) ;
    }
  free(preGroupArray) ;
  free(postGroupArray) ;
  
  return 1 ;
}
/**********************************************************************/
