#include <commands.h>
#include <simulator.h>
#include <xerion/minimize.h>
#include <xerion/utilsDCP.h>
#include "ffbp.h"
#include "ffbp-fm.h"
#include "ffbp-artic.h"

/*********************************************************************
 * 				Forward models
 *********************************************************************/
#define printOutputs(net)\
	{ netForAllUnits(net, OUTPUT, (UnitDataProc)printOutput, NULL) ; \
	  fprintf(dout, "\n") ; }

static void	printOutput(unit, data)
  Unit		unit ;
  void		*data ;
{
  fprintf(dout,"%.4f ", unit->output) ;
}

#define printInputs(net)\
	{ netForAllUnits(net, INPUT, (UnitDataProc)printOutput, NULL) ; \
	  fprintf(dout, "\n") ; }


#define printTargets(net)\
	{ netForAllUnits(net, OUTPUT, (UnitDataProc)printTarget, NULL) ; \
	  fprintf(dout, "\n") ; }

static void	printTarget(unit, data)
  Unit		unit ;
  void		*data ;
{
  fprintf(dout,"%.4f ", unit->target) ;

}

#define printDerivs(net,g)\
	{ groupForAllUnits(net->group[g], (UnitDataProc)printDeriv, NULL) ; \
	  fprintf(dout, "\n") ; }

static void	printDeriv(unit, data)
  Unit		unit ;
  void		*data ;
{
  fprintf(dout,"%.4f ", unit->outputDeriv) ;
}

static int	findFirstInputGroupNum(net)

  Net	net ;
{
  int	g ;
  Group	group ;

  for (g = 0, group=net->group[g] ; g < net->numGroups ; group=net->group[++g])
    if (group->type&INPUT) return g ;
  return 0 ;
}

static int	findLastOutputGroupNum(net)
  Net	net ;
{
  int	g ;
  Group	group ;

  for (g = net->numGroups-1, group=net->group[g] ; g > 0 ; group=net->group[--g])
    if (group->type&OUTPUT) return g ;
  return 0 ;
}


/**********************************************************************/
static void	swapPredictedOutcomeWithGoal (net)
  Net	net ;
{
  int idx ;
  Group outcomeGp, goalGp ;
  
  if (!MoutcomeLayer(net) || !MgoalLayer(net))
    IErrorAbort("outcome or goal layer not specified!") ;
  else {
    outcomeGp = net->group[MoutcomeLayer(net)] ;
    goalGp    = net->group[MgoalLayer(net)] ;
    if (outcomeGp->numUnits != goalGp->numUnits) 
      IErrorAbort("outcome layer (%d) and goal layer (%d) have different number of units!",
		  outcomeGp->numUnits, goalGp->numUnits) ;
  }
  
  for (idx=0; idx < outcomeGp->numUnits; idx++) {
    Real tmp                     = outcomeGp->unit[idx]->output ;
    outcomeGp->unit[idx]->output = goalGp->unit[idx]->output ;
    goalGp->unit[idx]->output    = tmp ;
  }
}

/********************************************************************/
static void	copyActualOntoPredicted (net)
  Net	net ;
{
  int idx ;
  Group outcomeGp ;
  
  if (!MoutcomeLayer(net))
    IErrorAbort("outcome layer not specified!") ;
  else {
    outcomeGp = net->group[MoutcomeLayer(net)] ;
  }
  
  for (idx=0; idx < outcomeGp->numUnits; idx++) {
    outcomeGp->unit[idx]->output = outcomeGp->unit[idx]->target ;

  }
}


/**********************************************************************/
static void	copyGoalDerivOntoAction (net)
  Net	net ;
{
  int idx ;
  Group outcomeGp, actionGp ;
  
  if (!MoutcomeLayer(net) || !MactionLayer(net))
    IErrorAbort("outcome or goal layer not specified!") ;
  else {
    outcomeGp = net->group[MoutcomeLayer(net)] ;
    actionGp    = net->group[MactionLayer(net)] ;
    if (outcomeGp->numUnits != actionGp->numUnits) 
      IErrorAbort("outcome layer (%d) and action layer (%d) have different number of units!",
		  outcomeGp->numUnits, actionGp->numUnits) ;
  }
  
  for (idx=0; idx < outcomeGp->numUnits; idx++) {
    actionGp->unit[idx]->outputDeriv = outcomeGp->unit[idx]->outputDeriv ;
  }
}

/**********************************************************************/
static void	swapPredictedOutcomeWithTarget (net)
  Net	net ;
{
  int idx ;
  Group outcomeGp ;
  
  if (!MoutcomeLayer(net))
    IErrorAbort("outcome layer not specified!") ;
  else {
    outcomeGp = net->group[MoutcomeLayer(net)] ;
  }
  
  for (idx=0; idx < outcomeGp->numUnits; idx++) {
    Real tmp                     = outcomeGp->unit[idx]->output ;
    outcomeGp->unit[idx]->output = outcomeGp->unit[idx]->target ;
    outcomeGp->unit[idx]->target = tmp ;
  }
}


extern void	storeOutputSlice( unit, data )
  Unit unit;
  void *data;
{
  int slice;
  
  slice = *(int *)data;
/*  if( !unit->extension->outputSlice ) 
    unit->extension->outputSlice = (Real *)malloc( MmaxSlice(unit->net) * sizeof( Real ) );
*/
  unit->outputHistory[slice] = unit->output;
}

extern void	restoreOutputSlice( unit, data )
  Unit unit;
  void *data;
{
  int slice;
  
  slice = *(int *)data;
/*  if( !unit->extension->outputSlice ) 
    IErrorAbort( "outputSlice not defined for %s!", unit->name );
*/
  unit->output = unit->outputHistory[slice];
}

extern void	swapOutputSlice(unit, data)
  Unit		unit ;
  void		*data ;
{
  Real 	temp;
  int	slice;
  
  slice = *(int *)data;
/*  if( !(unit->extension->outputSlice) )
    unit->extension->outputSlice = (Real *)malloc( MmaxSlice(unit->net) * sizeof( Real ) );
*/
  temp = unit->output;
  unit->output = unit->outputHistory[slice];
  unit->outputHistory[slice] = temp;
}

extern void	setOutcomeProc (net)
  Net	net;
{
  int code;
  
  code = MoutcomeCode(net);
  switch( code ) {
    case 1:  MoutcomeProc(net) = netOutcomeID; break;
    case 2:  MoutcomeProc(net) = netOutcomeAcoustics; break;
    case 3:  MoutcomeProc(net) = netOutcomeDistort4bit; break;
    case 4:  MoutcomeProc(net) = netOutcomeGaussID; break;
    default: MoutcomeProc(net) = netOutcomeID; break;
  }
}

extern void	setNormalizedOff(group, data)
  Group		group ;
  void		*data ;
{
  MalreadyNormed( group ) = FALSE;
}

extern void	forwardNormalized( unitArray, numUnits )
  Unit	*unitArray;
  int	numUnits;
{
  int	j;
  Real	*normArray, totalInput, normTotal, maxInput;
  
  normArray = (Real *)malloc( (numUnits * sizeof( Real )) );
  
    for( j = 0, maxInput = -MAXREAL; j < numUnits; j++ ) {
      normArray[j] = sumTotalInput( unitArray[j] );
      if( normArray[j] > maxInput ) maxInput = normArray[j];
    }
    for( j = 0, normTotal = 0.0; j < numUnits; j++ ) {
      normArray[j] = exp( (normArray[j] + (log(MAXREAL/numUnits) - maxInput - 2.0)) );
      normTotal += normArray[j];
    }
    for( j = 0; j < numUnits; j++ ) {
      unitArray[j]->output = normArray[j] / normTotal;
      if( !MinRange01( unitArray[j]->output ) )
        IErrorAbort( "normalized unit %s is out of range: %f\n", 
          unitArray[j]->name, unitArray[j]->output );
    }
    
  free( normArray ); 
}

extern void	netForwardFM (net, exampleSet)
  Net		net ;
  ExampleSet	exampleSet ;
{
  int i, slice;
  Group	actionGp;
  Example example;
  
  if (!MoutcomeLayer(net)) MoutcomeLayer(net) = findLastOutputGroupNum(net) ;
  if (!MgoalLayer(net))    MgoalLayer(net)    = findFirstInputGroupNum(net) ;
  actionGp = net->group[MactionLayer(net)];
  example = &(exampleSet->example[MexampleNum(net)]);
  slice = ((net->type&RECURRENT) ? net->currentTime : 0);
  
  if( !exampleSet )
    IErrorAbort( "empty example set passed to netForwardFM!" );
    
  if (MinputScaleFactor(net)>0.0)
    netForAllUnits (net, INPUT, (UnitDataProc)scaleInput, &MinputScaleFactor(net)) ;
  if (MinputNoise(net)>0.0)
    netForAllUnits (net, INPUT, (UnitDataProc)addInputNoise, &MinputNoise(net)) ;
  if (MtargetExternalInput(net) > 0.0) {
    netForAllOtherUnits (net, INPUT|BIAS, (UnitDataProc)forwardTargExtInput, &MtargetExternalInput(net)) ;
    PMSPtargExtInRange(net) ;
  } else {
     netForAllOtherUnits (net, INPUT|BIAS, (UnitDataProc)forward, NULL) ;
  }
  MoutputsIncorrect(net) = 0 ;
  
    /* compute actual outcome as targets for outcome layer */
    FMcomputeActualOutcome(net) ;
 
    if( example->numTargets != net->group[MoutcomeLayer(net)]->numUnits )
      IErrorAbort( "example target (%d) and outcome layer (%d) must have same # of units!",
      example->numTargets, net->group[MoutcomeLayer(net)]->numUnits );
      
    /* set these targets on the example */      
    for( i = 0; i < example->numTargets; i++ )
      example->targetVector[slice][i] = net->group[MoutcomeLayer(net)]->unit[i]->target;
     
    /* compute goal state, currently for converting ascii phonemes to acoustic targets */
    if( example->name[0] == '/' )
      if( !(net->type&RECURRENT) ) 
        computeAcousticGoals( net, example );
      else
        groupForAllUnits( net->group[MgoalLayer(net)], (UnitDataProc)restoreOutputFromSlice, (void *)slice );
    
    /* compute goal state for gaussianID if necessary */
    if( MoutcomeProc(net) == netOutcomeGaussID && !MrandomExampleType(net) )
      computeGaussianIDgoal( net, example );
    
    /* compute prediction error */
    groupForAllUnits(net->group[MoutcomeLayer(net)], (UnitDataProc)incrementError, &MpredictionError(net)) ;
    
    /* swap predicted outcome and goal states */
    swapPredictedOutcomeWithGoal(net) ;    
    
    /* for distort FM, compute the distortion of goal if necessary */
    if( MoutcomeProc(net) == netOutcomeDistort4bit && !MrandomExampleType(net) )
      distort4bit(net) ;

    /* compute performance error (including any other OUTPUT layers) */
    netForAllUnits(net, OUTPUT, (UnitDataProc)incrementError, &(net->error)) ;
    
    /* restore predicted outcome and goal states */
    swapPredictedOutcomeWithGoal(net) ;
}

/**********************************************************************/
extern void	netBackwardFM (net)
  Net		net ; 
{
  /* outcome targets calculated by FMcomputeActualOutcome in netForward */
  /* use these as states and goal as targets */
  swapPredictedOutcomeWithGoal(net) ;
  swapPredictedOutcomeWithTarget(net) ;

  /* zero all hidden unit derivs */
  netForAllOtherUnits(net, INPUT|OUTPUT|BIAS,(UnitDataProc)zeroOutputDeriv,NULL) ;

  /* inject error based on ACTUAL vs. GOAL outcome */
  netForAllUnits(net, OUTPUT, (UnitDataProc)injectError, NULL) ;

  /* possibly inject unit cost in inverse model (including action layer) */
  if (MunitCost(net) != 0.0)
    netDoUnits(net, MactionLayer(net), 1, INPUT|OUTPUT|BIAS, FALSE, 
	       (UnitDataProc)injectUnitCost, NULL) ;

  /* bp through entire network but don't increment wt derivs in FM */
  netDoUnits(net, net->numGroups-1, MactionLayer(net)+1, INPUT|BIAS, FALSE, 
	     (UnitDataProc)backwardNoWtDvs, NULL) ;
	     
  netDoUnits(net, MactionLayer(net), 1, INPUT|BIAS, FALSE, 
	     (UnitDataProc)backward, NULL) ;

  /* restore predicted outcome states and actual outcome as targets */
  swapPredictedOutcomeWithTarget(net) ;
  swapPredictedOutcomeWithGoal(net) ;

  /* re-zero hidden unit derivs in FM */
  netDoUnits(net, net->numGroups-1, MactionLayer(net)+1, INPUT|OUTPUT|BIAS, FALSE, 
	     (UnitDataProc)zeroOutputDeriv, NULL) ;

  /* inject error based on ACTUAL vs. PREDICTED outcome */
  groupForAllUnits(net->group[MoutcomeLayer(net)], (UnitDataProc)injectError, NULL) ;

  /* possibly inject unit cost to FM hidden units */
  if (MunitCost(net) != 0.0)
    netDoUnits(net, net->numGroups-1, MactionLayer(net)+1, 
	       INPUT|OUTPUT|BIAS, FALSE, 
	       (UnitDataProc)injectUnitCost, NULL) ;

  /* bp only in FM */
  netDoUnits(net, net->numGroups-1, MactionLayer(net)+1, INPUT|BIAS, FALSE, 
	     (UnitDataProc)backward, NULL) ;
}

