
/**********************************************************************
 * $Id: bp.c,v 1.9 91/09/26 11:05:14 drew Exp Locker: drew $
 **********************************************************************/

/**********************************************************************
 *	  Copyright 1990,1991 by The University of Toronto,
 *		      Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its  documentation for  any purpose is  hereby granted  without
 * fee, provided that the above copyright notice appears in all copies
 * and  that both the  copyright notice  and   this  permission notice
 * appear in   supporting documentation,  and  that the  name  of  The
 * University  of Toronto  not  be  used in  advertising or  publicity
 * pertaining   to  distribution   of  the  software without specific,
 * written prior  permission.   The  University of   Toronto makes  no
 * representations  about  the  suitability of  this software  for any
 * purpose.  It  is    provided   "as is"  without express or  implied
 * warranty.
 *
 * THE UNIVERSITY OF TORONTO DISCLAIMS  ALL WARRANTIES WITH REGARD  TO
 * THIS SOFTWARE,  INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT SHALL THE UNIVERSITY OF TORONTO  BE LIABLE
 * FOR ANY  SPECIAL, INDIRECT OR CONSEQUENTIAL  DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR  PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
 * OUT  OF  OR  IN  CONNECTION WITH  THE  USE  OR PERFORMANCE  OF THIS
 * SOFTWARE.
 *
 **********************************************************************/

#include <stdio.h>
#include <math.h>
#include <values.h>
#include <commands.h>
#include <xerion/minimize.h>
#include <xerion/utilsDCP.h>
#include "ffbp.h"
#include "ffbp-fm.h"
#include "ffbp-artic.h"

#ifndef ibm
#include <nan.h>
#endif

struct TRACE    afterExample ;
struct TRACE    forwardEachTime ;

static Proc	updateNetActivity      ARGS((Net	net)) ;
static Proc	updateNetGradient      ARGS((Net	net)) ;

static void	calculateNetErrorDeriv ARGS((Net net, ExampleSet exampleSet)) ;
static void	calculateNetError      ARGS((Net net, ExampleSet exampleSet)) ;

static void	initNet   		ARGS((Net	 net)) ;
static void	deinitNet 		ARGS((Net	 net)) ;
static void	initGroup    		ARGS((Group	 group)) ;
static void	deinitGroup    		ARGS((Group	 group)) ;
static void	initUnit    		ARGS((Unit	 unit)) ;
static void	deinitUnit  		ARGS((Unit	 unit)) ;
static void	initLink    		ARGS((Link	 link)) ;
static void	deinitLink  		ARGS((Link	 link)) ;
static void	initExampleSet    	ARGS((ExampleSet exampleSet)) ;
static void	deinitExampleSet	ARGS((ExampleSet exampleSet)) ;
   

/**********************************************************************
 *	Name:		main 
 *	Description:	the main function, used for the xerion simulator
 *	Parameters:	
 *		int	argc	- the number of input args
 *		char	**argv  - array of argument strings from command 
 *				  line
 *	Return Value:	
 *		int	main	- 0
 ***********************************************************************/
int main (argc, argv)
  int	argc ;
  char	**argv ;
{
  /* Insert any private initialization routines here */
  setCreateNetHook(initNet) ;
  setDestroyNetHook(deinitNet) ;
  setCreateGroupHook(initGroup) ;
  setDestroyGroupHook(deinitGroup) ;
  setCreateUnitHook(initUnit) ;
  setDestroyUnitHook(deinitUnit) ;
  setCreateLinkHook(initLink) ;
  setDestroyLinkHook(deinitLink) ;
  setCreateExampleSetHook(initExampleSet) ;
  setDestroyExampleSetHook(deinitExampleSet) ;

  /* Perform initialization of the simulator */
  IStandardInit(&argc, argv);

  /* Enter loop that reads commands and handles graphics */
  ICommandLoop(stdin, stdout, NULL);

  return 0 ;
}

/***********************************************************************/
static void	initNet (net)
  Net	net ;
{
  net->calculateErrorDerivProc = calculateNetErrorDeriv ;
  net->calculateErrorProc      = calculateNetError ;
  net->activityUpdateProc      = updateNetActivity ;
  net->gradientUpdateProc      = updateNetGradient ;

  net->extension = (NetExtension)calloc(1, sizeof(NetExtensionRec)) ;

  /* non-zero defaults */
  MbpToOutputs(net)		= 1 ;
  MinitialContextOutput(net)	= 0.2 ;
  MpredictOutput(net)		= 1 ;
  MweightCostDecrement(net)	= 0.9 ;
  MaveErrorFactor(net)		= 0.95 ;
  MlengthScaleDerivs(net)	= 1 ;
  /* FM */
  MtrainComp(net)		= 1 ;
}

static void	deinitNet (net)
  Net	net ;
{
  if (net->extension != NULL)
    free(net->extension) ;
}

/***********************************************************************/
static void	initGroup(group)
  Group	group ;
{
  group->extension = (GroupExtension)calloc(1, sizeof(GroupExtensionRec)) ;

  MminBias(group)		= -MAXFLOAT ;
  MmaxBias(group)		=  MAXFLOAT ;
  MminIncomingWeight(group)	= -MAXFLOAT ;
  MmaxIncomingWeight(group)	=  MAXFLOAT ;
  MminOutgoingWeight(group)	= -MAXFLOAT ;
  MmaxOutgoingWeight(group)	=  MAXFLOAT ;
}

static void	deinitGroup(group)
  Group	group ;
{
  if (group->extension != NULL)
    free(group->extension) ;
}

/***********************************************************************/
static void	initUnit(unit)
  Unit	unit ;
{
  unit->extension = (UnitExtension)calloc(1, sizeof(UnitExtensionRec)) ;
}

static void	deinitUnit(unit)
  Unit	unit ;
{
  if (unit->extension != NULL)
    free(unit->extension) ;
}

/***********************************************************************/
static void	initLink(link)
  Link	link ;
{
  link->extension = (LinkExtension)calloc(1, sizeof(LinkExtensionRec)) ;

  setLinkLength(link, 1.0) ;
}

static void	deinitLink(link)
  Link	link ;
{
  if (link->extension != NULL)
    free(link->extension) ;
}

/***********************************************************************
 *	Name:		initExampleSet
 ***********************************************************************/
static void	initExampleSet(exampleSet)
  ExampleSet  exampleSet ;
{
  exampleSet->extension = 
    (ExampleSetExtension)calloc(1, sizeof(ExampleSetExtensionRec)) ;
}

static void	deinitExampleSet(exampleSet)
  ExampleSet  exampleSet ;
{
  if (exampleSet->extension != NULL)
    free(exampleSet->extension) ;
}

void	*getExampleSetData(exampleSet)
  ExampleSet  exampleSet ;
{
  return exampleSet->extension->data ;
}

void	putExampleSetData(exampleSet, data)
  ExampleSet  exampleSet ;
  void	*data ;
{
 exampleSet->extension->data = data ;
}

/*********************************************************************
 * 				Link Length
 *********************************************************************/
void		setLinkLength(link, length)
  Link		link ;
  Real		length ;
{
  Mlength(link) = length ;
}

Real		getLinkLength(link)
  Link		link ;
{
  return Mlength(link) ;
}


/*********************************************************************
 * 				Lesioning
 *********************************************************************/
Boolean		isLesionedLink(link)
  Link		link ;
{
  return Mlesioned(link) ;
}

void		lesionLink(link)
  Link		link ;
{
  link->deriv = 0.0 ;
  MlesionedWeight(link) = link->weight ;
  link->weight = 0.0 ;
  Mlesioned(link) = TRUE ;
  Mcompiled(link->postUnit->net) = FALSE ;
}

void		linkLesion(link, data)
  Link		link ;
  void		*data ;
{
  lesionLink(link) ;
}

void		unlesionLink(link)
  Link		link ;
{
  if (Mlesioned(link)) link->weight = MlesionedWeight(link) ;
  Mlesioned(link) = FALSE ;
  Mcompiled(link->postUnit->net) = FALSE ;
}

void		linkUnlesion(link, data)
  Link		link ;
  void		*data ;
{
  unlesionLink(link) ;
}



Boolean		isLesionedUnit(unit)
  Unit		unit ;
{
  return Mlesioned(unit) ;
}

void		lesionUnit(unit)
  Unit		unit ;
{
  setNaN(unit->output) ;
  Mlesioned(unit) = TRUE ;
  unitForAllLinks(unit, ALL, (LinkDataProc)linkLesion, NULL) ;
  Mcompiled(unit->net) = FALSE ;
}

void		unlesionUnit(unit)
  Unit		unit ;
{
  Mlesioned(unit) = FALSE ;
  unit->output = 0.0 ;
  unitForAllLinks(unit, ALL, (LinkDataProc)linkUnlesion, NULL) ;
  Mcompiled(unit->net) = FALSE ;
}

/***********************************************************************
 *	Name:		unitError
 ***********************************************************************/
static Real	unitError(unit)
  const Unit	unit ;
{
  Net	net = unit->net ;
  Real	error, output = unit->output, target = unit->target ;

  if (isNaN(target) || fabs(output-target) <= MzeroErrorRadius(net))
    return 0.0 ;

  if (MnonExtremalTarget(net) > 0.0) 
    if (target == 0.0) {
      if (output <= MnonExtremalTarget(net)) return 0.0 ;
      else target = MnonExtremalTarget(net) ;
    } else {
      if (output >= 1.0-MnonExtremalTarget(net)) return 0.0 ;
      else target = 1.0-MnonExtremalTarget(net) ;
    }
  if (fabs(output-target) > McorrectTolerance(net)) MoutputsIncorrect(net)++ ;

/*  if (Mnormalize(unit->group)) {
      output = CLIP(output, MIN_UNIT_OUTPUT, MAX_UNIT_OUTPUT) ;
      error = DIVERGE(output, target);
  } else
*/
  switch (MerrorFunction(net)) {
  case CROSS_ENTROPY:
    output = CLIP(output, MIN_UNIT_OUTPUT, MAX_UNIT_OUTPUT) ;
    error = INFO(output, target) ;
    break ;
  case SQUARED_ERROR:
    error = (output - target) * (output - target) ;
    break ;
  default:
    IErrorAbort("Unknown error function: %d", MerrorFunction(net)) ;
  }      
  if (!MsampleExamples(net)) error *= MexampleFrequency(net) ;
  if (MnormError(net))       error /= (Real)MmaxSlice(net) ;
  return error ;
}

/**********************************************************************/
static Real	unitErrorDeriv(unit)
  const Unit	unit ;
{
  Net	net = unit->net ;
  Real	error, output = unit->output, target = unit->target ;

  if (isNaN(target) || fabs(output-target) <= MzeroErrorRadius(net))
    return 0.0 ;

  if (MnonExtremalTarget(net) > 0.0) 
    if (target == 0.0) {
      if (output <= MnonExtremalTarget(net)) return 0.0 ;
      else target = MnonExtremalTarget(net) ;
    } else {
      if (output >= 1.0-MnonExtremalTarget(net)) return 0.0 ;
      else target = 1.0-MnonExtremalTarget(net) ;
    }

/*  if (Mnormalize(unit->group) && (unit->group->type&OUTPUT)) 
      error = DIVERGE_DERIV(output, target);
    else
*/
  switch (MerrorFunction(net)) {
  case CROSS_ENTROPY:
    output = CLIP(output, MIN_UNIT_OUTPUT, MAX_UNIT_OUTPUT) ;
    error = INFO_DERIV(output, target) ; 
    break ;
  case SQUARED_ERROR:
    error = 2.0 * (output - target) ;
    break ;
  default:
    IErrorAbort("Unknown error function: %d", MerrorFunction(net)) ;
  }      

  if (!MsampleExamples(net)) error *= MexampleFrequency(net) ;
  if (MnormError(net))       error /= (Real)MmaxSlice(net) ;
  return error ;
}

/**********************************************************************/
void	incrementError(unit, data)
  Unit		unit ;
  void		*data ;
{
  *(Real *)data += unitError(unit) ;
}

Real		groupError(group)
  Group group ;
{
  Real	error = 0 ;
  groupForAllUnits(group, (UnitDataProc)incrementError, &error) ;
  return error ;
}

#define PRINT_OUTPUT	0
#define PRINT_DERIV	1
/**********************************************************************/
static void	unitPrintField(unit, data)
  Unit		unit ;
  void		*data ;
{
  Real	value ;
  switch ((int)data) {
  case PRINT_OUTPUT: value = unit->output ; break ;
  case PRINT_DERIV:  value = unit->outputDeriv ; break ;
  }
  fprintf(dout, "%f ", value) ;
}
static Real	groupPrintField(group, field)
  Group group ;
  int	field ;
{
  fprintf(dout, "%s ", group->name) ;
  switch (field) {
  case PRINT_OUTPUT: fprintf(dout, "output: ") ; break ;
  case PRINT_DERIV:  fprintf(dout, "deriv: ") ; break ;
  }
  groupForAllUnits(group, (UnitDataProc)unitPrintField, (void *)field) ;
  fprintf(dout,"\n") ;
}

/**********************************************************************/
void	zeroOutputDeriv(unit, data)
  Unit		unit ;
  void		*data ;
{
  unit->outputDeriv = 0.0 ;
}

/**********************************************************************/
void	injectError(unit, data)
  Unit		unit ;
  void		*data ;
{
  unit->outputDeriv = unitErrorDeriv(unit) ;
}

/**********************************************************************/
static void	incrementCost(unit, data)
  Unit		unit ;
  void		*data ;
{
  Real	output ;
  Real	ratio ;
  if (MnoUnitCost(unit->group)) return ;
  if (MunitPrior(unit->net) <= 0.0)
    /*
     * unbiased binary
     */
    *(Real *)data += MunitCost(unit->net) * 
                     unit->output * (1.0 - unit->output) ;
  else if (MunitPrior(unit->net) > 1.0)
    /*
     * biased binary
     */
    *(Real *)data += MunitCost(unit->net) * 
                     unit->output * (MunitPrior(unit->net) - unit->output) ;
  else {
    /* 
     * weight-elimination-based measure
     */
    ratio = (unit->output*unit->output)/(MunitPrior(unit->net)*MunitPrior(unit->net)) ;
    *(Real *)data += MunitCost(unit->net) * 
                     ratio / (1.0+ratio) ;
    /*
     * CE-based measure
     * output = CLIP(unit->output, MIN_UNIT_OUTPUT, MAX_UNIT_OUTPUT) ;
     * *(Real *)data += MunitCost(unit->net) * 
     *                  INFO(output,MunitPrior(unit->net)) ;
     */
  }
}

/**********************************************************************/
void	injectUnitCost(unit, data)
  Unit		unit ;
  void		*data ;
{
  Real	output ;
  Real	ratio ;
  if (MnoUnitCost(unit->group)) return ;
  if (MunitPrior(unit->net) <= 0.0)
    unit->outputDeriv += MunitCost(unit->net) * (1.0 - 2.0 * unit->output) ;
  else if (MunitPrior(unit->net) > 1.0)
    /*
     * biased binary
     */
    unit->outputDeriv += MunitCost(unit->net) * (MunitPrior(unit->net) - 2.0 * unit->output) ;
  else {
    /* 
     * weight-elimination-based measure
     */
    ratio = (unit->output*unit->output)/(MunitPrior(unit->net)*MunitPrior(unit->net)) ;
    unit->outputDeriv += MunitCost(unit->net) * 
                         2.0*unit->output/((1.0+ratio)*(1.0+ratio)) ;
    /*
     * CE-based measure
     * output = CLIP(unit->output, MIN_UNIT_OUTPUT, MAX_UNIT_OUTPUT) ;
     * unit->outputDeriv += MunitCost(unit->net) * 
     *                      INFO_DERIV(output,MunitPrior(unit->net)) ;
     */
  }
}

/**********************************************************************/
static void	zeroLinkDerivs(unit, data)
  Unit	unit ;
  void	*data ;
{
  Link	*linkPtr  = unit->incomingLink ;
  Link	*finalPtr = linkPtr + unit->numIncoming ;

  while (linkPtr != finalPtr) {
    Link	link = *(linkPtr++) ;
    link->deriv = 0.0 ;
  }
}

/**********************************************************************/
static void	lengthScaleDerivs(unit, data)
  Unit	unit ;
  void	*data ;
{
  Link	*linkPtr  = unit->incomingLink ;
  Link	*finalPtr = linkPtr + unit->numIncoming ;

  while (linkPtr != finalPtr) {
    Link	link   = *(linkPtr++) ;
    Real	length = getLinkLength(link) ;
    if (length > 0.0) link->deriv /= length ;
  }
}

/**********************************************************************/
static void	netAdaptWeightCost(net)
  Net		net ;
{
  MaveError(net) = (MaveError(net)<0.0 ? net->error :
		    net->error + MaveErrorFactor(net)*(MaveError(net)-net->error)) ;
  if (net->error < MlastError(net)	||
      net->error < MdesiredError(net)	||
      MlastError(net) < 0.0) 
    MweightCost(net) += MweightCostIncrement(net) ;
  else if (net->error < MaveError(net))
    MweightCost(net) -= MweightCostIncrement(net) ;
  else
    MweightCost(net) *= MweightCostDecrement(net) ;
  if (MweightCost(net) < 0.0) MweightCost(net) = 0.0 ;
  MlastError(net) = net->error ;
}

/**********************************************************************/
static void	unitDecayWts(unit, data)
  Unit		unit ;
  void		*data ;
{
  int	idx ;
  Net	net 		= unit->net ;
  int	numIncoming 	= unit->numIncoming ;
  Link	*incoming   	= unit->incomingLink ;
  Real	weightCost	= MweightCost(net) ;
  Real	wtCost ;
  Boolean lengthScaleCost = (! MlengthScaleDerivs(net)) ;

  if (MnoWeightCost(unit->group)) return ;

  for (idx = 0 ; idx < numIncoming ; ++idx) {
    Link link    = incoming[idx] ;
    if (MnoWeightCost(link->preUnit->group)) continue ;
    wtCost  = link->weight * weightCost ;
    if (lengthScaleCost) wtCost *= getLinkLength(link) ;
    net->cost   += wtCost * link->weight ;
    link->deriv += 2.0 * wtCost ;
  }
}

/**********************************************************************/
static void	unitMeltWts(unit, data)
  Unit		unit ;
  void		*data ;
{
  int	idx ;
  Net	net 		= unit->net ;
  int	numIncoming 	= unit->numIncoming ;
  Link	*incoming   	= unit->incomingLink ;
  Real	weightCost	= MweightCost(net) ;
  Real	Wo	   	= MstandardWeight(net) ;
  Real	wt, ratio ;
  Boolean lengthScaleCost = (! MlengthScaleDerivs(net)) ;

  if (MnoWeightCost(unit->group)) return ;

  for (idx = 0 ; idx < numIncoming ; ++idx) {
    Link link    = incoming[idx] ;
    if (MnoWeightCost(link->preUnit->group)) continue ;
    wt	 = link->weight ;
    ratio   = (wt * wt) / (Wo * Wo) ;
    if (lengthScaleCost) weightCost *= getLinkLength(link) ;
    net->cost   += weightCost * ratio / (1.0 + ratio) ;
    link->deriv += weightCost * 2.0 * wt / ((1.0 + ratio) * (1.0 + ratio)) ;
  }
}

/**********************************************************************/
static 	void	netUpdateWeightCost(net)
  Net	net ;
{
  if (MweightCost(net) == 0.0) return ;
  if (MstandardWeight(net) == 0.0) 
    netForAllUnits(net, ALL, (UnitDataProc)unitDecayWts, NULL) ;
  else
    netForAllUnits(net, ALL, (UnitDataProc)unitMeltWts, NULL) ;
}

/**********************************************************************/
static void	unitCompile(unit, data)
  Unit		unit ;
  void		*data ;
{
  int	idx, cdx, numIncoming = unit->numIncoming ;

  if (Mlesioned(unit)) {
    MnumCompiledIncoming(unit) = 0 ;
    return ;
  }
  if (!MpreOutputs(unit)) {
    MpreOutputs(unit)	   = ((Real **)calloc(numIncoming, sizeof(Real*))) ;
    MpreOutputDerivs(unit) = ((Real **)calloc(numIncoming, sizeof(Real*))) ;
    Mweights(unit)    	   = ((Real *) calloc(numIncoming, sizeof(Real))) ;
    Mderivs(unit)     	   = ((Real *) calloc(numIncoming, sizeof(Real))) ;
    McompiledLinks(unit)   = ((Link *) calloc(numIncoming, sizeof(Link))) ;
  }

  for (idx=0, cdx=0 ; idx < numIncoming ; idx++) {
    Link link = unit->incomingLink[idx] ;
    if (! (Mlesioned(link) || Mlesioned(link->preUnit))) {
      MpreOutputs(unit)[cdx] 	  = &(link->preUnit->output) ;
      MpreOutputDerivs(unit)[cdx] = &(link->preUnit->outputDeriv) ;
      Mweights(unit)[cdx]	  =   link->weight ;
      McompiledLinks(unit)[cdx]   =   link ;
      cdx++ ;
    }
  }
  MnumCompiledIncoming(unit) = cdx ; 
}

/***********************************************************************/
static void	netCompile(net)
  Net		net ;
{
  netForAllOtherUnits(net, INPUT|BIAS, (UnitDataProc)unitCompile, NULL) ;
  Mcompiled(net) = TRUE ;
}

/**********************************************************************/
static void	linkSumWt(link, data)
  Link		link ;
  void		*data ;
{
  if (link->preUnit->group->type&BIAS) return ;
  *(Real *)data += link->weight ;
}

/**********************************************************************/
static void	linkSetWt(link, data)
  Link		link ;
  void		*data ;
{
  if (link->preUnit->group->type&BIAS) return ;
  link->weight = *(Real *)data ;
}

/**********************************************************************/
static void	linkMinWt(link, data)
  Link		link ;
  void		*data ;
{
  if (link->preUnit->group->type&BIAS) return ;
  link->weight = MAX(link->weight, *(Real *)data) ;
}

/**********************************************************************/
static void	linkMaxWt(link, data)
  Link		link ;
  void		*data ;
{
  if (link->preUnit->group->type&BIAS) return ;
  link->weight = MIN(link->weight, *(Real *)data) ;

}
/**********************************************************************/
static void	unitConstrainWeights(unit, data)
  Unit	unit ;
  void	*data ;
{
  /* CONSTRAIN BIAS */
  if (unit->numIncoming > 0) {
    if (MmaxBias(unit->group)<MAXFLOAT)
      unit->incomingLink[0]->weight = 
	MIN(unit->incomingLink[0]->weight,MmaxBias(unit->group)) ;
    if (MminBias(unit->group)>-MAXFLOAT)
      unit->incomingLink[0]->weight = 
	MAX(unit->incomingLink[0]->weight,MminBias(unit->group)) ;
  }

  /* INCOMING WEIGHTS */
  if (unit->numIncoming > 1) {
    if (MequateIncomingWeights(unit)) {
      Real	weight = 0.0 ;
      unitForAllLinks(unit, INCOMING, (LinkDataProc)linkSumWt, (void *)&weight) ;
      weight /= (Real)(unit->numIncoming-1) ;
      unitForAllLinks(unit, INCOMING, (LinkDataProc)linkSetWt, (void *)&weight) ;
    }
    if (MmaxIncomingWeight(unit->group)<MAXFLOAT)
      unitForAllLinks(unit, INCOMING, (LinkDataProc)linkMaxWt, (void *)&MmaxIncomingWeight(unit->group)) ;
    if (MminIncomingWeight(unit->group)>-MAXFLOAT)
      unitForAllLinks(unit, INCOMING, (LinkDataProc)linkMinWt, (void *)&MminIncomingWeight(unit->group)) ;
  }

  /* OUTGOING WEIGHTS */
  if (unit->numOutgoing > 0) {
    if (MequateOutgoingWeights(unit)) {
      Real	weight = 0.0 ;
      unitForAllLinks(unit, OUTGOING, (LinkDataProc)linkSumWt, (void *)&weight) ;
      weight /= (Real)(unit->numOutgoing) ;
      unitForAllLinks(unit, OUTGOING, (LinkDataProc)linkSetWt, (void *)&weight) ;
    }
    if (MmaxOutgoingWeight(unit->group)<MAXFLOAT)
      unitForAllLinks(unit, OUTGOING, (LinkDataProc)linkMaxWt,(void *)&MmaxOutgoingWeight(unit->group)) ;
    if (MminOutgoingWeight(unit->group)>-MAXFLOAT)
      unitForAllLinks(unit, OUTGOING, (LinkDataProc)linkMinWt,(void *)&MminOutgoingWeight(unit->group)) ;
  }
}

/***********************************************************************/
int	command_applyWeightConstraints(tokc, tokv)
  int	tokc ;
  char	**tokv ;
{

  IUsage("");
  if (GiveHelp(tokc)) {
    ISynopsis("Apply constraints on weights and biases.");
    IHelp
      (IHelpArgs,
       "Apply constraints on weights and biases.",
       "EXAMPLES",
       "SEE ALSO",
       NULL) ;
    return 1 ;
  }

  if (currentNet == NULL)
    IErrorAbort("There is no current net.") ;

  netForAllUnits(currentNet, ALL, (UnitDataProc)unitConstrainWeights, NULL) ;

  return 1 ;
}

/**********************************************************************/
static void	getCompiledWeights(unit, data)
  Unit	unit ;
  void	*data ;
{
  int	idx, cdx ;
  int	numIncoming	= MnumCompiledIncoming(unit) ;
  Real	*wt, *dv;
  Link	*ln ;

  for (idx=0, wt=Mweights(unit), dv=Mderivs(unit), ln=McompiledLinks(unit) ;
       idx < numIncoming ; 
       idx++) {
    (*wt++) = (*ln++)->weight ;
    (*dv++) = 0.0 ;
  }
}

/**********************************************************************/
static void	putCompiledDerivs(unit, data)
  Unit	unit ;
  void	*data ;
{
  int	idx ;
  int	numIncoming = MnumCompiledIncoming(unit) ;
  Real	*dv ;
  Link	*ln ;

  for (idx=0, dv=Mderivs(unit), ln=McompiledLinks(unit) ;
       idx < numIncoming ; 
       idx++) 
    (*ln++)->deriv  = (*dv++) ;
}
/***********************************************************************
 *	Name:		command_compile
 ***********************************************************************/
int	command_compile(tokc, tokv)
  int	tokc ;
  char	**tokv ;
{

  IUsage("");
  if (GiveHelp(tokc)) {
    ISynopsis("Compiles network");
    IHelp
      (IHelpArgs,
       "Compiles network.",
       "EXAMPLES",
       "SEE ALSO",
       NULL) ;
    return 1 ;
  }

  if (currentNet == NULL)
    IErrorAbort("There is no current net.") ;

  netCompile(currentNet) ;

  return 1 ;
}

/***********************************************************************/
static	void	setWeight(link, data)
  Link	link ;
  void	*data ;
{
  link->weight = *(Real *)data ;
}

/**********************************************************************/
static void	groupSetInhibWeight(group, data)
  Group		group ;
  void		*data ;
{
  if (group->numUnits == 1 &&
      group->unit[0]->numIncoming > 1 &&
      MequateOutgoingWeights(group->unit[0]) &&
      MmaxOutgoingWeight(group) <= 0.0) {
    Unit	unit = group->unit[0] ;
    Real	bias = unit->incomingLink[0]->weight ;

    unitForAllLinks(unit, OUTGOING, (LinkDataProc)setWeight, (void *)&bias) ;
  }
}

/***********************************************************************/
int	command_setInhibWeights(tokc, tokv)
  int	tokc ;
  char	**tokv ;
{

  IUsage("");
  if (GiveHelp(tokc)) {
    ISynopsis("Sets outgoing weights of dedicated inhibitory units to their bias.") ;
    IHelp
      (IHelpArgs,
       "Sets outgoing weights of dedicated inhibitory units to their bias." ,
       "EXAMPLES",
       "SEE ALSO",
       NULL) ;
    return 1 ;
  }

  if (currentNet == NULL)
    IErrorAbort("There is no current net.") ;

  netForAllOtherGroups(currentNet, BIAS|INPUT, (GroupDataProc)groupSetInhibWeight, NULL) ;

  return 1 ;
}



/**********************************************************************
 *			netDoGroups/netDoUnits
 **********************************************************************/
static void	netDoGroups (net, first, last, mask, test, proc, data)
  Net		net ;
  int		first, last ;
  Mask		mask ;
  Boolean	test ;
  GroupDataProc	proc ;
  void		*data ;
{
  Group *groupPtr = net->group + first ;
  int	groupInc  = (first < last ? 1 : -1) ;
  Group *finalPtr = net->group + last + groupInc ;

  while (groupPtr != finalPtr) {
    Group	group = (*groupPtr) ;
    groupPtr += groupInc ;
    if (( test &&  (group->type & mask)) ||
	(!test && !(group->type & mask))) 
      (*proc) (group, data) ;
  }
}

/**********************************************************************/
void	netDoUnits (net, first, last, mask, test, proc, data)
  Net		net ;
  int		first, last ;
  Mask		mask ;
  Boolean	test ;
  UnitDataProc	proc ;
  void		*data ;
{
  Group *groupPtr = net->group + first ;
  int	groupInc  = (first < last ? 1 : -1) ;
  Group *finalPtr = net->group + last + groupInc ;

  while (groupPtr != finalPtr) {
    Group	group = (*groupPtr) ;
    groupPtr += groupInc ;
    if (( test &&  (group->type & mask)) ||
	(!test && !(group->type & mask))) {
      groupForAllUnits(group, proc, data) ;
    }
  }
}

/***********************************************************************
 *				netForward
 ***********************************************************************/
void	addInputNoise(unit, data)
  Unit	unit ;
  void	*data ;
{
  Real	noise = gaussianNoise(*(Real *)data) ;

  if      (unit->output == 0.0) unit->output += fabs(noise) ;
  else if (unit->output == 1.0) unit->output -= fabs(noise) ;
  else unit->output = CLIP(unit->output+noise, 0.0, 1.0) ;
}

void	scaleInput(unit, data)
  Unit	unit ;
  void	*data ;
{
  unit->output *= *(Real *)data ;
}

void	forwardTargExtInput(unit, data)
  Unit	unit ;
  void	*data ;
{
  Net	net	= unit->net ;
  Real	freq, factor, extIn ;
  int	numIncoming 	= MnumCompiledIncoming(unit) ;
  Real	totalInput  	= 0.0 ;
  Real	*wt, **preOutput ;
  int	idx ;

  if (Mlesioned(unit)) return ;
  if (unit->group->type&OUTPUT && !isNaN(unit->target)) {
    if (MtargExtInConstant(net)==0.0) {
      extIn = *(Real *)data ;
    } else {	
      switch (MfreqFunction(net)) {
      case 0: 	/* raw */
	freq = MexampleFrequency(net)*69973 ; 
	factor = net->mz->nIterations*LOG(freq) ;
	break ;
      case 1:		/* sqrt */
	freq = MexampleFrequency(net)*264.524 ;
	factor = net->mz->nIterations*LOG(freq*freq) ;
	break ;
      case 2:		/* log */
	factor = net->mz->nIterations*MexampleFrequency(net) ;
	break ;
      default:
	IErrorAbort("Unknown freqFunction %d", MfreqFunction(net)) ;
      }
      extIn = *(Real *)data * factor/(factor+MtargExtInConstant(net)) ;
    }
    totalInput += (unit->target > 0.5 ? extIn : -extIn) ;
  }
  for (idx=0, wt=Mweights(unit), preOutput=MpreOutputs(unit) ;
       idx < numIncoming ; 
       idx++) 
    totalInput += (*wt++) * *(*preOutput++) ;
  unit->output = sigmoid(totalInput) ;
}

/**********************************************************************/
void	PMSPtargExtInRange (net)
  Net		net ;
{
  Real factor ;
  factor = net->mz->nIterations*11.1559 ;	/* log(69973) */
  MtargExtInHi(net) = MtargetExternalInput(net) * factor/(factor+MtargExtInConstant(net)) ;
  factor = net->mz->nIterations*0.693147 ;	/* log(2) */
  MtargExtInLo(net) = MtargetExternalInput(net) * factor/(factor+MtargExtInConstant(net)) ;
}

/***********************************************************************/
/* DCP: new */
void	maybeCopyTargetsFromLayer(toGroup, slice)
  Group	toGroup ;
  void	*slice ;
{
  int i ;

  if (McopyTargetsFromLayer(toGroup)) {
    Net net = toGroup->net ;
    Group fromGroup = net->group[McopyTargetsFromLayer(toGroup)] ;
    if (toGroup->numUnits != fromGroup->numUnits)
      IErrorAbort("maybeCopyTargetsFromLayer: toGroup (%d) and fromGroup (%d) numUnits differ!", 
		  toGroup->numUnits, fromGroup->numUnits) ;
    for (i = 0 ; i < toGroup->numUnits ; i++) 
      if (!isNaN(toGroup->unit[i]->target)) {
	Real output = fromGroup->unit[i]->output ;
	toGroup->unit[i]->target = output ;
	if (net->type&RECURRENT) toGroup->unit[i]->targetHistory[(int)slice] = output ;
      }
  }
}

/**********************************************************************/
static void	netForward (net)
  Net		net ;
{
  if (!MoutcomeProc(net))  setOutcomeProc(net);
  
  if( isFM(net) ) {
    MexampleNum(net) = net->trainingExampleSet->currentIdx;
    netForwardFM( net, net->trainingExampleSet );		/* forward model */
    netForAllOtherGroups( net, BIAS, (GroupDataProc)setNormalizedOff, NULL );
    return;
    /* DCP: not forward model but want to calculate acoustic input */
    /* NOTE: this only works for TRAINING example set */
  } else if (MgoalLayer(net)) {
    MexampleNum(net) = net->trainingExampleSet->currentIdx;
    /* is now done in netIterate for recurrent 
       computeAcousticGoals( net, &(net->trainingExampleSet->example[MexampleNum(net)]) );  */
  }
  
  if (MinputScaleFactor(net)>0.0)
    netForAllUnits (net, INPUT, (UnitDataProc)scaleInput, &MinputScaleFactor(net)) ;
  if (MinputNoise(net)>0.0)
    netForAllUnits (net, INPUT, (UnitDataProc)addInputNoise, &MinputNoise(net)) ;
  if (MtargetExternalInput(net) > 0.0) {
    netForAllOtherUnits (net, INPUT|BIAS, (UnitDataProc)forwardTargExtInput, &MtargetExternalInput(net)) ;
    PMSPtargExtInRange(net) ;
  } else {
    netForAllOtherUnits (net, INPUT|BIAS, (UnitDataProc)forward, NULL) ;
  }
  MoutputsIncorrect(net) = 0 ;

  /* DCP */
  netForAllGroups(net, OUTPUT, (GroupDataProc)maybeCopyTargetsFromLayer, (void *)net->currentTime) ;
  netForAllUnits     (net, OUTPUT, (UnitDataProc)incrementError, &(net->error)) ;
  if (MunitCost(net) != 0.0)
    netForAllOtherUnits(net, INPUT|OUTPUT|BIAS, (UnitDataProc)incrementCost, &(net->cost)) ;
    
  netForAllOtherGroups( net, BIAS, (GroupDataProc)setNormalizedOff, NULL );
}

/**********************************************************************/
extern Real	sumTotalInput( unit )
  Unit unit;
{
  Real 	totalInput, *wt, **preOut;
  int	numIncoming, k;
  
  numIncoming = MnumCompiledIncoming(unit) ;
  totalInput  = (MunitNoise(unit->net) ? gaussianNoise(MunitNoise(unit->net)) : 0.0);
  
  for (k=0, wt=Mweights(unit), preOut=MpreOutputs(unit); k < numIncoming ; k++) 
    totalInput += (*wt++) * *(*preOut++) ;
    
  return totalInput;
}

/**********************************************************************/
void	forward(unit, data)
  Unit	unit ;
  void	*data ;
{
  if (Mlesioned(unit)) return ;
    
  if( Mnormalize( unit->group ) ) {
    if( MalreadyNormed( unit->group ) ) return;
    
    if( MgaussianSize( unit->group ) )	
      forwardArticNormalized( unit->net, unit->group );
    else
      forwardNormalized( unit->group->unit, unit->group->numUnits );
    MalreadyNormed( unit->group ) = TRUE;
    return;
  }
  unit->totalInput = sumTotalInput( unit ) ;
  unit->output = sigmoid( unit->totalInput ) ;
}

/**********************************************************************/
static void	saveTarget(unit, data)
  Unit	unit ;
  void	*data ;
{
  Msaved(unit) = (unit->net->type&RECURRENT ? unit->targetHistory[(int)data] : unit->target) ;
}
/**********************************************************************/
static void	instateActualOutcome(unit, data)
  Unit	unit ;
  void	*data ;
{
  Real	tmp = Msaved(unit) ;	/* save goal */
  Msaved(unit) = unit->output ;	/* predicted -> saved  */
  unit->output = unit->target ;	/* actual    -> output */
  unit->target = tmp ;		/* goal      -> target */
  if (unit->net->type&RECURRENT) unit->targetHistory[(int)data] = tmp ;
}
/**********************************************************************/
static void	instatePredictedOutcome(unit, data)
  Unit	unit ;
  void	*data ;
{
  Real	tmp = unit->output ;	/* save actual outcome */
  unit->output = Msaved(unit) ;	/* predicted -> output */ 
  Msaved(unit) = unit->target ;	/* goal      -> saved  */
  unit->target = tmp ;		/* actual    -> target */
  if (unit->net->type&RECURRENT) unit->targetHistory[(int)data] = tmp ;
}
/********************************************************************/
extern void	copyOutputsToOutputs (net, fromGroupIdx, toGroupIdx)
  Net	net ;
  int	fromGroupIdx, toGroupIdx ;
{
  int idx, s ;
  Group from = net->group[fromGroupIdx] ;
  Group to   = net->group[toGroupIdx] ;
  
  if (from->numUnits != to->numUnits)
    IErrorAbort ("numUnits for from group (%d) and to group (%d) unequal!",
		 from->numUnits, to->numUnits) ;
  for (idx=0; idx < from->numUnits; idx++) 
    to->unit[idx]->output = from->unit[idx]->output ;
}
/********************************************************************/
static void	copyOutputHistoryToOutputs (net, fromGroupIdx, toGroupIdx, slice)
  Net	net ;
  int	fromGroupIdx, toGroupIdx ;
{
  int idx, s ;
  Group from = net->group[fromGroupIdx] ;
  Group to   = net->group[toGroupIdx] ;
  
  if (from->numUnits != to->numUnits)
    IErrorAbort ("numUnits for from group (%d) and to group (%d) unequal!",
		 from->numUnits, to->numUnits) ;
  for (idx=0; idx < from->numUnits; idx++) 
    to->unit[idx]->output = from->unit[idx]->outputHistory[slice] ;
}
/********************************************************************/
static void	copyOutputsToTargets (net, fromGroupIdx, toGroupIdx, copySlices)
  Net	net ;
  int	fromGroupIdx, toGroupIdx ;
  Boolean copySlices ;
{
  int idx, s ;
  Group from = net->group[fromGroupIdx] ;
  Group to   = net->group[toGroupIdx] ;
  
  if (from->numUnits != to->numUnits)
    IErrorAbort ("numUnits for from group (%d) and to group (%d) unequal!",
		 from->numUnits, to->numUnits) ;
  for (idx=0; idx < from->numUnits; idx++) {
    /* 
     * NOTE: targets are copied slice by slice
     *       could have copied final output to target for all slices 
     */
    if (copySlices && net->type&RECURRENT)
      for (s = 0 ; s < MmaxSlice(net) ; s++)
	to->unit[idx]->targetHistory[s] = from->unit[idx]->outputHistory[s] ;
    else
      to->unit[idx]->target = from->unit[idx]->output ;
  }
}
/********************************************************************/
static void	copyOutputDerivsToOutputDerivs (net, fromGroupIdx, toGroupIdx)
  Net	net ;
  int	fromGroupIdx, toGroupIdx ;
{
  int idx, s ;
  Group from = net->group[fromGroupIdx] ;
  Group to   = net->group[toGroupIdx] ;
  
  if (from->numUnits != to->numUnits)
    IErrorAbort ("numUnits for from group (%d) and to group (%d) unequal!",
		 from->numUnits, to->numUnits) ;
  for (idx=0; idx < from->numUnits; idx++) 
    to->unit[idx]->outputDeriv = from->unit[idx]->outputDeriv ;
}

/**********************************************************************/
static void	netTrainFM (net, Acoustics, Articulation)
  Net		net ;
  int		Acoustics, Articulation ;
{
  netDoUnits(net, Acoustics, Articulation+1, ALL, TRUE,
	     (UnitDataProc)zeroOutputDeriv, NULL) ;
  /* reinstate predicted outcome as outputs and actual outcome as targets */
  netDoUnits(net, Acoustics, Acoustics, ALL, TRUE,
	     (UnitDataProc)instatePredictedOutcome, NULL) ;
  groupForAllUnits(net->group[Acoustics], (UnitDataProc)injectError, NULL) ;
  netDoUnits(net, Acoustics, Articulation+1, INPUT|BIAS, FALSE,
	     (UnitDataProc)backward, NULL) ;
  /* reinstate actual outcome as outputs and goal as targets */
  netDoUnits(net, Acoustics, Acoustics, ALL, TRUE,
	     (UnitDataProc)instateActualOutcome, NULL) ;
}

/**********************************************************************/
static void	groupForward(group, slice)
  Group		group ;
  void		*slice ;
{
  Net	net = group->net ;
  int	idx ;

  if( Mnormalize( group ) ) {
    if( MgaussianSize( group ) ) forwardArticNormalized( net, group );
    else			 forwardNormalized( group->unit, group->numUnits );
    MalreadyNormed( group ) = TRUE;
  } else
    groupForAllUnits(group, (UnitDataProc)forward, NULL) ;

  if (MactionLayer(net)) {
    /* copy computed Articulation onto CArticulation for first slice */
    if (slice==0 && !strcmp(net->group[MactionLayer(net)]->name, group->name))
      if ((idx=copiedFromLayer(net, MactionLayer(net))))
	copyOutputsToOutputs(net, MactionLayer(net), idx) ;
    if (!strcmp(net->group[MoutcomeLayer(net)]->name, group->name)) {
      /* save (possible) goal targets */
      groupForAllUnits(group, (UnitDataProc)saveTarget, slice) ;
      /* compute actual outcome as targets */
      /* NOTE: this does not save resulting targets in current example */
      FMcomputeActualOutcome(net) ;
      /* compute prediction error */
      groupForAllUnits(group, (UnitDataProc)incrementError, &MpredictionError(net)) ;

      /* save predicted outcome as activation over Goal layer (if any) */
      if (MgoalLayer(net)) {
	if (net->group[MgoalLayer(net)]->numUnits != group->numUnits)
	  IErrorAbort("goal layer \"%s\" (%d) and outcome layer \"%s\" (%d) must be same size",
		      net->group[MgoalLayer(net)]->name, net->group[MgoalLayer(net)]->numUnits,
		      group->name, group->numUnits) ;
	for(idx = 0; idx < group->numUnits; idx++) 
	  net->group[MgoalLayer(net)]->unit[idx]->output = group->unit[idx]->output ;
      }

      /* instate actual outcome as outputs, reinstate goal as targets, and */
      /* save predicted outcome as Msaved(unit) (in case FM is to be trained) */
      groupForAllUnits(group, (UnitDataProc)instateActualOutcome, slice) ;
    }
  }
  maybeCopyTargetsFromLayer(group, slice) ;
}

/***********************************************************************
 *				netBackward
 ***********************************************************************/
static void	netBackward (net)
  Net		net ;
{
  if( isFM(net) ) {
    netBackwardFM(net);			/* forward model */
    return;
  }
  
  netForAllOtherUnits(net, INPUT|OUTPUT, (UnitDataProc)zeroOutputDeriv,NULL) ;
  if (MunitCost(net) != 0.0)
    netForAllOtherUnits(net, INPUT|OUTPUT|BIAS,  (UnitDataProc)injectUnitCost, NULL) ;
  if (MbpToOutputs(net)) {
    netForAllUnits          (net, OUTPUT,     (UnitDataProc)injectError, NULL) ;
    netForAllOtherUnitsBack (net, INPUT|BIAS, (UnitDataProc)backward,	 NULL) ;
  } else
    netForAllOtherUnitsBack (net, INPUT|BIAS, (UnitDataProc)backwardRecomputeOutputError, NULL) ;
}

/**********************************************************************/
void	backward (unit, data)
  Unit	unit ;
  void	*data ;
{
  int	numIncoming = MnumCompiledIncoming(unit) ;
  Real	inputDeriv ;
  Real 	*wt, *dv, **preOut, **preDv ;
  int	idx ;

  if (Mlesioned(unit)) return ;
  
  inputDeriv = sigmoidDeriv(unit->output) * unit->outputDeriv ;

  for (idx=0, wt=Mweights(unit), dv=Mderivs(unit), preOut=MpreOutputs(unit),
       preDv=MpreOutputDerivs(unit) ;
       idx < numIncoming ; 
       idx++) {
    *(*preDv++) += inputDeriv *  (*wt++) ;
     (*dv++)    += inputDeriv * *(*preOut++) ;
  }
}

/**********************************************************************/
void	backwardNoWtDvs(unit, data)
  Unit	unit ;
  void	*data ;
{
  int	numIncoming = MnumCompiledIncoming(unit) ;
  Real	inputDeriv ;
  Real 	*wt, **preDv ;
  int	idx ;

  if (Mlesioned(unit)) return ;
  inputDeriv = sigmoidDeriv(unit->output) * unit->outputDeriv ;

  for (idx=0, wt=Mweights(unit), preDv=MpreOutputDerivs(unit) ;
       idx < numIncoming ; 
       idx++) {
    *(*preDv++) += inputDeriv *  (*wt++) ;
  }
}

/***********************************************************************/
static	void	backwardRecomputeOutputError(unit, data)
  Unit	unit ;
  void	*data ;
{
  /*
   *  only called if MbpToOutputs(net) = FALSE
   */
  int	numIncoming = MnumCompiledIncoming(unit) ;
  Real	inputDeriv ;
  Real 	*wt, *dv, **preOut, **preDv ;
  int	idx ;

  if (Mlesioned(unit)) return ;

  /*
   * Calculate error derivatives for an output unit
   * (overwriting any backpropagated derivatives)
   */
  if (unit->group->type&OUTPUT) unit->outputDeriv = unitErrorDeriv(unit) ;

  inputDeriv = sigmoidDeriv(unit->output) * unit->outputDeriv ;

  for (idx=0, wt=Mweights(unit), dv=Mderivs(unit), preOut=MpreOutputs(unit),
       preDv=MpreOutputDerivs(unit) ;
       idx < numIncoming ; 
       idx++) {
    *(*preDv++) += inputDeriv *  (*wt++) ;
     (*dv++)    += inputDeriv * *(*preOut++) ;
  }
}

/**********************************************************************
 *				Utilities
 **********************************************************************/
static Proc	updateNetGradient (net)
  Net	net ;
{
  netBackward(net) ;
}

/***********************************************************************/
static	Boolean	getExample(net, exampleSet, next) 
  Net		net ;
  ExampleSet	exampleSet ;
  int		next ;
{
    
  if (MbatchSize(net) < exampleSet->numExamples) {
    /* select random example according to it's probability */
    do {
      MexampleNum(net) = randomInteger(exampleSet->numExamples);
      MexampleFrequency(net) = exampleSet->example[MexampleNum(net)].frequency ;
    } while (MsampleExamples(net) && !WITH_PROBABILITY(MexampleFrequency(net))) ;
  } else {
    MexampleNum(net) = next ;
    MexampleFrequency(net) = exampleSet->example[next].frequency ;
    if (MsampleExamples(net) && !WITH_PROBABILITY(MexampleFrequency(net)))
      return FALSE ;       /* skip this example */
  }
  
  /* Overwrite current example with random values if sampling action space */
  /* NOT HERE: DONE INSIDE netIterate... 
  if (MrandomExampleType(net))
    makeRandomExample( net, &(exampleSet->example[MexampleNum(net)]) );  
  */

  MgetNumber(exampleSet, MexampleNum(net)) ;
  
  return TRUE ;
}

/**********************************************************************/
static void	initContextOutput(unit, data)
  Unit		unit ;
  void		*data ;
{
  if (McopyFromLayer(unit->group))
    unit->output = MinitialContextOutput(unit->net) ;
}

/**********************************************************************/
static void	clampInput(unit, slice)
  Unit		unit ;
  void		*slice ;
{
  unit->output = unit->clampingHistory[(int)slice] ;
}

/**********************************************************************/
static void	setTarget(unit, slice)
  Unit		unit ;
  void		*slice ;
{
  unit->target = unit->targetHistory[(int)slice] ;
}
/**********************************************************************/
static void	setTargetHistory(unit, slice)
  Unit		unit ;
  void		*slice ;
{
  unit->targetHistory[(int)slice] = unit->target ;
}

/**********************************************************************/
static void	removeTarget(unit, slice)
  Unit		unit ;
  void		*slice ;
{
  setNaN(unit->target) ;
  setNaN(unit->targetHistory[(int)slice]) ;
}

/**********************************************************************/
extern void	saveOutputToSlice(unit, slice)
  Unit		unit ;
  void		*slice ;
{
  unit->outputHistory[(int)slice] = unit->output ;
}

/**********************************************************************/
extern void	restoreOutputFromSlice(unit, slice)
  Unit		unit ;
  void		*slice ;
{
  unit->output = unit->outputHistory[(int)slice] ;
}

static void	putExampleOnInputs(group, data)
  Group group;
  void	*data;
{
  int slice, i;
  Example example;
  
  example = &(group->net->trainingExampleSet->example[group->net->trainingExampleSet->currentIdx]);
  slice = *(int *)data;
  if( group->numUnits != example->numInputs )
    IErrorAbort( "input group has different number of units than example!" );
  for( i = 0; i < group->numUnits; i++ )
    group->unit[i]->output = example->inputVector[slice][i];
}

static void	putExampleOnTargets(group, data)
  Group group;
  void	*data;
{
  int slice, i;
  Example example;
  
  example = &(group->net->trainingExampleSet->example[group->net->trainingExampleSet->currentIdx]);
  slice = *(int *)data;
  if( group->numUnits != example->numTargets )
    IErrorAbort( "output group has different number of units than example!" );
  for( i = 0; i < group->numUnits; i++ )
    group->unit[i]->target = example->targetVector[slice][i];
}

/**********************************************************************/
static void	displayStates(net) 
  Net		net ;
{
  int i ;
  IDoTrace(&forwardEachTime) ;
  for (i=0 ; i<MdelayCount(net)*1000000 ; i++) ;
}

/**********************************************************************/
static void	copyContext(group, data)
  Group		group ;
  void		*data ;
{
  if (McopyFromLayer(group)) {
    Group	fromGroup = group->net->group[McopyFromLayer(group)] ;
    if (group->numUnits != fromGroup->numUnits) {
      IErrorAbort("Source %s numUnits (%d) != Context %s numUnits (%d)",
		  fromGroup->name, fromGroup->numUnits, group->name, group->numUnits) ;
    } else {
      Unit	*toPtr   = group->unit ;
      Unit	*endPtr  = toPtr + group->numUnits ;
      Unit	*fromPtr = fromGroup->unit ;
      while (toPtr != endPtr) {
	Unit	to   = *(toPtr++) ;
	Unit	from = *(fromPtr++) ;
	to->output = from->output ;
      }
    }
  }
}

/***********************************************************************/
static	void	initStats(net)
  Net	net ;
{
  if (net->currentEpoch==0) MlastError(net) = MaveError(net) = -1.0 ;
  if (net->mz->repetitionCount==0 || net->mz->nIterations==0 ||
      (net->mz->nIterations-1)%net->mz->repetitionCount==0) {
    net->error 		   	= 0.0 ;
    net->cost 		   	= 0.0 ;
    MpExampleCorrect(net)	= 0.0 ;
    MpSliceCorrect(net)		= 0.0 ;
    MaveFixations(net)		= 0.0 ;
    /* FM parameters */
    MpredictionError(net)	= 0.0 ;
    McomprehensionError(net) = MimitationError(net) = MintentionError(net) = 0.0 ;
  }
}

/***********************************************************************/
static	void	normStats(net)
  Net	net ;
{
  if (net->mz->repetitionCount==0 || 
      net->mz->nIterations%net->mz->repetitionCount==0) {
    Real	scale = 1.0 ;
    if (net->mz->nIterations && net->mz->repetitionCount)
      scale = (Real)net->mz->repetitionCount ;

    if (MphonologyFM(net))
      net->error += McomprehensionError(net) + MimitationError(net) + MintentionError(net) ;

    net->error		     	/= scale ;
    net->cost		     	/= scale ;
    /* FM/phonology measures */
    MpredictionError(net)	/= scale ;
    McomprehensionError(net)	/= scale ;
    MimitationError(net)	/= scale ;
    MintentionError(net)	/= scale ;
    scale 			*= net->batchSize ;
    MpSliceCorrect(net)		/= scale ;
    MpExampleCorrect(net)	/= scale ;
    MaveFixations(net)		/= scale ;
  }
}

/***********************************************************************/
static void	netForwardLayers (net, first, last, slice, training)
  Net		net ;
  int		first, last, slice ;
  Boolean	training ;
{
  netDoUnits(net,first,last,INPUT|BIAS,FALSE,(UnitDataProc)forward,       NULL) ;
  netDoUnits(net,first,last,OUTPUT,    TRUE, (UnitDataProc)setTarget,     (void *)slice) ;
  /* DCP */
  netForAllGroups(net, OUTPUT, (GroupDataProc)maybeCopyTargetsFromLayer, (void *)slice) ;
  netDoUnits(net,first,last,OUTPUT,    TRUE, (UnitDataProc)incrementError,(void *)&(net->error)) ;
  if (training) {
    netDoUnits(net,first,last,INPUT|OUTPUT, FALSE,(UnitDataProc)zeroOutputDeriv, NULL) ;
    netDoUnits(net,first,last,OUTPUT,       TRUE, (UnitDataProc)injectError, NULL) ;
    if (MunitCost(net) != 0.0)
      netDoUnits(net,first,last,INPUT|OUTPUT|BIAS,FALSE,(UnitDataProc)injectUnitCost, NULL) ;
  }
}

/**********************************************************************/
static Boolean	outputCorrectBridge(net, first, last, tolerance)
  int	first, last ;
  Net	net ;
  double tolerance ;
{
  int	groupIdx ;
  Group	*group ;

  if (tolerance <= 0.0) return FALSE ;

  for (groupIdx=first, group= &(net->group[first]);
       groupIdx <= last ;
       groupIdx++, group++) {
    if ((*group)->type&OUTPUT &&
	!groupOutputCorrect((*group), tolerance)) return FALSE ;
  }
  return TRUE ;
}

/***********************************************************************
 *			netIterate[Bridge/Refixate/PhonologyFM]
 ***********************************************************************/
static void	netIterateBridge(net, training, exampleSet)
  Net		net ;
  Boolean	training ;
  ExampleSet	exampleSet ;
{
  int	inSlice, outSlice ;
  int	bridge = MbridgeLayer(net), lastGroup = net->numGroups-1 ;
  Real	tolerance = McorrectTolerance(net) ;
  Boolean abort = FALSE ;

  netDoUnits(net, 1, bridge-1, BIAS, TRUE,
	      (UnitDataProc)initContextOutput, NULL) ;
  for (inSlice = 0 ; inSlice < MmaxSlice(net) && !abort ; inSlice++) {
    int	maxOutSlice = (MpredictOutput(net) ? MmaxSlice(net) : inSlice+1) ;
    int numOutCorrect  = 0 ;
    netForAllUnits(net, INPUT,  (UnitDataProc)clampInput, (void *)inSlice) ;
    netForwardLayers(net, 1, bridge, inSlice, training) ;
    netDoUnits(net, bridge+1, lastGroup, BIAS, TRUE,
		(UnitDataProc)initContextOutput, NULL) ;

    for (outSlice = 0 ; outSlice < maxOutSlice && !abort ; outSlice++) {
      netForwardLayers(net, bridge+1, lastGroup, outSlice, training) ;
      if (!Mrunning(net)) {
	netForAllUnits(net, ALL, (UnitDataProc)saveOutputToSlice, (void *)outSlice) ;
	displayStates(net) ;
      }
      if (training) 
	netDoUnits(net, lastGroup, bridge+1, INPUT|BIAS, FALSE, (UnitDataProc)backward, NULL) ;
      if (tolerance > 0.0 && outputCorrectBridge(net, bridge+1, lastGroup, tolerance)) numOutCorrect++ ;
      if (MpredictOutput(net)==2 && numOutCorrect <= outSlice) break ;
      if (numOutCorrect <= outSlice) {
	if (MpredictOutput(net)==2) break ;
	else if (MabortIfIncorrect(net)) {
	  abort = TRUE ; break ;
	}
      }
      if (outSlice < MmaxSlice(net)-1)
	netDoGroups(net, bridge+1, lastGroup, BIAS, TRUE, (GroupDataProc)copyContext, NULL) ;
    } /* outSlice */
    MpSliceCorrect(net)  += numOutCorrect/((Real)(maxOutSlice*MmaxSlice(net))) ;

    if (training) 
      netDoUnits(net, bridge, 1, INPUT|BIAS, FALSE, (UnitDataProc)backward, NULL) ;
    if (MpredictOutput(net)==2 && numOutCorrect <= inSlice) break ;
    if (inSlice < MmaxSlice(net)-1)
      netDoGroups(net, 1, bridge, BIAS, TRUE, (GroupDataProc)copyContext, NULL) ;
    else if (numOutCorrect == maxOutSlice) 
      MpExampleCorrect(net)++ ;
  } /* inSlice */
}

/***********************************************************************/
static void	netIterateRefixate(net, training, exampleSet)
  Net		net ;
  Boolean	training ;
  ExampleSet	exampleSet ;
{
  int	slice, numCorrect = 0, outCorrect ;
  int	fixation=0, attention=0, saccade=0, outputOK, done, attOK ; 
  Group	Input, Attention, AttentionOut, Done ;

  if (!MinputGroupName(net))     MinputGroupName(net)     = "Letters" ;
  /* note that this is the INPUT Attention group */
  if (!MattentionGroupName(net)) MattentionGroupName(net) = "cAttention" ;
  if (!MdoneGroupName(net))      MdoneGroupName(net)      = "Done" ;

  if ((Input    =groupFromName(net, MinputGroupName(net)))     == NULL) 
    IErrorAbort("no \"%s\" group", MinputGroupName(net)) ;
  if ((Attention=groupFromName(net, MattentionGroupName(net))) == NULL) 
    IErrorAbort("no \"%s\" group", MattentionGroupName(net)) ;
  AttentionOut = net->group[Attention->extension->copyFromLayer] ;
  if ((Done     =groupFromName(net, MdoneGroupName(net)))      == NULL) 
    IErrorAbort("no \"%s\" group",MdoneGroupName(net)) ;
  
  /* CONTEXT GROUPS MUST BE TYPE BIAS */
  netForAllUnits(net, BIAS, (UnitDataProc)initContextOutput, NULL) ;
  MaveFixations(net)++ ;

  /* FOR EACH TIMESLICE - ONE PER PHONEME */
  for (slice = 0 ; slice < MmaxSlice(net) ; slice++) {
    net->currentTime = slice ;
    /* this is applied to the OUTPUT group because targets are shifted */
    Fixate(Input, AttentionOut, saccade, attention, slice) ;
    copyContext(Attention, NULL) ;
    netForAllUnits(net, INPUT,  (UnitDataProc)clampInput, (void *)slice) ;
    netForAllUnits(net, OUTPUT, (UnitDataProc)setTarget,  (void *)slice) ;

    /* FORWARD AND BACKWARD PASSES */
    netForward(net) ;
    if (!Mrunning(net)) {
      netForAllUnits(net, ALL, (UnitDataProc)saveOutputToSlice, (void *)slice) ;
      displayStates(net) ;
    }
    if (outCorrect=outputCorrect(net, McorrectTolerance(net))) numCorrect++ ;
    if (training) {
      outputOK = outCorrect ;
      done     = (outputOK && Done->unit[0]->target==1.0) ;
      netBackward(net) ;
    } else {
      outputOK = outputClean(net, McorrectTolerance(net)) ;
      done     = (Done->unit[0]->output > maxOutput(AttentionOut)) ;
    }
    if (Mdebug(net)) 
      fprintf(dout, "%d\tfixate %d(%d)\t", slice, fixation, attention) ;

    /* DONE - GET OUT */
    if (done) {
      if (Mdebug(net)) fprintf(dout, "done    %s\n", 
			       (attention ? "in periphery" : "at fixation ")) ;
      break ;

    /* CORRECT - TRY NEXT PHONEME WITHOUT REFIXATING */
    } else if (outputOK) {
      if (Mdebug(net)) fprintf(dout, "correct %s - no saccade\n", 
			       (attention ? "in periphery" : "at fixation ")) ;
      saccade   = fixation ;	/* no change */
      attention = (training ? maxTargetPosition(AttentionOut) : maxOutputPosition(AttentionOut)) ;
      netForAllGroups(net, BIAS, (GroupDataProc)copyContext, NULL) ;

    /* FAILED ON FIXATED GRAPHEME - REFIXATE NEXT GRAPHEME */
    } else if (attention == 0) { 
      attention = (training ? maxTargetPosition(AttentionOut) : maxOutputPosition(AttentionOut)) ;
      if (Mdebug(net)) fprintf(dout, "fail    at fixation  - saccade %d\n", attention) ;
      fixation += attention ;
      saccade   = fixation ;
      attention = 0 ;
      MaveFixations(net)++ ;
      if (slice < MmaxSlice(net)-1)
	netForAllGroups(net, BIAS, (GroupDataProc)copyContext, NULL) ;

    /* FAILED ON NONFIXATED GRAPHEME - REFIXATE IT AND TRY AGAIN */
    } else {
      if (Mdebug(net)) fprintf(dout, "fail    in periphery - saccade %d and repeat\n", attention) ;
      fixation += attention ;
      saccade   = attention ;		/* applied to OLD states/targets */
      attention = 0 ;
      MaveFixations(net)++ ;
      slice-- ;
      netForAllUnits     (net, BIAS,       (UnitDataProc)initContextOutput,      NULL) ;
      netForAllOtherUnits(net, INPUT|BIAS, (UnitDataProc)restoreOutputFromSlice, (void *)slice) ;
    }
  } /* slice */
  if (numCorrect == MmaxSlice(net)) MpExampleCorrect(net)++ ;
  MpSliceCorrect(net)  += numCorrect/((Real) MmaxSlice(net)) ;
}

/***********************************************************************/
/**********************************************************************/
static	void	linkLoadWeight(link, data)
  Link	link ;
  void	*data ;
{
  FILE	*fp = (FILE *)data ;
  double wt ;

  if (fscanf(fp, "%lf", &wt)==EOF) {
    ICloseFile(fp,NULL) ;
    IErrorAbort("Error reading weight for link %s", link->name) ;
  } else 
    link->weight = (Real)wt ;
}
static void	unitLoadWeights(unit, fp)
  Unit		unit ;
  void		*fp ;
{
  unitForAllLinks(unit, INCOMING, (LinkDataProc)linkLoadWeight, fp) ;
}
static	void	linkSaveWeight(link, data)
  Link	link ;
  void	*data ;
{
  FILE	*fp = (FILE *)data ;
  fprintf(fp, "%f\n", link->weight) ;
}
static void	unitSaveWeights(unit, fp)
  Unit		unit ;
  void		*fp ;
{
  unitForAllLinks(unit, INCOMING, (LinkDataProc)linkSaveWeight, fp) ;
}

int	command_comprehensionWeights(tokc, tokv)
  int	tokc ;
  char	**tokv ;
{
  int	SemanticsOut, PhonologyOut, Acoustics, PhonologyIn, SemanticsIn ;
  String file_name = NULL;
  FILE	*fp ;
  struct ARG_DESC *argd ;

  argd = StartArgs(tokv[0]);
  Args(argd, "[<file>%s%?]", &file_name, "file to read from");
  EndArgs(argd);
  if (GiveHelp(tokc)) {
    ISynopsis("Load/save comprehension weights (phonologyFM nets).") ;
    IHelp
      (tokc, tokv[0], NULL, synopsis,
       NULL);
    return 0;
  }
  (void) ParseArgs(argd, tokc, tokv, 0);

  if (currentNet == NULL) IErrorAbort("There is no current net.") ;

  SemanticsOut	= groupIdxFromName(currentNet, "SemanticsOut") ;
  PhonologyOut	= groupIdxFromName(currentNet, "PhonologyOut") ;
  Acoustics	= groupIdxFromName(currentNet, "Acoustics") ;
  PhonologyIn	= groupIdxFromName(currentNet, "PhonologyIn") ;
  SemanticsIn	= groupIdxFromName(currentNet, "SemanticsIn") ;

  if (file_name) {
    fprintf(dout, "Loading comprehension weights from \"%s\"...", file_name) ;
    fp = IOpenFileOrAbort(file_name, "r", NULL) ;
    netDoUnits(currentNet, SemanticsOut+1, PhonologyOut, ALL, TRUE,
	       (UnitDataProc)unitLoadWeights, (void *)fp) ;
    netDoUnits (currentNet, Acoustics+1, SemanticsIn, ALL, TRUE,
		(UnitDataProc)unitLoadWeights, (void *)fp) ;
    ICloseFile(fp, NULL) ;
    fprintf(dout, "done\n") ;
  } else {
    netDoUnits(currentNet, SemanticsOut+1, PhonologyOut, ALL, TRUE,
	       (UnitDataProc)unitSaveWeights, (void *)dout) ;
    netDoUnits (currentNet, Acoustics+1, SemanticsIn, ALL, TRUE,
		(UnitDataProc)unitSaveWeights, (void *)dout) ;
  }    
  return 1 ;
}
int	command_forwardModelWeights(tokc, tokv)
  int	tokc ;
  char	**tokv ;
{
  int	Articulation, Acoustics ;
  String file_name = NULL;
  FILE	*fp ;
  struct ARG_DESC *argd ;

  argd = StartArgs(tokv[0]);
  Args(argd, "[<file>%s%?]", &file_name, "file to read from");
  EndArgs(argd);
  if (GiveHelp(tokc)) {
    ISynopsis("Load/save forward model weights (phonologyFM nets).") ;
    IHelp
      (tokc, tokv[0], NULL, synopsis,
       NULL);
    return 0;
  }
  (void) ParseArgs(argd, tokc, tokv, 0);

  if (currentNet == NULL) IErrorAbort("There is no current net.") ;

  Articulation	= groupIdxFromName(currentNet, "Articulation") ;
  Acoustics	= groupIdxFromName(currentNet, "Acoustics") ;

  if (file_name) {
    fprintf(dout, "Loading forward model weights from \"%s\"...", file_name) ;
    fp = IOpenFileOrAbort(file_name, "r", NULL) ;
    netDoUnits(currentNet, Articulation+1, Acoustics, ALL, TRUE,
	       (UnitDataProc)unitLoadWeights, (void *)fp) ;
    ICloseFile(fp, NULL) ;
    fprintf(dout, "done\n") ;
  } else {
    netDoUnits(currentNet, Articulation+1, Acoustics, ALL, TRUE,
	       (UnitDataProc)unitSaveWeights, (void *)dout) ;
  }    
  return 1 ;
}
/**********************************************************************/
extern void	netBackwardFromContext(net, last, first)
  Net		net ;
  int		last, first ;	/* these are group indexes; last must have context layer */
{
  int	idx, context ;
  int	slice = net->currentTime ;

  /* NOTE: REQUIRES OUTPUTS TO BE SAVED INTO outputHistory */

  if (slice == 0) return ;	/* makes no sense for first slice */
  if (last < first )
    IErrorAbort("index of last group \"%s\" (%d) must not be less than that of first group \"%s\" (%d)",
		net->group[last]->name, last, net->group[first]->name, first) ;
  if (!(context=copiedFromLayer(net, last)))
    IErrorAbort("last group \"%s\" (%d) has no context layer", net->group[last]->name, last) ;
  
  /* reinstate previous outputs */
  for (idx = last ; idx >= first ; idx--) 
    copyOutputHistoryToOutputs (net, idx, idx, slice-1) ;

  /* copy derivatives from context layer to last */
  copyOutputDerivsToOutputDerivs(net, context, last) ;

  /* zero output derivatives between last and first (last will get set; first should accumulate) */
  netDoUnits(net, last-1, first+1, ALL, TRUE, (UnitDataProc)zeroOutputDeriv, NULL) ;

  /* back-prop from last to first */
  netDoUnits(net, last, first+1, INPUT|BIAS, FALSE, (UnitDataProc)backward, NULL) ;

  /* reinstate current outputs */
  for (idx = last ; idx >= first ; idx--) 
    copyOutputHistoryToOutputs (net, idx, idx, slice) ;
}
/**********************************************************************/
static void	netIteratePhonologyFM(net, training, exampleSet)
  Net		net ;
  Boolean	training ;
  ExampleSet	exampleSet ;
{
  int	slice ;
  Real	tolerance = McorrectTolerance(net) ;
  int	anticipate = Manticipate(net), perseverate = Mperseverate(net) ;
  int	SemanticsOut, PhonologyOut, CArticulation, Articulation, Acoustics, PhonologyIn, SemanticsIn ;

  /* NOTE: make Acoustics an OUTPUT group in order use adult Acoustics 
   *       as "goal" targets for outcome layer 
   *
   * Runs only comprehension steps if there is no Articulation (actionLayer) group
   */
  SemanticsOut		= groupIdxFromName(net, "SemanticsOut") ;	/* I */
  PhonologyOut		= groupIdxFromName(net, "PhonologyOut") ;	/* O */
  CArticulation		= groupIdxFromName(net, "CArticulation") ;
  Articulation		= groupIdxFromName(net, "Articulation") ;
  Acoustics		= groupIdxFromName(net, "Acoustics") ;
  PhonologyIn		= groupIdxFromName(net, "PhonologyIn") ;	/* O */
  SemanticsIn		= groupIdxFromName(net, "SemanticsIn") ;	/* O */
  MactionLayer(net)	= (Articulation < 0 ? 0 : Articulation) ;
  MoutcomeLayer(net)	= Acoustics ;
  if (!MoutcomeProc(net)) setOutcomeProc(net);

  if (MonlyBabble(net)) goto babble ;

  /**************************************************************
   * STEP 1: Comprehension of adult (Acoustics->SemanticsIn)	
   **************************************************************/

  /* NOTE: this sets both outputs and targets of outcome layer */
  /* targets will not be used unless outcome layer (Acoustics) is of type OUTPUT */
  computeAcousticGoals( net, &(exampleSet->example[exampleSet->currentIdx]) );

  /* can't be sure of numbering of context layers, so initialize them all */
  netForAllUnits(net, BIAS, (UnitDataProc)initContextOutput, NULL) ;
  for (slice = 0 ; slice < MmaxSlice(net) ; slice++) {
    int	lastLayer = (MonlyTrainImitate(net) && !MtrainComp(net) ? PhonologyIn : SemanticsIn) ;
    net->currentTime = slice ;
    /* Acoustic input set by computeAcousticGoals */
    groupForAllUnits(net->group[Acoustics], (UnitDataProc)clampInput, (void *)slice) ;
    netDoUnits (net, Acoustics+1, lastLayer, OUTPUT, TRUE, (UnitDataProc)setTarget, (void *)slice) ;
    netDoGroups(net, Acoustics+1, lastLayer, INPUT|BIAS, FALSE, (GroupDataProc)groupForward, (void *)slice) ;
    netDoUnits (net, Acoustics+1, lastLayer, OUTPUT, TRUE, (UnitDataProc)incrementError, &McomprehensionError(net)) ;
    /* need to save so we can copy PhonIn outputs to PhonIn targets for imitation slice by slice */
    groupForAllUnits(net->group[PhonologyIn], (GroupDataProc)saveOutputToSlice, (void *)slice) ;
    if (!Mrunning(net)) {
      netForAllUnits(net, ALL, (UnitDataProc)saveOutputToSlice, (void *)slice) ;
      displayStates(net) ;
    }
    if (training && MtrainComp(net)) {
      netDoUnits(net, lastLayer, Acoustics+1, ALL, TRUE, (UnitDataProc)zeroOutputDeriv, NULL) ;
      netDoUnits(net, lastLayer, Acoustics+1, OUTPUT, TRUE, (UnitDataProc)injectError, NULL) ;
      netDoUnits(net, lastLayer, Acoustics+1, INPUT|BIAS, FALSE, (UnitDataProc)backward, NULL) ;
    }
    if (slice < MmaxSlice(net)-1)
      netForAllGroups(net, BIAS, (GroupDataProc)copyContext, NULL) ;
  }

  /*****************************************************************
   * STEP 2: Imitation of adult Phonology (Phon->Art->Acous->Phon) 
   *****************************************************************/

  /* copy final PhonologyIn pattern from adult to PhonologyOut to use as input for imitation */
  copyOutputsToOutputs(net, PhonologyIn, PhonologyOut) ;
  if (MactionLayer(net)) {
    int lastLayer = (MstopAtAcoustics(net) && net->group[Acoustics]->type&OUTPUT ? Acoustics : PhonologyIn) ;
    if (lastLayer == PhonologyIn)
      copyOutputsToTargets(net, PhonologyIn, PhonologyIn, TRUE) ; /* slice-by-slice */
    netForAllUnits(net, BIAS, (UnitDataProc)initContextOutput, NULL) ;

    for (slice = 0 ; slice < MmaxSlice(net) ; slice++) {
      net->currentTime = slice ;
      /* this will set acoustic targets if Acoustics is OUTPUT layer */
      netDoUnits (net, PhonologyOut+1, lastLayer, OUTPUT, TRUE, (UnitDataProc)setTarget, (void *)slice) ;
      /* NOTE: Acoustics is set to actual outcome of Articulation */
      netDoGroups(net, PhonologyOut+1, lastLayer, INPUT|BIAS, FALSE, (GroupDataProc)groupForward, (void *)slice) ;
      /* run to semantics for last slice */
      if (lastLayer == PhonologyIn && slice == MmaxSlice(net)-1)
	netDoGroups(net, PhonologyIn+1, SemanticsIn, INPUT|BIAS, FALSE, (GroupDataProc)groupForward, (void *)slice) ;
      netDoUnits (net, PhonologyOut+1, lastLayer, OUTPUT, TRUE, (UnitDataProc)incrementError, &MimitationError(net)) ;
      netForAllUnits(net, ALL, (UnitDataProc)saveOutputToSlice, (void *)slice) ;
      if (!Mrunning(net)) displayStates(net) ;
      if (training) {
	/* NOTE: zeroing derivs for PhonOut in case adaptComprehension=1 */
	/*       and zeroing context layers */
	netDoUnits(net, lastLayer, PhonologyOut, ALL, TRUE, (UnitDataProc)zeroOutputDeriv, NULL) ;
	netDoUnits(net, lastLayer, PhonologyOut+1, OUTPUT, TRUE, (UnitDataProc)injectError, NULL) ;
	/* NOTE: Does not adapt comprehension based on own articulations directly */
	netDoUnits(net, lastLayer, Articulation+1, INPUT|BIAS, FALSE, (UnitDataProc)backwardNoWtDvs, NULL) ;
	netDoUnits(net, Articulation, PhonologyOut+1, INPUT|BIAS, FALSE, (UnitDataProc)backward, NULL) ;
	/* adapt articulation from prior slice as well */
	netBackwardFromContext(net, Articulation, PhonologyOut) ;
	/* maybe copy derivatives of PhonOut to PhonIn and adapt Acous->PhonIn */
	if (MadaptComprehension(net) && lastLayer == PhonologyIn) {
	  copyOutputDerivsToOutputDerivs(net, PhonologyOut, PhonologyIn) ;
	  netDoUnits(net, PhonologyIn-1, Acoustics+1, ALL, TRUE, (UnitDataProc)zeroOutputDeriv, NULL) ;
	  netDoUnits(net, PhonologyIn, Acoustics+1, INPUT|BIAS, FALSE, (UnitDataProc)backward, NULL) ;
	}
	if (MtrainFM(net)) netTrainFM(net, Acoustics, Articulation) ;
      }
      if (slice < MmaxSlice(net)-1)
	netForAllGroups(net, BIAS, (GroupDataProc)copyContext, NULL) ;
    }
  }

  /********************************************************************
   * STEP 3: Intentional utterance (Sem->Phon->Art->Acous->Phon->Sem) 
   ********************************************************************/

  if (MonlyTrainImitate(net)) goto babble ;

  /* outputs for PhonOut were generated by comprehension and copied from PhonIn */
  copyOutputsToTargets(net, PhonologyOut, PhonologyOut, FALSE) ; /* only final output */
  /* run SemOut->PhonOut only once; first get SemOut input states */
  groupForAllUnits(net->group[SemanticsOut], (UnitDataProc)clampInput, (void *)0) ;
  netDoGroups(net, SemanticsOut+1, PhonologyOut, INPUT|BIAS, FALSE, (GroupDataProc)groupForward, (void *)0) ;
  netDoUnits (net, SemanticsOut+1, PhonologyOut, OUTPUT, TRUE, (UnitDataProc)incrementError, &McomprehensionError(net)) ;
  if (training && MtrainComp(net)) {
    netDoUnits(net, PhonologyOut, SemanticsOut+1, ALL, TRUE, (UnitDataProc)zeroOutputDeriv, NULL) ;
    netDoUnits(net, PhonologyOut, SemanticsOut+1, OUTPUT, TRUE, (UnitDataProc)injectError, NULL) ;
    netDoUnits(net, PhonologyOut, SemanticsOut+1, INPUT|BIAS, FALSE, (UnitDataProc)backward, NULL) ;
  }
  if (MactionLayer(net)) {
    netForAllUnits(net, BIAS, (UnitDataProc)initContextOutput, NULL) ;
    for (slice = 0 ; slice < MmaxSlice(net) ; slice++) {
      net->currentTime = slice ;
      /* NOTE: no targets for PhonIn, and don't permit (adult) goal for Acoustics */
      groupForAllUnits(net->group[PhonologyIn], (UnitDataProc)removeTarget, (void *)slice) ;
      groupForAllUnits(net->group[Acoustics],   (UnitDataProc)removeTarget, (void *)slice) ;
      /* NOTE: Acoustics is set to actual outcome of Articulation */
      netDoGroups(net, PhonologyOut+1, SemanticsIn, INPUT|BIAS, FALSE, (GroupDataProc)groupForward, (void *)slice) ;
      netDoUnits (net, PhonologyOut+1, SemanticsIn, OUTPUT, TRUE, (UnitDataProc)incrementError, &MintentionError(net)) ;
      if (!Mrunning(net)) {
	netForAllUnits(net, ALL, (UnitDataProc)saveOutputToSlice, (void *)slice) ;
	displayStates(net) ;
      }
      if (training) {
	/* NOTE: zeroing BIAS (context) layers */
	netDoUnits(net, SemanticsIn, SemanticsOut, ALL, TRUE, (UnitDataProc)zeroOutputDeriv, NULL) ;
	netDoUnits(net, SemanticsIn, PhonologyOut+1, OUTPUT, TRUE, (UnitDataProc)injectError, NULL) ;
	/* NOTE: Does not adapt comprehension based on own articulations directly */
	netDoUnits(net, SemanticsIn, Articulation+1, INPUT|BIAS, FALSE, (UnitDataProc)backwardNoWtDvs, NULL) ;
	netDoUnits(net, Articulation, PhonologyOut+1, INPUT|BIAS, FALSE, (UnitDataProc)backward, NULL) ;
	/* adapt articulation from prior slice as well */
	netBackwardFromContext(net, Articulation, PhonologyOut) ;
	if (MadaptComprehension(net)) {
	  /* bp all the way back to SemanticsOut */
	  netDoUnits(net, PhonologyOut, SemanticsOut+1, INPUT|BIAS, FALSE, (UnitDataProc)backward, NULL) ;
	  /* copy derivatives of PhonOut to PhonIn and adapt Acous->PhonIn */
	  copyOutputDerivsToOutputDerivs(net, PhonologyOut, PhonologyIn) ;
	  netDoUnits(net, PhonologyIn-1, Acoustics+1, ALL, TRUE, (UnitDataProc)zeroOutputDeriv, NULL) ;
	  netDoUnits(net, PhonologyIn, Acoustics+1, INPUT|BIAS, FALSE, (UnitDataProc)backward, NULL) ;
	} 
	if (MtrainFM(net)) netTrainFM(net, Acoustics, Articulation) ;
      }
      if (slice < MmaxSlice(net)-1)
	netForAllGroups(net, BIAS, (GroupDataProc)copyContext, NULL) ;
    }
  } else if (!Mrunning(net)) {
    netForAllUnits(net, ALL, (UnitDataProc)saveOutputToSlice, (void *)(MmaxSlice(net)-1)) ;
    displayStates(net) ;
    groupForAllUnits(net->group[PhonologyOut], (UnitDataProc)setTargetHistory, (void *)(MmaxSlice(net)-1)) ;
  }

  /********************************************************************
   * STEP 4: Babble (Art->Acous) 
   ********************************************************************/

 babble:
  if (!Mbabble(net) || !MactionLayer(net)) return ;
  /* this is done last because makeRandomExample overwrites Acoustic targets */
  Manticipate(net) = Mperseverate(net) = 0 ;
  makeRandomExample( net, &(exampleSet->example[MexampleNum(net)]) );  
  Manticipate(net) = anticipate; Mperseverate(net) = perseverate ;
  copyOutputsToOutputs(net, Articulation, CArticulation) ;
  for (slice = 0 ; slice < MmaxSlice(net) ; slice++) {
    net->currentTime = slice ;
    groupForAllUnits(net->group[Articulation], (UnitDataProc)clampInput, (void *)slice) ;
    groupForAllUnits(net->group[Acoustics],    (UnitDataProc)setTarget,  (void *)slice) ;
    /* NOTE: Acoustics is set to actual outcome of Articulation */
    /*       and incrementError is called in groupForward */
    netDoGroups(net, Articulation+1, Acoustics, INPUT|BIAS, FALSE,
		(GroupDataProc)groupForward, (void *)slice) ;
    if (!Mrunning(net)) {
	netForAllUnits(net, ALL, (UnitDataProc)saveOutputToSlice, (void *)slice) ;
	displayStates(net) ;
      }
    if (training) netTrainFM(net, Acoustics, Articulation) ;
    if (slice < MmaxSlice(net)-1)
      netForAllGroups(net, BIAS, (GroupDataProc)copyContext, NULL) ;
  }
}

/***********************************************************************/
static void	netIterate(net, training, exampleSet)
  Net		net ;
  Boolean	training ;
  ExampleSet	exampleSet ;
{
  int slice, numCorrect = 0, i, contextNum ;
  Real	tolerance = McorrectTolerance(net) ;
  Group	goalGp, actionGp, outcomeGp;
  
  MmaxSlice(net) = lastTargetSlice(net) ;

  if (MbridgeLayer(net))
    netIterateBridge(net, training, exampleSet) ;
  else if (Mrefixate(net))
    netIterateRefixate(net, training, exampleSet) ;
  else if (MphonologyFM(net))
    netIteratePhonologyFM(net, training, exampleSet) ;
  else {
    if (MrandomExampleType(net))
      makeRandomExample( net, &(exampleSet->example[MexampleNum(net)]) );  

    /* 
     * assumes context groups are type BIAS
     */
    netForAllUnits(net, BIAS, (UnitDataProc)initContextOutput, NULL) ;
        
    /* compute goal states, currently for converting ascii phonemes to acoustic targets 
       do this here and now instead of in netForwardFM */
    if( exampleSet->example[net->trainingExampleSet->currentIdx].name[0] == '/' ) 
      computeAcousticGoals( net, &(exampleSet->example[net->trainingExampleSet->currentIdx]) );

    for (slice = 0 ; slice < MmaxSlice(net) ; slice++) {
      net->currentTime = slice ;
      netForAllUnits(net, INPUT,  (UnitDataProc)clampInput, (void *)slice) ;
      netForAllUnits(net, OUTPUT, (UnitDataProc)setTarget,  (void *)slice) ;

      /* assume some prep time by copying the action/articulatory layer to its context for first slice
	 NOTE: this only works when the articulatory layer is an input layer */
      if (slice == 0 &&
	  MactionLayer(net) &&
	  net->group[MactionLayer(net)]->type&INPUT &&
	  (contextNum=copiedFromLayer(net, MactionLayer(net))))
	copyOutputsToOutputs(net, MactionLayer(net), contextNum);

      netForward(net) ;
      netForAllUnits(net, ALL, (UnitDataProc)saveOutputToSlice, (void *)slice) ;
      if (!Mrunning(net)) displayStates(net) ;
      if (training) netBackward(net) ;
      if (tolerance > 0.0 && outputCorrect(net, tolerance)) numCorrect++ ;
      if (MpredictOutput(net)==2 && numCorrect <= slice) break ;
      if (slice < MmaxSlice(net)-1)
	netForAllGroups(net, BIAS, (GroupDataProc)copyContext, NULL) ;
      else if (numCorrect == MmaxSlice(net)) 
	MpExampleCorrect(net)++ ;
    }
    MpSliceCorrect(net)  += numCorrect/((Real) MmaxSlice(net)) ;
  }
}

static Real	_epsilon ;
static Real	_alpha ;
/**********************************************************************/
void		linkZeroDeltaWeight(link, data)
  Link		link ;
  void		*data ;
{
  link->deltaWeight = 0.0 ;
}
/**********************************************************************/
static void	unitZeroDeltaWeights(unit, data)
  Unit		unit ;
  void		*data ;
{
  unitForAllLinks(unit, INCOMING, (LinkDataProc)linkZeroDeltaWeight, NULL) ;
}
/**********************************************************************/
static void	unitChangeWeights(unit, data)
  Unit		unit ;
  void		*data ;
{
  int	idx ;
  int	numIncoming 	= unit->numIncoming ;
  Link	*incoming   	= unit->incomingLink ;

  if (_alpha > 0.0)
    for (idx = 0 ; idx < numIncoming ; ++idx) {
      Link link		= incoming[idx] ;
      link->deltaWeight = _alpha * link->deltaWeight - _epsilon * link->deriv ;
      link->weight     +=  link->deltaWeight ;
    }
  else
    for (idx = 0 ; idx < numIncoming ; ++idx) {
      Link link		= incoming[idx] ;
      link->deltaWeight = -_epsilon * link->deriv ;
      link->weight     += link->deltaWeight ;
    }
}

static Real	_g2 ;
static Real	_w2 ;
static Real	_gw ;
/**********************************************************************/
void		linkCosAngle(link, data)
  Link		link ;
  void		*data ;
{
  _g2 += link->deriv       * link->deriv ;
  _w2 += link->deltaWeight * link->deltaWeight ;
  _gw += link->deriv       * link->deltaWeight ;
}
/**********************************************************************/
static void	unitCosAngle(unit, data)
  Unit		unit ;
  void		*data ;
{
  unitForAllLinks(unit, INCOMING, (LinkDataProc)linkCosAngle, NULL) ;
}
/***********************************************************************
 *			netMinimize
 ***********************************************************************/
static void	netMinimize(net, exampleSet)
  Net		net ;
  ExampleSet	exampleSet ;
{
  int numExamples ;
  Real	cosAngle ;

  fprintf(dout, "WARNING: not returning to minimize\n") ;

  if (!Mcompiled(net)) netCompile(net) ;

  if (net->currentEpoch == 0) {
    fprintf(dout, "Initializing momentum direction.\n") ;
    netForAllUnits(net, ALL, (UnitDataProc)unitZeroDeltaWeights, NULL) ;
  }
  for ( ; net->mz->maxIterations==0 || net->mz->nIterations <= net->mz->maxIterations ; 
       net->mz->nIterations++, net->currentEpoch++) {

    netForAllUnits     (net, ALL,       (UnitDataProc)unitConstrainWeights, NULL) ;
    netForAllOtherUnits(net, INPUT|BIAS, (UnitDataProc)getCompiledWeights, NULL) ;

    Mrunning(net) = TRUE ;
    initStats(net) ;
    initEpoch(net, MalphaOnset(net)) ;
    _epsilon = net->mz->epsilon ;
    _alpha   = net->mz->alpha ;
    netForAllUnits(net, ALL, (UnitDataProc)zeroLinkDerivs, net) ;

    for (numExamples = 0 ; numExamples < MbatchSize(net) ; ++numExamples) {
      if (!getExample(net, exampleSet, numExamples)) continue ;

      if (net->type & RECURRENT)			/* elman network */
	netIterate(net, TRUE, exampleSet) ;
      else {				
	netForward(net) ;
	netBackward(net) ;
      }
    } /* examples */
    if (numExamples <= 0)
      IErrorAbort("netCalculate: no examples processed") ;

    /* compiled derivatives are put back first because cost functions 
       currently use standard data structures */
    netForAllOtherUnits(net, INPUT|BIAS, (UnitDataProc)putCompiledDerivs, NULL) ;
    if (MlengthScaleDerivs(net))
      netForAllUnits(net, ALL, (UnitDataProc)lengthScaleDerivs, net) ;

    /* update the cost after everything else is done */
    if (MweightCostIncrement(net) > 0.0) 
      netAdaptWeightCost(net) ;
    netUpdateWeightCost(net) ;
    normStats(net) ;

    /* going to be time to print - want to do this BEFORE weight changes */
    if (net->mz->repetitionCount==0 || net->currentEpoch%net->mz->repetitionCount == 0) {
      _g2 = _w2 = _gw = 0.0 ;
      netForAllUnits(net, ALL, (UnitDataProc)unitCosAngle, NULL) ;
      cosAngle = (_g2*_w2 <= 0.0 ? 0.0 :_gw/sqrt(_g2*_w2)) ;
    }
    /* all but last iteration */
    if (net->mz->maxIterations==0 || net->mz->nIterations < net->mz->maxIterations) {
      IDoTrace(&minimizeEndIteration) ;
      /* NOTE: only does gradient/momentum descent */
      netForAllUnits(net, ALL, (UnitDataProc)unitChangeWeights, NULL) ;
      if (net->mz->repetitionCount==0 || net->currentEpoch%net->mz->repetitionCount == 0) 
	IDoTrace(&minimizeEndRepetition) ;
    }
    /* time to print */
    if (net->mz->repetitionCount==0 || net->currentEpoch%net->mz->repetitionCount == 0)
      fprintf(dout, "epoch %d err= %f cos= %f\n", net->currentEpoch, net->error, cosAngle) ;
  } /* epoch */
  Mrunning(net) = FALSE ;
}

/***********************************************************************
 *			netCalculate
 ***********************************************************************/
static void	netCalculate(net, exampleSet, training)
  Net		net ;
  ExampleSet	exampleSet ;
  Boolean	training ;
{
  int numExamples ;

  if (!Mcompiled(net)) netCompile(net) ;
  netForAllUnits     (net, ALL,       (UnitDataProc)unitConstrainWeights, NULL) ;
  netForAllOtherUnits(net, INPUT|BIAS, (UnitDataProc)getCompiledWeights, NULL) ;

  Mrunning(net) = TRUE ;
  if (training) {
    initStats(net) ;
    initEpoch(net, MalphaOnset(net)) ;
    netForAllUnits(net, ALL, (UnitDataProc)zeroLinkDerivs, net) ;
  }
  for (numExamples = 0 ; numExamples < MbatchSize(net) ; ++numExamples) {
    if (!getExample(net, exampleSet, numExamples)) continue ;

    if (net->type & RECURRENT) {			/* elman network */
      netIterate(net, training, exampleSet) ;
      } else {				
      netForward(net) ;
      if (training) netBackward(net) ;
    }
  }
  if (numExamples <= 0)
    IErrorAbort("netCalculate: no examples processed") ;

  if (training) {
    /* compiled derivatives are put back first because cost functions 
       currently use standard data structures */
    netForAllOtherUnits(net, INPUT|BIAS, (UnitDataProc)putCompiledDerivs, NULL) ;
    if (MlengthScaleDerivs(net))
      netForAllUnits(net, ALL, (UnitDataProc)lengthScaleDerivs, net) ;

    /* update the cost after everything else is done */
    if (MweightCostIncrement(net) > 0.0) 
      netAdaptWeightCost(net) ;
    netUpdateWeightCost(net) ;
    normStats(net) ;
  }
  Mrunning(net) = FALSE ;
}

/***********************************************************************/
static void	calculateNetErrorDeriv(net, exampleSet)
  Net		net ;
  ExampleSet	exampleSet ;
{
  if (MphonologyFM(net))
    netMinimize(net, exampleSet) ;
  else
    netCalculate(net, exampleSet, TRUE) ;
}

/*********************************************************************/
static void	calculateNetError(net, exampleSet)
  Net		net ;
  ExampleSet	exampleSet ;
{
  netCalculate(net, exampleSet, FALSE) ;
}

/**********************************************************************/
static Proc	updateNetActivity (net)
  Net	net ;
{
  if (!Mcompiled(net)) netCompile(net) ;
  netForAllUnits     (net, ALL,       (UnitDataProc)unitConstrainWeights, NULL) ;
  netForAllOtherUnits(net, INPUT|BIAS, (UnitDataProc)getCompiledWeights, NULL) ;

  Mrunning(net) = FALSE ;
  net->error = MpredictionError(net) = MlastError(net) = MaveError(net) = 0.0 ;
  MoutputsIncorrect(net) = 0 ;
  MpSliceCorrect(net) = MpExampleCorrect(net) =  MaveFixations(net) = 0.0 ;

  if (MtargetExternalInput(net)>0.0 && MtargExtInConstant(net)>0 && net->cost > 0.0) 
    /* stored by pronounce in SM.c --- NET->ERROR WILL NOT BE ACCURATE */
    MexampleFrequency(net) = net->cost ;
  else
    MexampleFrequency(net) = 1.0 ;

  /* kluge for now: assume that training set is the selected one */
  MexampleNum(net) = net->trainingExampleSet->currentIdx;
  
  if (net->type & RECURRENT)  	/* elman network */
    netIterate(net, FALSE, net->trainingExampleSet) ;
  else
    netForward(net) ;
    
  IDoTrace(&afterExample) ;
}
/***********************************************************************/
