
/**********************************************************************
 * $Id: bp.c,v 1.9 91/09/26 11:05:14 drew Exp Locker: drew $
 **********************************************************************/

/**********************************************************************
 *	  Copyright 1990,1991 by The University of Toronto,
 *		      Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its  documentation for  any purpose is  hereby granted  without
 * fee, provided that the above copyright notice appears in all copies
 * and  that both the  copyright notice  and   this  permission notice
 * appear in   supporting documentation,  and  that the  name  of  The
 * University  of Toronto  not  be  used in  advertising or  publicity
 * pertaining   to  distribution   of  the  software without specific,
 * written prior  permission.   The  University of   Toronto makes  no
 * representations  about  the  suitability of  this software  for any
 * purpose.  It  is    provided   "as is"  without express or  implied
 * warranty.
 *
 * THE UNIVERSITY OF TORONTO DISCLAIMS  ALL WARRANTIES WITH REGARD  TO
 * THIS SOFTWARE,  INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT SHALL THE UNIVERSITY OF TORONTO  BE LIABLE
 * FOR ANY  SPECIAL, INDIRECT OR CONSEQUENTIAL  DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR  PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
 * OUT  OF  OR  IN  CONNECTION WITH  THE  USE  OR PERFORMANCE  OF THIS
 * SOFTWARE.
 *
 **********************************************************************/

#include <stdio.h>
#include <math.h>
#include <values.h>
#include <commands.h>
#include <xerion/minimize.h>
#include <xerion/utilsDCP.h>
#include "ffbp.h"

#ifndef linux
#include <nan.h>
#endif

#define DB(x)	if (Mdebug(currentNet)) { x }

static	Boolean	_presentedAssociate ;
static	Real	_numOutputs ;

struct TRACE    afterExample ;
struct TRACE    forwardEachTime ;
struct TRACE    inCascade ;

static void	forward				ARGS((Unit unit, void *data)) ;
static void	backward			ARGS((Unit unit, void *data)) ;
static void	backwardNoWtDvs			ARGS((Unit unit, void *data)) ;
static void	backwardRecomputeOutputError    ARGS((Unit unit, void *data)) ;

static void	netForward	ARGS((Net	 net)) ;
static void	netBackward	ARGS((Net	 net)) ;

static Proc	updateNetActivity      ARGS((Net	net)) ;
static Proc	updateNetGradient      ARGS((Net	net)) ;

static void	calculateNetErrorDeriv ARGS((Net net, ExampleSet exampleSet)) ;
static void	calculateNetError      ARGS((Net net, ExampleSet exampleSet)) ;

static void	initNet   		ARGS((Net	 net)) ;
static void	deinitNet 		ARGS((Net	 net)) ;
static void	initGroup    		ARGS((Group	 group)) ;
static void	deinitGroup    		ARGS((Group	 group)) ;
static void	initUnit    		ARGS((Unit	 unit)) ;
static void	deinitUnit  		ARGS((Unit	 unit)) ;
static void	initLink    		ARGS((Link	 link)) ;
static void	deinitLink  		ARGS((Link	 link)) ;
static void	initExample	    	ARGS((Example    example)) ;
static void	deinitExample		ARGS((Example    example)) ;
static void	initExampleSet    	ARGS((ExampleSet exampleSet)) ;
static void	deinitExampleSet	ARGS((ExampleSet exampleSet)) ;


/**********************************************************************
 *	Name:		main 
 *	Description:	the main function, used for the xerion simulator
 *	Parameters:	
 *		int	argc	- the number of input args
 *		char	**argv  - array of argument strings from command 
 *				  line
 *	Return Value:	
 *		int	main	- 0
 ***********************************************************************/
int main (argc, argv)
  int	argc ;
  char	**argv ;
{
  /* Insert any private initialization routines here */
  setCreateNetHook(initNet) ;
  setDestroyNetHook(deinitNet) ;
  setCreateGroupHook(initGroup) ;
  setDestroyGroupHook(deinitGroup) ;
  setCreateUnitHook(initUnit) ;
  setDestroyUnitHook(deinitUnit) ;
  setCreateLinkHook(initLink) ;
  setDestroyLinkHook(deinitLink) ;
  setCreateExampleHook(initExample) ;
  setDestroyExampleHook(deinitExample) ;
  setCreateExampleSetHook(initExampleSet) ;
  setDestroyExampleSetHook(deinitExampleSet) ;

  /* Perform initialization of the simulator */
  IStandardInit(&argc, argv);

  /* Enter loop that reads commands and handles graphics */
  ICommandLoop(stdin, stdout, NULL);

  return 0 ;
}

/***********************************************************************/
static void	initNet (net)
  Net	net ;
{
  net->calculateErrorDerivProc = calculateNetErrorDeriv ;
  net->calculateErrorProc      = calculateNetError ;
  net->activityUpdateProc      = updateNetActivity ;
  net->gradientUpdateProc      = updateNetGradient ;

  net->extension = (NetExtension)calloc(1, sizeof(NetExtensionRec)) ;

  /* non-zero defaults */
  MbpToOutputs(net)		= TRUE ;
  MinitializeContext(net)	= TRUE ;
  MlastExample(net)		= -1 ;
  MinitialContextOutput(net)	= 0.2 ;
  MpredictOutput(net)		= TRUE ;
  MweightCostDecrement(net)	= 0.9 ;
  MaveErrorFactor(net)		= 0.95 ;
  MlengthScaleDerivs(net)	= TRUE ;
  Mdt(net)			= 1.0 ;
  MinitializeUnits(net)		= TRUE ;
  MinitialUnitOutput(net)	= 0.2 ;
  MmaxTime(net)			= 10.0 ;
  MuseStoppingCriteria(net)	= TRUE ;
}

static void	deinitNet (net)
  Net	net ;
{
  if (net->extension != NULL)
    free(net->extension) ;
}

/***********************************************************************/
static void	initGroup(group)
  Group	group ;
{
  group->extension = (GroupExtension)calloc(1, sizeof(GroupExtensionRec)) ;

  MminBias(group)		= -MAXFLOAT ;
  MmaxBias(group)		=  MAXFLOAT ;
  MminIncomingWeight(group)	= -MAXFLOAT ;
  MmaxIncomingWeight(group)	=  MAXFLOAT ;
  MminOutgoingWeight(group)	= -MAXFLOAT ;
  MmaxOutgoingWeight(group)	=  MAXFLOAT ;
  MgroupXPosition(group)	= -1 ;
  MgroupYPosition(group)	= -1 ;
  MinputNoise(group)		= -1.0 ;
}

static void	deinitGroup(group)
  Group	group ;
{
  if (group->extension != NULL)
    free(group->extension) ;
}

/***********************************************************************/
static void	initUnit(unit)
  Unit	unit ;
{
  unit->extension = (UnitExtension)calloc(1, sizeof(UnitExtensionRec)) ;
}

static void	deinitUnit(unit)
  Unit	unit ;
{
  if (unit->extension != NULL)
    free(unit->extension) ;
}

/***********************************************************************/
static void	initLink(link)
  Link	link ;
{
  link->extension = (LinkExtension)calloc(1, sizeof(LinkExtensionRec)) ;

  setLinkLength(link, 1.0) ;
}

static void	deinitLink(link)
  Link	link ;
{
  if (link->extension != NULL)
    free(link->extension) ;
}

/***********************************************************************/
static void	initExample(example)
  Example  example ;
{
  int i ;
  example->extension = 
    (ExampleExtension)calloc(1, sizeof(ExampleExtensionRec)) ;

  example->extension->numAssocs = 0 ;
  for (i = 0 ; i < MAX_ASSOCS ; i++)
    example->extension->assoc[i] = NO_ASSOC ;
}

static void	deinitExample(example)
  Example  example ;
{
  if (example->extension != NULL)
    free(example->extension) ;
}

/***********************************************************************/
static void	initExampleSet(exampleSet)
  ExampleSet  exampleSet ;
{
  exampleSet->extension = 
    (ExampleSetExtension)calloc(1, sizeof(ExampleSetExtensionRec)) ;
}

static void	deinitExampleSet(exampleSet)
  ExampleSet  exampleSet ;
{
  if (exampleSet->extension != NULL)
    free(exampleSet->extension) ;
}

void	*getExampleSetData(exampleSet)
  ExampleSet  exampleSet ;
{
  return exampleSet->extension->data ;
}

void	putExampleSetData(exampleSet, data)
  ExampleSet  exampleSet ;
  void	*data ;
{
 exampleSet->extension->data = data ;
}

/*********************************************************************
 * 				Link Length
 *********************************************************************/
void		setLinkLength(link, length)
  Link		link ;
  Real		length ;
{
  Mlength(link) = length ;
}

Real		getLinkLength(link)
  Link		link ;
{
  return Mlength(link) ;
}

int		unitXPosition(unit)
  Unit	unit ;
{
  return (MgroupXPosition(unit->group) >= 0 ? 
	  MgroupXPosition(unit->group) : MxPosition(unit)) ;
}

int		unitYPosition(unit)
  Unit	unit ;
{
  return (MgroupYPosition(unit->group) >= 0 ? 
	  MgroupYPosition(unit->group) : MyPosition(unit)) ;
}

/*********************************************************************
 * 				Lesioning
 *********************************************************************/
Boolean		isLesionedLink(link)
  Link		link ;
{
  return Mlesioned(link) ;
}

void		lesionLink(link)
  Link		link ;
{
  if (Mlesioned(link)) return ;
  link->deriv = 0.0 ;
  MlesionedWeight(link) = link->weight ;
  link->weight = 0.0 ;
  Mlesioned(link) = TRUE ;
  Mcompiled(link->postUnit->net) = FALSE ;
}

void		unlesionLink(link)
  Link		link ;
{
  if (!Mlesioned(link)) return ;
  link->weight = MlesionedWeight(link) ;
  Mlesioned(link) = FALSE ;
  Mcompiled(link->postUnit->net) = FALSE ;
}


Boolean		isLesionedUnit(unit)
  Unit		unit ;
{
  return Mlesioned(unit) ;
}

void		lesionUnit(unit)
  Unit		unit ;
{
  if (Mlesioned(unit)) return ;
  setNaN(unit->output) ;
  Mlesioned(unit) = TRUE ;
  unitForAllLinks(unit, ALL, (LinkDataProc)lesionLink, NULL) ;
  Mcompiled(unit->net) = FALSE ;
}

void		unlesionUnit(unit)
  Unit		unit ;
{
  if (!Mlesioned(unit)) return ;
  Mlesioned(unit) = FALSE ;
  unit->output = 0.0 ;
  unitForAllLinks(unit, ALL, (LinkDataProc)unlesionLink, NULL) ;
  Mcompiled(unit->net) = FALSE ;
}

/***********************************************************************
 *	Name:		unitError
 ***********************************************************************/
static Real	unitError(unit)
  const Unit	unit ;
{
  Net	net = unit->net ;
  Real	error, output = unit->output, target = unit->target ;

  if (isNaN(target) || fabs(output-target) <= MzeroErrorRadius(net))
    return 0.0 ;

  if (MsymmetricSigmoid(net)) { output = SYM2ASYM(output) ; target = SYM2ASYM(target) ; }

  if (MnonExtremalZeroTarget(net) > 0.0 && target == 0.0) {
    if (output <= MnonExtremalZeroTarget(net)) return 0.0 ;
    else target = MnonExtremalZeroTarget(net) ;
  }
  if (MnonExtremalOneTarget(net) > 0.0 && target == 1.0) {
    if (output >= MnonExtremalOneTarget(net)) return 0.0 ;
    else target = MnonExtremalOneTarget(net) ;
  }

  if (fabs(output-target) > McorrectTolerance(net)) MoutputsIncorrect(net)++ ;

  switch (MerrorFunction(net)) {
  case CROSS_ENTROPY:
    output = CLIP(output, MIN_UNIT_OUTPUT, MAX_UNIT_OUTPUT) ;
    error = INFO(output, target) ;
    break ;
  case SQUARED_ERROR:
    error = (output - target) * (output - target) ;
    break ;
  default:
    IErrorAbort("Unknown error function: %d", MerrorFunction(net)) ;
  }      
  if (!MsampleExamples(net)) error *= MexampleFrequency(net) ;
  if (MnormError(net))       error /= (Real)MmaxSlice(net) ;
  return error ;
}

/**********************************************************************/
static Real	unitErrorDeriv(unit)
  const Unit	unit ;
{
  Net	net = unit->net ;
  Real	error, output = unit->output, target = unit->target ;

  if (isNaN(target) || fabs(output-target) <= MzeroErrorRadius(net))
    return 0.0 ;

  if (MsymmetricSigmoid(net)) { output = SYM2ASYM(output) ; target = SYM2ASYM(target) ; }

   if (MnonExtremalZeroTarget(net) > 0.0 && target == 0.0) {
    if (output <= MnonExtremalZeroTarget(net)) return 0.0 ;
    else target = MnonExtremalZeroTarget(net) ;
  }
  if (MnonExtremalOneTarget(net) > 0.0 && target == 1.0) {
    if (output >= MnonExtremalOneTarget(net)) return 0.0 ;
    else target = MnonExtremalOneTarget(net) ;
  }

  switch (MerrorFunction(net)) {
  case CROSS_ENTROPY:
    output = CLIP(output, MIN_UNIT_OUTPUT, MAX_UNIT_OUTPUT) ;
    error = INFO_DERIV(output, target) ; 
    break ;
  case SQUARED_ERROR:
    error = 2.0 * (output - target) ;
    break ;
  default:
    IErrorAbort("Unknown error function: %d", MerrorFunction(net)) ;
  }      
  if (!MsampleExamples(net)) error *= MexampleFrequency(net) ;
  if (MnormError(net))       error /= (Real)MmaxSlice(net) ;
  return error ;
}

/**********************************************************************/
static void	incrementError(unit, data)
  Unit		unit ;
  void		*data ;
{
  *(Real *)data += unitError(unit) ;
}

Real		groupError(group)
  Group group ;
{
  Real	error = 0 ;
  groupForAllUnits(group, (UnitDataProc)incrementError, &error) ;
  return error ;
}

/**********************************************************************/
static void	setOutput(unit, data)
  Unit		unit ;
  void		*data ;
{
  unit->output = *(Real *)data ;
}

/**********************************************************************/
static void	zeroOutputDeriv(unit, data)
  Unit		unit ;
  void		*data ;
{
  unit->outputDeriv = 0.0 ;
}

/**********************************************************************/
static void	injectError(unit, data)
  Unit		unit ;
  void		*data ;
{
  unit->outputDeriv = unitErrorDeriv(unit) ;
}

/**********************************************************************/
static void	incrementCost(unit, data)
  Unit		unit ;
  void		*data ;
{
  Real	output = unit->output ;
  Real	ratio ;
  if (MnoUnitCost(unit->group)) return ;

  if (MsymmetricSigmoid(unit->net)) output = SYM2ASYM(output) ;

  if (MunitPrior(unit->net) < 0.0) {
    /*
     * avoid mean
     */
    ratio = MmeanOutput(unit->group) - output ;	/* temp var */
    *(Real *)data += MunitCost(unit->net) * (1.0 - ratio * ratio) ;
  } else if (MunitPrior(unit->net) > 1.0) {
    /*
     * biased binary
     */
    *(Real *)data += MunitCost(unit->net) * output * (MunitPrior(unit->net) - output) ;
  } else {
    /* 
     * weight-elimination-based measure
     */
    ratio = (output*output)/(MunitPrior(unit->net)*MunitPrior(unit->net)) ;
    *(Real *)data += MunitCost(unit->net) * ratio / (1.0+ratio) ;
    /*
     * CE-based measure
     * output = CLIP(output, MIN_UNIT_OUTPUT, MAX_UNIT_OUTPUT) ;
     * *(Real *)data += MunitCost(unit->net) * INFO(output,MunitPrior(unit->net)) ;
     */
  }
}

/**********************************************************************/
static void	injectUnitCost(unit, data)
  Unit		unit ;
  void		*data ;
{
  Real	output ;
  Real	ratio ;
  if (MnoUnitCost(unit->group)) return ;
  if (MunitPrior(unit->net) < 0.0)
    /* 
     * avoid mean
     */
    unit->outputDeriv += MunitCost(unit->net) * 2.0 * (MmeanOutput(unit->group) - unit->output) ;
  else if (MunitPrior(unit->net) > 1.0)
    /*
     * biased binary
     */
    unit->outputDeriv += MunitCost(unit->net) * (MunitPrior(unit->net) - 2.0 * unit->output) ;
  else {
    /* 
     * weight-elimination-based measure
     */
    ratio = (unit->output*unit->output)/(MunitPrior(unit->net)*MunitPrior(unit->net)) ;
    unit->outputDeriv += MunitCost(unit->net) * 
                         2.0*unit->output/((1.0+ratio)*(1.0+ratio)) ;
    /*
     * CE-based measure
     * output = CLIP(unit->output, MIN_UNIT_OUTPUT, MAX_UNIT_OUTPUT) ;
     * unit->outputDeriv += MunitCost(unit->net) * 
     *                      INFO_DERIV(output,MunitPrior(unit->net)) ;
     */
  }
}

/**********************************************************************/
static void	zeroLinkDerivs(unit, data)
  Unit	unit ;
  void	*data ;
{
  Link	*linkPtr  = unit->incomingLink ;
  Link	*finalPtr = linkPtr + unit->numIncoming ;

  while (linkPtr != finalPtr) {
    Link	link = *(linkPtr++) ;
    link->deriv = 0.0 ;
  }
}

/**********************************************************************/
static void	maybeFixBias(unit, data)
  Unit	unit ;
  void	*data ;
{
  if (MfixedBias(unit->group)) 
    if (unit->incomingLink[0]->preUnit->group->type & BIAS)
      unit->incomingLink[0]->deriv = 0.0 ;
}

/**********************************************************************/
static void	lengthScaleDerivs(unit, data)
  Unit	unit ;
  void	*data ;
{
  Link	*linkPtr  = unit->incomingLink ;
  Link	*finalPtr = linkPtr + unit->numIncoming ;

  while (linkPtr != finalPtr) {
    Link	link   = *(linkPtr++) ;
    Real	length = getLinkLength(link) ;
    if (length > 0.0) link->deriv /= length ;
  }
}

/**********************************************************************/
static void	netAdaptWeightCost(net)
  Net		net ;
{
  MaveError(net) = (MaveError(net)<0.0 ? net->error :
		    net->error + MaveErrorFactor(net)*(MaveError(net)-net->error)) ;
  if (net->error < MlastError(net)	||
      net->error < MdesiredError(net)	||
      MlastError(net) < 0.0) 
    MweightCost(net) += MweightCostIncrement(net) ;
  else if (net->error < MaveError(net))
    MweightCost(net) -= MweightCostIncrement(net) ;
  else
    MweightCost(net) *= MweightCostDecrement(net) ;
  if (MweightCost(net) < 0.0) MweightCost(net) = 0.0 ;
  MlastError(net) = net->error ;
}

/**********************************************************************/
static void	unitDecayWts(unit, data)
  Unit		unit ;
  void		*data ;
{
  int	idx ;
  Net	net 		= unit->net ;
  int	numIncoming 	= unit->numIncoming ;
  Link	*incoming   	= unit->incomingLink ;
  Real	weightCost	= MweightCost(net) ;
  Real	wtCost ;
  Boolean lengthScaleCost = (! MlengthScaleDerivs(net)) ;

  if (MnoWeightCost(unit->group)) return ;

  for (idx = 0 ; idx < numIncoming ; ++idx) {
    Link link    = incoming[idx] ;
    if (MnoWeightCost(link->preUnit->group)) continue ;
    wtCost  = link->weight * weightCost ;
    if (lengthScaleCost) wtCost *= getLinkLength(link) ;
    net->cost   += wtCost * link->weight ;
    link->deriv += 2.0 * wtCost ;
  }
}

/**********************************************************************/
static void	unitMeltWts(unit, data)
  Unit		unit ;
  void		*data ;
{
  int	idx ;
  Net	net 		= unit->net ;
  int	numIncoming 	= unit->numIncoming ;
  Link	*incoming   	= unit->incomingLink ;
  Real	weightCost	= MweightCost(net) ;
  Real	Wo	   	= MstandardWeight(net) ;
  Real	wt, ratio ;
  Boolean lengthScaleCost = (! MlengthScaleDerivs(net)) ;

  if (MnoWeightCost(unit->group)) return ;

  for (idx = 0 ; idx < numIncoming ; ++idx) {
    Link link    = incoming[idx] ;
    if (MnoWeightCost(link->preUnit->group)) continue ;
    wt	 = link->weight ;
    ratio   = (wt * wt) / (Wo * Wo) ;
    if (lengthScaleCost) weightCost *= getLinkLength(link) ;
    net->cost   += weightCost * ratio / (1.0 + ratio) ;
    link->deriv += weightCost * 2.0 * wt / ((1.0 + ratio) * (1.0 + ratio)) ;
  }
}

/**********************************************************************/
static 	void	netUpdateWeightCost(net)
  Net	net ;
{
  if (MweightCost(net) == 0.0) return ;
  if (MstandardWeight(net) == 0.0) 
    netForAllUnits(net, ALL, (UnitDataProc)unitDecayWts, NULL) ;
  else
    netForAllUnits(net, ALL, (UnitDataProc)unitMeltWts, NULL) ;
}

/**********************************************************************/
static void	unitCompile(unit, data)
  Unit		unit ;
  void		*data ;
{
  int	idx, cdx, numIncoming = unit->numIncoming ;

  if (Mlesioned(unit)) {
    MnumCompiledIncoming(unit) = 0 ;
    return ;
  }
  if (!MpreOutputs(unit)) {
    MpreOutputs(unit)	   = ((Real **)calloc(numIncoming, sizeof(Real*))) ;
    MpreOutputDerivs(unit) = ((Real **)calloc(numIncoming, sizeof(Real*))) ;
    Mweights(unit)    	   = ((Real *) calloc(numIncoming, sizeof(Real))) ;
    Mderivs(unit)     	   = ((Real *) calloc(numIncoming, sizeof(Real))) ;
    McompiledLinks(unit)   = ((Link *) calloc(numIncoming, sizeof(Link))) ;
  }

  for (idx=0, cdx=0 ; idx < numIncoming ; idx++) {
    Link link = unit->incomingLink[idx] ;
    if (! (Mlesioned(link) || Mlesioned(link->preUnit))) {
      MpreOutputs(unit)[cdx] 	  = &(link->preUnit->output) ;
      MpreOutputDerivs(unit)[cdx] = &(link->preUnit->outputDeriv) ;
      Mweights(unit)[cdx]	  =   link->weight ;
      McompiledLinks(unit)[cdx]   =   link ;
      cdx++ ;
    }
  }
  MnumCompiledIncoming(unit) = cdx ; 
}

/***********************************************************************/
static void	netCompile(net)
  Net		net ;
{
  netForAllOtherUnits(net, INPUT|BIAS, (UnitDataProc)unitCompile, NULL) ;
  Mcompiled(net) = TRUE ;
}

/**********************************************************************/
static void	linkSumWt(link, data)
  Link		link ;
  void		*data ;
{
  if (link->preUnit->group->type&BIAS) return ;
  *(Real *)data += link->weight ;
}

/**********************************************************************/
static void	linkSetWt(link, data)
  Link		link ;
  void		*data ;
{
  if (link->preUnit->group->type&BIAS) return ;
  link->weight = *(Real *)data ;
}

/**********************************************************************/
static void	linkMinWt(link, data)
  Link		link ;
  void		*data ;
{
  if (link->preUnit->group->type&BIAS) return ;
  link->weight = MAX(link->weight, *(Real *)data) ;
}

/**********************************************************************/
static void	linkMaxWt(link, data)
  Link		link ;
  void		*data ;
{
  if (link->preUnit->group->type&BIAS) return ;
  link->weight = MIN(link->weight, *(Real *)data) ;

}
/**********************************************************************/
static void	unitConstrainWeights(unit, data)
  Unit	unit ;
  void	*data ;
{
  /* CONSTRAIN BIAS */
  if (unit->numIncoming > 0) {
    if (MmaxBias(unit->group)<MAXFLOAT)
      unit->incomingLink[0]->weight = 
	MIN(unit->incomingLink[0]->weight,MmaxBias(unit->group)) ;
    if (MminBias(unit->group)>-MAXFLOAT)
      unit->incomingLink[0]->weight = 
	MAX(unit->incomingLink[0]->weight,MminBias(unit->group)) ;
  }

  /* INCOMING WEIGHTS */
  if (unit->numIncoming > 1) {
    if (MequateIncomingWeights(unit)) {
      Real	weight = 0.0 ;
      unitForAllLinks(unit, INCOMING, (LinkDataProc)linkSumWt, (void *)&weight) ;
      weight /= (Real)(unit->numIncoming-1) ;
      unitForAllLinks(unit, INCOMING, (LinkDataProc)linkSetWt, (void *)&weight) ;
    }
    if (MmaxIncomingWeight(unit->group)<MAXFLOAT)
      unitForAllLinks(unit, INCOMING, (LinkDataProc)linkMaxWt, (void *)&MmaxIncomingWeight(unit->group)) ;
    if (MminIncomingWeight(unit->group)>-MAXFLOAT)
      unitForAllLinks(unit, INCOMING, (LinkDataProc)linkMinWt, (void *)&MminIncomingWeight(unit->group)) ;
  }

  /* OUTGOING WEIGHTS */
  if (unit->numOutgoing > 0) {
    if (MequateOutgoingWeights(unit)) {
      Real	weight = 0.0 ;
      unitForAllLinks(unit, OUTGOING, (LinkDataProc)linkSumWt, (void *)&weight) ;
      weight /= (Real)(unit->numOutgoing) ;
      unitForAllLinks(unit, OUTGOING, (LinkDataProc)linkSetWt, (void *)&weight) ;
    }
    if (MmaxOutgoingWeight(unit->group)<MAXFLOAT)
      unitForAllLinks(unit, OUTGOING, (LinkDataProc)linkMaxWt,(void *)&MmaxOutgoingWeight(unit->group)) ;
    if (MminOutgoingWeight(unit->group)>-MAXFLOAT)
      unitForAllLinks(unit, OUTGOING, (LinkDataProc)linkMinWt,(void *)&MminOutgoingWeight(unit->group)) ;
  }
}

/***********************************************************************/
int	command_applyWeightConstraints(tokc, tokv)
  int	tokc ;
  char	**tokv ;
{

  IUsage("");
  if (GiveHelp(tokc)) {
    ISynopsis("Apply constraints on weights and biases.");
    IHelp
      (IHelpArgs,
       "Apply constraints on weights and biases.",
       "EXAMPLES",
       "SEE ALSO",
       NULL) ;
    return 1 ;
  }

  if (currentNet == NULL)
    IErrorAbort("There is no current net.") ;

  netForAllUnits(currentNet, ALL, (UnitDataProc)unitConstrainWeights, NULL) ;

  return 1 ;
}

/**********************************************************************/
static void	getCompiledWeights(unit, data)
  Unit	unit ;
  void	*data ;
{
  int	idx, cdx ;
  int	numIncoming	= MnumCompiledIncoming(unit) ;
  Real	*wt, *dv;
  Link	*ln ;

  for (idx=0, wt=Mweights(unit), dv=Mderivs(unit), ln=McompiledLinks(unit) ;
       idx < numIncoming ; 
       idx++) {
    (*wt++) = (*ln++)->weight ;
    (*dv++) = 0.0 ;
  }
}

/**********************************************************************/
static void	putCompiledDerivs(unit, data)
  Unit	unit ;
  void	*data ;
{
  int	idx ;
  int	numIncoming = MnumCompiledIncoming(unit) ;
  Real	*dv ;
  Link	*ln ;

  for (idx=0, dv=Mderivs(unit), ln=McompiledLinks(unit) ;
       idx < numIncoming ; 
       idx++) 
    (*ln++)->deriv  = (*dv++) ;
}
/***********************************************************************
 *	Name:		command_compile
 ***********************************************************************/
int	command_compile(tokc, tokv)
  int	tokc ;
  char	**tokv ;
{

  IUsage("");
  if (GiveHelp(tokc)) {
    ISynopsis("Compiles network");
    IHelp
      (IHelpArgs,
       "Compiles network.",
       "EXAMPLES",
       "SEE ALSO",
       NULL) ;
    return 1 ;
  }

  if (currentNet == NULL)
    IErrorAbort("There is no current net.") ;

  netCompile(currentNet) ;

  return 1 ;
}

/***********************************************************************/
static	void	setWeight(link, data)
  Link	link ;
  void	*data ;
{
  link->weight = *(Real *)data ;
}

/**********************************************************************/
static void	groupSetInhibWeight(group, data)
  Group		group ;
  void		*data ;
{
  if (group->numUnits == 1 &&
      group->unit[0]->numIncoming > 1 &&
      MequateOutgoingWeights(group->unit[0]) &&
      MmaxOutgoingWeight(group) <= 0.0) {
    Unit	unit = group->unit[0] ;
    Real	bias = unit->incomingLink[0]->weight ;

    unitForAllLinks(unit, OUTGOING, (LinkDataProc)setWeight, (void *)&bias) ;
  }
}

/**********************************************************************/
static void	copyOutputsToTargets(from, to)
  Group		from, to ;
{
  int i ;
  Boolean rec = from->net->type & RECURRENT ;
  
  if (from->numUnits != to->numUnits)
    IErrorAbort("must have same numbers of units (%d -> %d)", from->numUnits, to->numUnits) ;

  for (i = 0 ; i < from->numUnits ; i++ ) {
    Real out = from->unit[i]->output ;
    to->unit[i]->target = out ;
    if (rec) to->unit[i]->targetHistory[from->net->currentTime] = out ;
  }
}

/**********************************************************************/
static void	copyTargetsToOutputs(from, to)
  Group		from, to ;
{
  int i ;
  
  if (from->numUnits != to->numUnits)
    IErrorAbort("must have same numbers of units (%d -> %d)", from->numUnits, to->numUnits) ;

  for (i = 0 ; i < from->numUnits ; i++ ) 
    to->unit[i]->output = from->unit[i]->target ;
}

/***********************************************************************/
int	command_setInhibWeights(tokc, tokv)
  int	tokc ;
  char	**tokv ;
{

  IUsage("");
  if (GiveHelp(tokc)) {
    ISynopsis("Sets outgoing weights of dedicated inhibitory units to their bias.") ;
    IHelp
      (IHelpArgs,
       "Sets outgoing weights of dedicated inhibitory units to their bias." ,
       "EXAMPLES",
       "SEE ALSO",
       NULL) ;
    return 1 ;
  }

  if (currentNet == NULL)
    IErrorAbort("There is no current net.") ;

  netForAllOtherGroups(currentNet, BIAS|INPUT, (GroupDataProc)groupSetInhibWeight, NULL) ;

  return 1 ;
}

/*********************************************************************
 * 				Forward models
 *********************************************************************/
#define printOutputs(net)\
	{ netForAllUnits(net, OUTPUT, (UnitDataProc)printOutput, NULL) ; \
	  fprintf(dout, "\n") ; }

static void	printOutput(unit, data)
  Unit		unit ;
  void		*data ;
{
  fprintf(dout,"%.4f ", unit->output) ;
}

#define printTargets(net)\
	{ netForAllUnits(net, OUTPUT, (UnitDataProc)printTarget, NULL) ; \
	  fprintf(dout, "\n") ; }

static void	printTarget(unit, data)
  Unit		unit ;
  void		*data ;
{
  fprintf(dout,"%.4f ", unit->target) ;

}

#define printDerivs(net,g)\
	{ groupForAllUnits(net->group[g], (UnitDataProc)printDeriv, NULL) ; \
	  fprintf(dout, "\n") ; }

static void	printDeriv(unit, data)
  Unit		unit ;
  void		*data ;
{
  fprintf(dout,"%.4f ", unit->outputDeriv) ;
}

static int	findFirstInputGroupNum(net)

  Net	net ;
{
  int	g ;
  Group	group ;

  for (g = 0, group=net->group[g] ; g < net->numGroups ; group=net->group[++g])
    if (group->type&INPUT) return g ;
  return 0 ;
}

static int	findLastOutputGroupNum(net)
  Net	net ;
{
  int	g ;
  Group	group ;

  for (g = net->numGroups-1, group=net->group[g] ; g > 0 ; group=net->group[--g])
    if (group->type&OUTPUT) return g ;
  return 0 ;
}

/**********************************************************************/
static void	initContextOutput(unit, data)
  Unit		unit ;
  void		*data ;
{
  if (McopyFromLayer(unit->group))
    unit->output = MinitialContextOutput(unit->net) ;
}

/**********************************************************************/
static	void	addInputNoise(unit, data)
  Unit	unit ;
  void	*data ;
{
  Real	noise, output ;
  Real  value = *(Real *)data ;
  Real	groupNoise = MinputNoise(unit->group) ;

  if (groupNoise == 0.0) return ;
  if (groupNoise > 0.0) value = groupNoise ;

/*
  noise = uniformNoise(-value, -value) ;
*/
  noise = gaussianNoise(value) ;
  output = unit->output ;

  if (MsymmetricSigmoid(unit->net)) output = SYM2ASYM(output) ;

  if      (output == 0.0) output += fabs(noise) ;
  else if (output == 1.0) output -= fabs(noise) ;
  else output = CLIP(output+noise, 0.0, 1.0) ;

  if (MsymmetricSigmoid(unit->net)) output = ASYM2SYM(output) ;
  unit->output = output ;
}

/**********************************************************************/
static	void	addTargetNoise(unit, data)
  Unit	unit ;
  void	*data ;
{
/*
  Real	noise = uniformNoise(-*(Real *)data, *(Real *)data) ;
*/
  Real	noise = gaussianNoise(*(Real *)data) ;
  Real	target = unit->target ;
  if (MsymmetricSigmoid(unit->net)) target = SYM2ASYM(target) ;

  if      (target == 0.0) target += fabs(noise) ;
  else if (target == 1.0) target -= fabs(noise) ;
  else target = CLIP(target+noise, 0.0, 1.0) ;
  if (MsymmetricSigmoid(unit->net)) target = ASYM2SYM(target) ;
  unit->target = target ;
  if (unit->net->type & RECURRENT)
    unit->targetHistory[unit->net->currentTime] = unit->target ;
}

/**********************************************************************/
static	void	scaleInput(unit, data)
  Unit	unit ;
  void	*data ;
{
  unit->output *= *(Real *)data ;
}

/**********************************************************************/
static void	clampInput(unit, slice)
  Unit		unit ;
  void		*slice ;
{
  Net net = unit->net ;

  if (net->type&RECURRENT)
    unit->output = unit->clampingHistory[(int)slice] ;
  if (MinputScaleFactor(net)>0.0) scaleInput(unit, (void *)&MinputScaleFactor(net)) ;
  if (MinputNoise(net)>0.0)       addInputNoise(unit, (void *)&MinputNoise(net)) ;
}

/**********************************************************************/
static void	setTarget(unit, slice)
  Unit		unit ;
  void		*slice ;
{
  unit->target = unit->targetHistory[(int)slice] ;
}

/**********************************************************************/
static void	saveTarget(unit, slice)
  Unit		unit ;
  void		*slice ;
{
  unit->targetHistory[(int)slice] = unit->target ;
}

/**********************************************************************/
static void	saveOutputToSlice(unit, slice)
  Unit		unit ;
  void		*slice ;
{
  unit->outputHistory[(int)slice] = unit->output ;
}

/**********************************************************************/
static void	saveTargetToClamping(unit, slice)
  Unit		unit ;
  void		*slice ;
{
  unit->clampingHistory[(int)slice] = unit->target ;
}

/**********************************************************************/
static void	restoreOutputFromSlice(unit, slice)
  Unit		unit ;
  void		*slice ;
{
  unit->output = unit->outputHistory[(int)slice] ;
}

/**********************************************************************/
static void	groupZeroMeanOutput(group, data)
  Group		group ;
  void		*data ;
{
  MmeanOutput(group) = 0.0 ;
}

/**********************************************************************/
static void	groupNormMeanOutput(group, data)
  Group		group ;
  void		*data ;
{
  MmeanOutput(group) /= (Real)(group->numUnits) ;
}

/**********************************************************************
 *			marcus
 **********************************************************************/
#define MARCUS_N_FEATURES 15
#define MARCUS_N_PHONEMES 34
#define MARCUS_LAST_VOWEL 10
static int	marcus_features[MARCUS_N_PHONEMES][MARCUS_N_FEATURES] = {
{ 1,  -1,  -1,   1,   1,   1,   1,   1,   1,  -1,  -1,  -1,  -1,  -1,   1}, /* @  0 */
{-1,  -1,  -1,   1,   1,   1,   1,   1,   1,  -1,  -1,  -1,  -1,  -1,   1}, /* a  1 */
{-1,   1,  -1,   1,   1,   1,   1,   1,   1,  -1,  -1,  -1,  -1,  -1,   1}, /* ^  2 */
{ 1,  -1,  -1,  -1,   1,   1,   1,   1,   1,  -1,  -1,  -1,  -1,  -1,   1}, /* e  3 */
{-1,  -1,  -1,  -1,   1,   1,   1,   1,   1,  -1,  -1,  -1,  -1,  -1,   1}, /* 6  4 */
{-1,   1,  -1,  -1,   1,   1,   1,   1,   1,  -1,  -1,  -1,  -1,  -1,   1}, /* o  5 */
{ 1,  -1,  -1,  -1,  -1,   1,   1,   1,   1,  -1,  -1,  -1,  -1,  -1,   1}, /* I  6 */
{ 1,  -1,   1,  -1,  -1,   1,   1,   1,   1,  -1,  -1,  -1,  -1,  -1,   1}, /* i  7 */
{-1,   1,  -1,  -1,  -1,   1,   1,   1,   1,  -1,  -1,  -1,  -1,  -1,   1}, /* U  8 */
{-1,   1,   1,  -1,  -1,   1,   1,   1,   1,  -1,  -1,  -1,  -1,  -1,   1}, /* u  9 */
{-1,  -1,  -1,  -1,  -1,  -1,   1,   1,   1,  -1,  -1,  -1,  -1,  -1,   1}, /* j 10 */
{-1,   1,  -1,  -1,  -1,  -1,   1,   1,   1,  -1,  -1,  -1,  -1,  -1,   1}, /* w 11 */
{-1,  -1,  -1,  -1,  -1,  -1,   1,   1,   1,  -1,  -1,  -1,  -1,  -1,  -1}, /* h 12 */
{-1,  -1,  -1,  -1,  -1,  -1,   1,   1,   1,  -1,   1,  -1,  -1,  -1,   1}, /* r 13 */
{-1,  -1,  -1,  -1,  -1,  -1,  -1,   1,   1,  -1,   1,  -1,  -1,  -1,   1}, /* l 14 */
{-1,  -1,  -1,  -1,  -1,  -1,  -1,   1,   1,   1,  -1,  -1,   1,  -1,   1}, /* m 15 */
{-1,  -1,  -1,  -1,  -1,  -1,  -1,   1,   1,  -1,   1,  -1,   1,  -1,   1}, /* n 16 */
{-1,  -1,  -1,  -1,  -1,  -1,  -1,   1,   1,  -1,  -1,   1,   1,  -1,   1}, /* N 17 */
{-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,   1,   1,  -1,  -1,  -1,  -1,  -1}, /* f 18 */
{-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,   1,   1,  -1,  -1,  -1,  -1,   1}, /* v 19 */
{-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,   1,  -1,   1,  -1,  -1,   1,  -1}, /* s 20 */
{-1,  -1,  -1,  -1,  -1,   1,  -1,  -1,   1,  -1,   1,  -1,  -1,   1,  -1}, /* S 21 */
{-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,   1,  -1,   1,  -1,  -1,   1,   1}, /* z 22 */
{-1,  -1,  -1,  -1,  -1,   1,  -1,  -1,   1,  -1,   1,  -1,  -1,   1,   1}, /* Z 23 */
{-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,   1,  -1,   1,  -1,  -1,  -1,  -1}, /* T 24 */
{-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,   1,  -1,   1,  -1,  -1,  -1,   1}, /* D 25 */
{-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,   1,  -1,  -1,  -1,  -1,  -1}, /* p 26 */
{-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,   1,  -1,  -1,  -1,  -1,   1}, /* b 27 */
{-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,   1,  -1,  -1,  -1,  -1}, /* t 28 */
{-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,   1,  -1,  -1,   1,  -1}, /* C 29 */
{-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,   1,  -1,  -1,  -1,   1}, /* d 30 */
{-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,   1,  -1,  -1,   1,   1}, /* J 31 */
{-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,   1,  -1,  -1,  -1}, /* k 32 */
{-1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,   1,  -1,  -1,   1}  /* g 33 */
} ;

static int _marcus_last_C = -1 ;
static int _marcus_last_V = -1 ;

/* set targets on MARCUS_NEXT group */
static void	marcusPickNextSyllable (net)
  Net		net ;
{
  Group	next = net->group[MARCUS_NEXT] ;
  int	u, f ;

  if (Mmarcus(net) < 0) {
    int nFeatures = -Mmarcus(net) ;
    if (! WITH_PROBABILITY(MmarcusRepeatProb(net))) {		/* maybe keep last CV */
      if (WITH_PROBABILITY(MmarcusSilenceProb(net)))		/* silence */
	_marcus_last_C = -1 ;			
      else {
	int max = 1<<nFeatures ;
	_marcus_last_C = randomInteger(max) ;
      }
    }
    for (f = 0 ; f < nFeatures ; f++)
      next->unit[f]->target = (_marcus_last_C == -1 ? 
			       0.0 :
			       (_marcus_last_C / (1<<f)) % 2) ;
    return ;
  }
  if (next->numUnits != 2 * MARCUS_N_FEATURES)
    IErrorAbort("MARCUS_NEXT group does not have 2*nfeatures (%d) units (%d)",
		2*MARCUS_N_FEATURES, next->numUnits) ;
  if (! WITH_PROBABILITY(MmarcusRepeatProb(net))) {		/* maybe keep last CV */
    if (WITH_PROBABILITY(MmarcusSilenceProb(net)))		/* silence */
      _marcus_last_C = _marcus_last_V = -1 ;			
    else {
      _marcus_last_V = randomInteger(MARCUS_LAST_VOWEL) ;
      _marcus_last_C = MARCUS_LAST_VOWEL + randomInteger(MARCUS_N_PHONEMES-MARCUS_LAST_VOWEL) ;
    }
  }
  for (f = 0 ; f < 2*MARCUS_N_FEATURES ; f++)				/* silence */
    next->unit[f]->target = (_marcus_last_C == -1 ? 
			     0.0 :
			     (f < MARCUS_N_FEATURES ? 
			      marcus_features[_marcus_last_C][f] :
			      marcus_features[_marcus_last_V][f-MARCUS_N_FEATURES])) ;
}
/**********************************************************************
 *			netDoGroups/netDoUnits
 **********************************************************************/
static void	netDoGroups (net, first, last, mask, test, proc, data)
  Net		net ;
  int		first, last ;
  Mask		mask ;
  Boolean	test ;
  GroupDataProc	proc ;
  void		*data ;
{
  Group *groupPtr = net->group + first ;
  int	groupInc  = (first < last ? 1 : -1) ;
  Group *finalPtr = net->group + last + groupInc ;

  while (groupPtr != finalPtr) {
    Group	group = (*groupPtr) ;
    groupPtr += groupInc ;
    if (( test &&  (group->type & mask)) ||
	(!test && !(group->type & mask)))
      (*proc) (group, data) ;
  }
}

/**********************************************************************/
static void	netDoUnits (net, first, last, mask, test, proc, data)
  Net		net ;
  int		first, last ;
  Mask		mask ;
  Boolean	test ;
  UnitDataProc	proc ;
  void		*data ;
{
  Group *groupPtr = net->group + first ;
  int	groupInc  = (first < last ? 1 : -1) ;
  Group *finalPtr = net->group + last + groupInc ;

  while (groupPtr != finalPtr) {
    Group	group = (*groupPtr) ;
    groupPtr += groupInc ;
    if (( test &&  (group->type & mask)) ||
	(!test && !(group->type & mask))) {
      groupForAllUnits(group, proc, data) ;
    }
  }
}

/***********************************************************************
 *				netForward
 ***********************************************************************/
static	void	forwardTargExtInput(unit, data)
  Unit	unit ;
  void	*data ;
{
  Net	net	= unit->net ;
  Real	freq, factor, extIn ;
  Real	dt		= Mdt(unit->net) ;
  int	numIncoming 	= MnumCompiledIncoming(unit) ;
  Real	totalInput  	= 0.0 ;
  Real	*wt, **preOutput ;
  int	idx ;
  Real	output = unit->output ;
  if (MsymmetricSigmoid(net)) output = SYM2ASYM(output) ;

  if (Mlesioned(unit)) return ;
  if (unit->group->type&OUTPUT && !isNaN(unit->target)) {
    if (MtargExtInConstant(net)==0.0) {
      extIn = *(Real *)data ;
    } else {	
      switch (MfreqFunction(net)) {
      case 0: 	/* raw */
	freq = MexampleFrequency(net)*69973 ; 
	factor = net->mz->nIterations*LOG(freq) ;
	break ;
      case 1:		/* sqrt */
	freq = MexampleFrequency(net)*264.524 ;
	factor = net->mz->nIterations*LOG(freq*freq) ;
	break ;
      case 2:		/* log */
	factor = net->mz->nIterations*MexampleFrequency(net) ;
	break ;
      default:
	IErrorAbort("Unknown freqFunction %d", MfreqFunction(net)) ;
      }
      extIn = *(Real *)data * factor/(factor+MtargExtInConstant(net)) ;
    }
    totalInput += (unit->target > 0.5 ? extIn : -extIn) ;
  }
  for (idx=0, wt=Mweights(unit), preOutput=MpreOutputs(unit) ;
       idx < numIncoming ; 
       idx++) 
    totalInput += (*wt++) * *(*preOutput++) ;

  if (dt < 1.0) {
    Real change = sigmoid(totalInput) - output ;
    output += dt * change ;
    if (unit->group->type&OUTPUT) {
      change = fabs(change) ;
      if (change > MoutputChange(net)) MoutputChange(net) = change ;
    }
  } else
    output = sigmoid(totalInput) ;
  if (MsymmetricSigmoid(net)) output = ASYM2SYM(output) ;
  unit->output = output ;
}

/**********************************************************************/
static void	PMSPtargExtInRange (net)
  Net		net ;
{
  Real factor ;
  factor = net->mz->nIterations*11.1559 ;	/* log(69973) */
  MtargExtInHi(net) = MtargetExternalInput(net) * factor/(factor+MtargExtInConstant(net)) ;
  factor = net->mz->nIterations*0.693147 ;	/* log(2) */
  MtargExtInLo(net) = MtargetExternalInput(net) * factor/(factor+MtargExtInConstant(net)) ;
}

/**********************************************************************/
static void	netForward (net)
  Net		net ;
{
  if (!(net->type&RECURRENT))
    netForAllUnits(net, INPUT,  (UnitDataProc)clampInput, (void *)0) ;
/****** MOVED TO clampInput
  if (MinputScaleFactor(net)>0.0)
    netForAllUnits (net, INPUT, (UnitDataProc)scaleInput, &MinputScaleFactor(net)) ;
  if (MinputNoise(net)>0.0)
    netForAllUnits (net, INPUT, (UnitDataProc)addInputNoise, &MinputNoise(net)) ;
*******/
  netForAllOtherGroups(net, INPUT|BIAS, (GroupDataProc)groupZeroMeanOutput, NULL) ;
  if (MtargetExternalInput(net) > 0.0) {
    netForAllOtherUnits (net, INPUT|BIAS, (UnitDataProc)forwardTargExtInput, &MtargetExternalInput(net)) ;
    PMSPtargExtInRange(net) ;
  } else {
    if (Mdt(net) < 1.0) {
      Real time ;
      if (MinitializeUnits(net))
	netForAllOtherUnits (net, INPUT|BIAS, (UnitDataProc)setOutput, (void *)&MinitialUnitOutput(net)) ;
      CASCADE(net,time) {
	netForAllOtherUnits (net, INPUT|BIAS, (UnitDataProc)forward, NULL) ;
	if (!Mrunning(net)) DISPLAY_CASCADE(net) ;
      }
      MaveTime(net) += time ;
    } else
      netForAllOtherUnits (net, INPUT|BIAS, (UnitDataProc)forward, NULL) ;
  }
  netForAllOtherGroups(net, INPUT|BIAS, (GroupDataProc)groupNormMeanOutput, NULL) ;
  MoutputsIncorrect(net) = 0 ;
  netForAllUnits     (net, OUTPUT, (UnitDataProc)incrementError, &(net->error)) ;
  if (MunitCost(net) != 0.0)
    netForAllOtherUnits(net, INPUT|OUTPUT|BIAS, (UnitDataProc)incrementCost, &(net->cost)) ;
}

/***********************************************************************/
static void	netForwardLayers (net, first, last, slice, training)
  Net		net ;
  int		first, last, slice ;
  Boolean	training ;
{
  if (Mdt(net) < 1.0) {
    Real time ;
      if (MinitializeUnits(net))
	netDoUnits(net,first,last,INPUT|BIAS,FALSE,(UnitDataProc)setOutput, (void *)&MinitialUnitOutput(net)) ;
    CASCADE(net,time) {
      netDoUnits(net,first,last,INPUT|BIAS,FALSE,(UnitDataProc)forward, NULL) ;
      if (!Mrunning(net)) DISPLAY_CASCADE(net) ;
    }
    MaveTime(net) += time ;
  } else
    netDoUnits(net,first,last,INPUT|BIAS,FALSE,(UnitDataProc)forward,       NULL) ;
  netDoUnits(net,first,last,OUTPUT,    TRUE, (UnitDataProc)setTarget,     (void *)slice) ;
  netDoUnits(net,first,last,OUTPUT,    TRUE, (UnitDataProc)incrementError,(void *)&(net->error)) ;
  if (training) {
    netDoUnits(net,first,last,INPUT|OUTPUT|BIAS, FALSE,(UnitDataProc)zeroOutputDeriv, NULL) ;
    netDoUnits(net,first,last,OUTPUT,            TRUE, (UnitDataProc)injectError, NULL) ;
    if (MunitCost(net) != 0.0)
      netDoUnits(net,first,last,INPUT|OUTPUT|BIAS,FALSE,(UnitDataProc)injectUnitCost, NULL) ;
  }
}

/**********************************************************************/
static	void	forward(unit, data)
  Unit	unit ;
  void	*data ;
{
  Net	net		= unit->net ;
  int	numIncoming 	= MnumCompiledIncoming(unit) ;
  Real	dt		= Mdt(net) ;
  Real	totalInput  	= 0.0 ;
  Real	*wt, **preOut ;
  int	idx ;
  Real	output = unit->output ;
  if (MsymmetricSigmoid(net)) output = SYM2ASYM(output) ;

  if (Mlesioned(unit)) return ;
  if (MunitNoise(net))
    totalInput = gaussianNoise(MunitNoise(net)) ;

  for (idx=0, wt=Mweights(unit), preOut=MpreOutputs(unit) ;
       idx < numIncoming ; 
       idx++)
    totalInput += (*wt++) * *(*preOut++) ;

  if (dt < 1.0) {
    Real change = sigmoid(totalInput) - output ;
    output += dt * change ;
    if (unit->group->type&OUTPUT) {
      change = fabs(change) ;
      if (MoutputChangeUseAve(net)) {
	MoutputChange(net) += change ;   _numOutputs++ ;
      } else if (change > MoutputChange(net)) {
	MoutputChange(net)  = change ;
      }
    }
  } else
    output = sigmoid(totalInput) ;
  if (MsymmetricSigmoid(net)) output = ASYM2SYM(output) ;
  unit->output = output ;
  MmeanOutput(unit->group) += output ;
}

/***********************************************************************
 *				netBackward
 ***********************************************************************/
static void	netBackward (net)
  Net		net ;
{
  netForAllOtherUnits(net, INPUT|OUTPUT|BIAS,(UnitDataProc)zeroOutputDeriv,NULL) ;
  if (MunitCost(net) != 0.0)
    netForAllOtherUnits(net, INPUT|OUTPUT|BIAS,  (UnitDataProc)injectUnitCost, NULL) ;
  if (MbpToOutputs(net)) {
    netForAllUnits          (net, OUTPUT,     (UnitDataProc)injectError, NULL) ;
    netForAllOtherUnitsBack (net, INPUT|BIAS, (UnitDataProc)backward,	 NULL) ;
  } else
    netForAllOtherUnitsBack (net, INPUT|BIAS, (UnitDataProc)backwardRecomputeOutputError, NULL) ;
}

/**********************************************************************/
static	void	backward(unit, data)
  Unit	unit ;
  void	*data ;
{
  int	numIncoming = MnumCompiledIncoming(unit) ;
  Real	inputDeriv ;
  Real 	*wt, *dv, **preOut, **preDv ;
  int	idx ;
  Real	output = unit->output ;
  if (MsymmetricSigmoid(unit->net)) output = SYM2ASYM(output) ;

  if (Mlesioned(unit)) return ;
  inputDeriv = sigmoidDeriv(output) * unit->outputDeriv ;

  for (idx=0, wt=Mweights(unit), dv=Mderivs(unit), preOut=MpreOutputs(unit),
       preDv=MpreOutputDerivs(unit) ;
       idx < numIncoming ; 
       idx++) {
    *(*preDv++) += inputDeriv *  (*wt++) ;
    (*dv++)    += inputDeriv * *(*preOut++) ;
  }
}

/**********************************************************************/
static	void	backwardNoWtDvs(unit, data)
  Unit	unit ;
  void	*data ;
{
  int	numIncoming = MnumCompiledIncoming(unit) ;
  Real	inputDeriv ;
  Real 	*wt, **preDv ;
  int	idx ;
  Real	output = unit->output ;
  if (MsymmetricSigmoid(unit->net)) output = SYM2ASYM(output) ;

  if (Mlesioned(unit)) return ;
  inputDeriv = sigmoidDeriv(output) * unit->outputDeriv ;

  for (idx=0, wt=Mweights(unit), preDv=MpreOutputDerivs(unit) ;
       idx < numIncoming ; 
       idx++) {
    *(*preDv++) += inputDeriv *  (*wt++) ;
  }
}

/***********************************************************************/
static	void	backwardRecomputeOutputError(unit, data)
  Unit	unit ;
  void	*data ;
{
  /*
   *  only called if MbpToOutputs(net) = FALSE
   */
  int	numIncoming = MnumCompiledIncoming(unit) ;
  Real	inputDeriv ;
  Real 	*wt, *dv, **preOut, **preDv ;
  int	idx ;
  Real	output = unit->output ;
  if (MsymmetricSigmoid(unit->net)) output = SYM2ASYM(output) ;

  if (Mlesioned(unit)) return ;

  /*
   * Calculate error derivatives for an output unit
   * (overwriting any backpropagated derivatives)
   */
  if (unit->group->type&OUTPUT) unit->outputDeriv = unitErrorDeriv(unit) ;

  inputDeriv = sigmoidDeriv(output) * unit->outputDeriv ;

  for (idx=0, wt=Mweights(unit), dv=Mderivs(unit), preOut=MpreOutputs(unit),
       preDv=MpreOutputDerivs(unit) ;
       idx < numIncoming ; 
       idx++) {
    *(*preDv++) += inputDeriv *  (*wt++) ;
     (*dv++)    += inputDeriv * *(*preOut++) ;
  }
}

/**********************************************************************
 *				Utilities
 **********************************************************************/
static Proc	updateNetGradient (net)
  Net	net ;
{
  netBackward(net) ;
}

/***********************************************************************/
#define MAXLETTERS   27
#define MAXBANKWIDTH 25
#define SQ(x) ((x)*(x))
#define GAUSS(x,s) exp(-SQ(x)/SQ(s))
#define GAUSS_SD (0.1)

String _primaryStress = "'" ;
#define PRIMARY_STRESS (_primaryStress[0])
#define SECONDARY_STRESS '"'

Real _gaussValue[MAXLETTERS][MAXBANKWIDTH] ;

static	Boolean	computeExample(net, exampleSet, next) 
  Net		net ;
  ExampleSet	exampleSet ;
  int		next ;
{
  int	ex, i, s, p, wLen, pLen, nLet, gIdx ;
  String word, phon ;
  String hebrew_letters = "" ;
  int	nLetters = strlen(hebrew_letters), bankWidth = 20 ;
  SMlist *list = (SMlist *)getExampleSetData(exampleSet) ;
  ExampleRec *exRec ;
  Group  *group, inGroup, outGroup ;

  /* chose a word */
  while (!WITH_PROBABILITY(list->entries[ex=randomInteger(list->numEntries)].frequency)) ;
  word = list->entries[ex].graphemes ;
  wLen = strlen(word) ;
  phon = list->entries[ex].phonemes ;
  pLen = strlen(phon) ;
  /* load in example 0 (which should be 1 slice and all zeros) */
  MexampleNum(net) = 0 ;
  MexampleFrequency(net) = list->entries[ex].frequency ;
  MgetNumber(exampleSet, MexampleNum(net)) ;
  exampleSet->example[0].name = list->entries[ex].graphemes ;

  switch (McomputeExample(net)) {
  case 1:	/* sequential input and output for "bridge" network */
    nLet = strlen(hebrew_letters) ;
    for (gIdx=0, group=net->group; gIdx < net->numGroups; gIdx++, group++)
      if ((*group)->type&INPUT)       inGroup  = (*group) ;
      else if ((*group)->type&OUTPUT) outGroup = (*group) ;
    for (i = 0 ; i < nLet ; i++) {
      Unit inUnit  = inGroup->unit[i] ;
      Unit outUnit = outGroup->unit[i] ;
      for (s = 0 ; s < wLen ; s++)
	inUnit->clampingHistory[s] =
	  outUnit->targetHistory[s] = 
	  (word[wLen-s-1] == hebrew_letters[i] ? 1.0 : 0.0) ;
      /* all 0.0 at end */
      inUnit->clampingHistory[wLen] = outUnit->targetHistory[wLen] = 0.0 ;
    }
    return TRUE ;
  case 2:	/* random position over localist slots */
    {int nSlots = 8 ;
    int lSlots = floor((nSlots-wLen)/2.0) ;
    int rSlots = ceil((nSlots-wLen)/2.0) ;
    int lshift = (wLen-1 > lSlots ? lSlots : wLen-1) ;
    int rshift = (wLen-1 > rSlots ? rSlots : wLen-1) ;
    int shift  = lSlots - lshift + randomInteger(rshift+lshift+1) ;
    nLet = strlen(hebrew_letters) ;
    for (gIdx=0, group=net->group; gIdx < net->numGroups; gIdx++, group++)
      if ((*group)->type&INPUT)       inGroup  = (*group) ;
      else if ((*group)->type&OUTPUT) outGroup = (*group) ;
    for (s = 0 ; s < wLen ; s++) {
      int incr   = (s+shift)*nLet ;
      for (i = 0 ; i < nLet ; i++)
	if (word[wLen-s-1] == hebrew_letters[i]) 
	  inGroup->unit[i+incr]->output = outGroup->unit[i+incr]->target = 1.0 ;
    }}
    return TRUE ;
  case 3:	/* like 2 but with different input and output positions */
    {int nSlots = 8 ;
    int lSlots = floor((nSlots-wLen)/2.0) ;
    int rSlots = ceil((nSlots-wLen)/2.0) ;
    int lshift = (wLen-1 > lSlots ? lSlots : wLen-1) ;
    int rshift = (wLen-1 > rSlots ? rSlots : wLen-1) ;
    int inshift   = lSlots - lshift + randomInteger(rshift+lshift+1) ;
    int outshift  = lSlots - lshift + randomInteger(rshift+lshift+1) ;
    Group inRightGroup, inLeftGroup, outRightGroup, outLeftGroup ;
    nLet = strlen(hebrew_letters) ;
    /* NOTE: groups names are hard-wired */
    if ((inGroup      =groupFromName(net, "Input"))     == NULL) IErrorAbort("No group \"Input\"\n") ;
    if ((inRightGroup =groupFromName(net, "InputRight"))  == NULL) IErrorAbort("No group \"InputRight\"\n") ;
    if ((inLeftGroup  =groupFromName(net, "InputLeft"))  == NULL) IErrorAbort("No group \"InputLeft\"\n") ;
    if ((outRightGroup=groupFromName(net, "OutputRight"))  == NULL) IErrorAbort("No group \"OutputRight\"\n") ;
    if ((outLeftGroup =groupFromName(net, "OutputLeft"))  == NULL) IErrorAbort("No group \"OutputLeft\"\n") ;
    if ((outGroup     =groupFromName(net, "Output"))    == NULL) IErrorAbort("No group \"Output\"\n") ;
    inRightGroup->unit[inshift+wLen-1]->output   = 1.0 ;
    inLeftGroup->unit[inshift]->output           = 1.0 ;
    outRightGroup->unit[outshift+wLen-1]->output = 1.0 ;
    outLeftGroup->unit[outshift]->output         = 1.0 ;
    for (s = 0 ; s < wLen ; s++) {
      int in_incr  = (s+inshift)*nLet ;
      int out_incr = (s+outshift)*nLet ;
      for (i = 0 ; i < nLet ; i++)
	if (word[wLen-s-1] == hebrew_letters[i]) {
	  inGroup->unit[i+in_incr]->output = 1.0 ;
	  outGroup->unit[i+out_incr]->target = 1.0 ;
	}
    }}
    return TRUE ;
  case 4:	/* gaussian activity over banks of Hebrew letter units */
    {int u, l ;
    char *c ;
    Real wordWidth, sumWidth, mean ;
    for (l=0 ; l<MAXLETTERS ; l++) for (u=0 ; u<MAXBANKWIDTH ; u++) _gaussValue[l][u] = 0.0 ;
    wordWidth = (Real)strlen(word) ;
    /* accumulate gaussian activity in letter banks */
    for (c=word, sumWidth=0.0 ; *c ; c++) {
      /* find letter */
      for (l=0 ; hebrew_letters[l] != *c && l<nLetters; l++) ;
      if (l == nLetters) { IErrorAbort("'%c' (...%s) in %s not found", *c, c, word) ;  exit(0) ; }
      /* sumWidth is sum of earlier letter widths */
      mean = (sumWidth + 0.5) / wordWidth ;
      for (u=0 ; u<bankWidth ; u++) {
	Real x = (0.5+((float)u))/((float)bankWidth) ;
	_gaussValue[l][u] += GAUSS(x-mean,GAUSS_SD) ;
      }
      sumWidth++ ;
    }
    for (gIdx=0, group=net->group; gIdx < net->numGroups; gIdx++, group++)
      if ((*group)->type&INPUT)       inGroup  = (*group) ;
      else if ((*group)->type&OUTPUT) outGroup = (*group) ;
    for (l=0, i=0 ; l<nLetters ; l++)
      for (u=0 ; u<bankWidth ; u++, i++)
	inGroup->unit[i]->output = outGroup->unit[i]->target = MIN(_gaussValue[l][u], 1.0) ;
    }
    return TRUE ;
  case 5:	/* CELEX word boundaries */
    {String phonemes = "#$123456789@DEHIJNPQRSTUVZ_bdfghijklmnprstuvwz{" ;
    Group Phonemes, Stress, Boundary ;
    int p, s = 0, nPhonemes = strlen(phonemes) ;
    /* NOTE: groups names are hard-wired */
    if ((Phonemes =groupFromName(net, "Phonemes")) == NULL) IErrorAbort("No group \"Phonemes\"\n") ;
    if ((Stress   =groupFromName(net, "Stress"))   == NULL) IErrorAbort("No group \"Stress\"\n") ;
    if ((Boundary =groupFromName(net, "Boundary")) == NULL) IErrorAbort("No group \"Boundary\"\n") ;
    while (phon) {
      if (*phon == PRIMARY_STRESS) {
	Stress->unit[0]->clampingHistory[s] = 1.0 ; phon++ ;
      } else if (*phon == SECONDARY_STRESS) {
	Stress->unit[0]->clampingHistory[s] = 0.5 ; phon++ ;
      } else
	Stress->unit[0]->clampingHistory[s] = 0.0 ; 
      for (p = 0 ; p < nPhonemes ; p++) 
	Phonemes->unit[p]->clampingHistory[s] = (*phon == phonemes[p] ? 1.0 : 0.0) ;
      Stress->unit[0]->targetHistory[s] = (s == 0 ? 1.0 : 0.0) ;
      phon++ ; s++ ;
    }}
    return TRUE ;
  default:
    IErrorAbort("Unknown computeExample function %d", McomputeExample(net)) ;
  }
  return FALSE ;
}

static	Boolean	modifyExample(net, exampleSet, ex) 
  Net		net ;
  ExampleSet	exampleSet ;
  int		ex ;
{
  SMlist *list = (SMlist *)getExampleSetData(exampleSet) ;
  String word = list->entries[ex].graphemes ;

  switch (MmodifyExample(net)) {
  case 1:	/* gaussian activity over pseudo-Hebrew Orth input */
    {int i, u, l ;
    char *c ;
    char *letters  = "hmntvyABCDEFGHIJ" ;
    int bankWidth, nLetters = strlen(letters) ;
    int nLetMin = 3, nLetMax = 9 ;
    Real wordWidth, sumWidth, mean, sd ;
    Group Orthography ;
    /* WARNING: group name is hard-coded */
    if ((Orthography=groupFromName(net, "Orth"))==NULL) IErrorAbort("No group \"Orthography\"\n") ;
    bankWidth = Orthography->numUnits/nLetters ;
    wordWidth = (Real)strlen(word) ;
    sd = 0.15 ;		/* WARNING: hard-coded parameter */
    for (l=0 ; l<MAXLETTERS ; l++) for (u=0 ; u<MAXBANKWIDTH ; u++) _gaussValue[l][u] = 0.0 ;
    /* accumulate gaussian activity in letter banks */
    sumWidth = gaussianNoise(sd) ;		/* gaussian noise added to global position */
    for (c=word ; *c ; c++) {
      /* find letter */
      for (l=0 ; letters[l] != *c && l<nLetters; l++) ;
      if (l == nLetters) { IErrorAbort("'%c' (...%s) in %s not found", *c, c, word) ;  exit(0) ; }
      /* sumWidth is sum of earlier letter widths */
      mean = (sumWidth + 0.5) / wordWidth ;
      for (u=0 ; u<bankWidth ; u++) {
	Real x = (0.5+((float)u))/((float)bankWidth) ;
	_gaussValue[l][u] += GAUSS(x-mean,sd) ;
      }
      sumWidth++ ;
    }
    for (l=0, i=0 ; l<nLetters ; l++)
      for (u=0 ; u<bankWidth ; u++, i++)
	Orthography->unit[i]->output = MIN(_gaussValue[l][u], 1.0) ;
    }
    return TRUE ;
  default:
    IErrorAbort("Unknown modifyExample function %d", MmodifyExample(net)) ;
  }
  return FALSE ;
}
/***********************************************************************/
static	Boolean	getExample(net, exampleSet, next) 
  Net		net ;
  ExampleSet	exampleSet ;
  int		next ;
{
  if (McomputeExample(net)) return computeExample(net, exampleSet, next) ;
  _presentedAssociate = FALSE ;
  if (MbatchSize(net) < exampleSet->numExamples && !MincTraining(net)) {
    int last      = MexampleNum(net) ;
    int numAssocs = exampleSet->example[last].extension->numAssocs ;
    if (MassocProb(net) > 0.0 && numAssocs > 0 && WITH_PROBABILITY(MassocProb(net))) {
      /* select associated example */
      if (numAssocs == 1)
	MexampleNum(net) = exampleSet->example[last].extension->assoc[0] ;
      else
	do {
	  MexampleNum(net) = exampleSet->example[last].extension->assoc[randomInteger(numAssocs)] ;
	} while (MexampleNum(net) == NO_ASSOC || !WITH_PROBABILITY(exampleSet->example[MexampleNum(net)].frequency)) ;
      _presentedAssociate = TRUE ;
    } else
      /* select random example according to its frequency */
      do {
	MexampleNum(net) = randomInteger(exampleSet->numExamples);
	MexampleFrequency(net) = exampleSet->example[MexampleNum(net)].frequency ;
      } while (MsampleExamples(net) && !WITH_PROBABILITY(MexampleFrequency(net))) ;
  } else {
    MexampleNum(net) = next ;
    MexampleFrequency(net) = exampleSet->example[next].frequency ;
    if (MsampleExamples(net) && !WITH_PROBABILITY(MexampleFrequency(net)))
      return FALSE ;       /* skip this example */
  }
  exampleSet->example[MexampleNum(net)].extension->presentations++ ;
  MgetNumber(exampleSet, MexampleNum(net)) ;
  if (MmodifyExample(net)) modifyExample(net, exampleSet, MexampleNum(net)) ;
  return TRUE ;
}

/**********************************************************************/
static void	displayStates(net, slice) 
  Net		net ;
  int		slice ;
{
  int i ;
  net->currentTime = slice ;
  netForAllUnits(net, ALL, (UnitDataProc)saveOutputToSlice, (void *)slice) ;
  IDoTrace(&forwardEachTime) ;
  for (i=0 ; i<MdelayCount(net)*1000000 ; i++) ;
}

/**********************************************************************/
static void	copyContext(group, data)
  Group		group ;
  void		*data ;
{
  if (McopyFromLayer(group)) {
    Group	fromGroup = group->net->group[McopyFromLayer(group)] ;
    if (group->numUnits != fromGroup->numUnits) {
      IErrorAbort("Source %s numUnits (%d) != Context %s numUnits (%d)",
		  fromGroup->name, fromGroup->numUnits, group->name, group->numUnits) ;
    } else if (McopyWTA(group)) {
      int idx, maxIdx ;  Real this, max = -MAXFLOAT ;
      for (idx = 0 ; idx < fromGroup->numUnits ; idx++)
	if ((this=fromGroup->unit[idx]->output) > max) { max = this ; maxIdx = idx ; }
      for (idx = 0 ; idx < maxIdx ; idx++)
	group->unit[idx]->output = 0.0 ;
      group->unit[maxIdx]->output = 1.0 ;
      for (idx = maxIdx+1 ; idx < group->numUnits ; idx++) 
	group->unit[idx]->output = 0.0 ;
    } else {
      Unit	*toPtr   = group->unit ;
      Unit	*endPtr  = toPtr + group->numUnits ;
      Unit	*fromPtr = fromGroup->unit ;
      while (toPtr != endPtr) {
	Unit	to   = *(toPtr++) ;
	Unit	from = *(fromPtr++) ;
	to->output = from->output ;
      }
    }
  }
}

/***********************************************************************/
static	void	initStats(net)
  Net	net ;
{
  if (net->currentEpoch==0) MlastError(net) = MaveError(net) = -1.0 ;
  if (net->mz->repetitionCount==0 || net->mz->nIterations==0 ||
      (net->mz->nIterations-1)%net->mz->repetitionCount==0) {
    net->error 		   	= 0.0 ;
    net->cost 		   	= 0.0 ;
    MpExampleCorrect(net)	= 0.0 ;
    MpSliceCorrect(net)		= 0.0 ;
    MaveTime(net)		= 0.0 ;
    MaveFixations(net)		= 0.0 ;
  }
}

/***********************************************************************/
static	void	normStats(net)
  Net	net ;
{
  if (net->mz->repetitionCount==0 || 
      net->mz->nIterations%net->mz->repetitionCount==0) {
    Real	scale = 1.0 ;
    if (net->mz->nIterations && net->mz->repetitionCount)
      scale = (Real)net->mz->repetitionCount ;
    net->error		     	/= scale ;
    net->cost		     	/= scale ;
    scale 			*= net->batchSize ;
    MpSliceCorrect(net)		/= scale ;
    MpExampleCorrect(net)	/= scale ;
    MaveTime(net)		/= scale ;
    MaveFixations(net)		/= scale ;
  }
}


/**********************************************************************/
static Boolean	outputCorrectBridge(net, first, last, tolerance)
  int	first, last ;
  Net	net ;
  double tolerance ;
{
  int	groupIdx ;
  Group	*group ;

  if (tolerance <= 0.0) return FALSE ;

  for (groupIdx=first, group= &(net->group[first]);
       groupIdx <= last ;
       groupIdx++, group++) {
    if ((*group)->type&OUTPUT &&
	!groupOutputCorrect((*group), tolerance)) return FALSE ;
  }
  return TRUE ;
}

/***********************************************************************
 *			netIterate[Bridge/Refixate]
 ***********************************************************************/
static void	netIterateBridge(net, training)
  Net		net ;
  Boolean	training ;
{
  int	inSlice, outSlice ;
  int	bridge = MbridgeLayer(net), lastGroup = net->numGroups-1 ;
  Real	tolerance = McorrectTolerance(net) ;
  Boolean abort = FALSE ;

  /* 
     MpredictOutput(net) = 
       0  Only run outSlices that correspond to presented inSlices
       1  Run all outSlices for each inSlice
       2  Only run additional predicted outSlices if current one is correct
          (generally means run only one extra outSlice)

     MabortIfIncorrect(net) = 1  Stop running inSlices if corresponding outSlices are incorrect
  */

  netDoUnits(net, 1, bridge-1, BIAS, TRUE, (UnitDataProc)initContextOutput, NULL) ;
  for (inSlice = 0 ; inSlice < MmaxSlice(net) && !abort ; inSlice++) {
    int	maxOutSlice = (MpredictOutput(net) ? MmaxSlice(net) : inSlice+1) ;
    int numOutCorrect  = 0 ;
    netForAllUnits(net, INPUT,  (UnitDataProc)clampInput, (void *)inSlice) ;
    netForwardLayers(net, 1, bridge, inSlice, training) ;
    netDoUnits(net, bridge+1, lastGroup, BIAS, TRUE, (UnitDataProc)initContextOutput, NULL) ;

    for (outSlice = 0 ; outSlice < maxOutSlice && !abort ; outSlice++) {
      netForwardLayers(net, bridge+1, lastGroup, outSlice, training) ;
      if (!Mrunning(net)) displayStates(net, outSlice) ;
      if (training) 
	netDoUnits(net, lastGroup, bridge+1, INPUT|BIAS, FALSE, (UnitDataProc)backward, NULL) ;
      if (tolerance > 0.0 && outputCorrectBridge(net, bridge+1, lastGroup, tolerance))
	numOutCorrect++ ;
      if (numOutCorrect <= outSlice) {
	if (MabortIfIncorrect(net) && outSlice <= inSlice) { abort = TRUE ; break ; }
	if (MpredictOutput(net)==2) break ;
      }
      if (outSlice < MmaxSlice(net)-1) 
	netDoGroups(net, bridge+1, lastGroup, BIAS, TRUE, (GroupDataProc)copyContext, NULL) ;
    } /* outSlice */

    MpSliceCorrect(net)  += numOutCorrect/((Real)(maxOutSlice*MmaxSlice(net))) ;
    if (training) 
      netDoUnits(net, bridge, 1, INPUT|BIAS, FALSE, (UnitDataProc)backward, NULL) ;
    if (inSlice < MmaxSlice(net)-1)
      netDoGroups(net, 1, bridge, BIAS, TRUE, (GroupDataProc)copyContext, NULL) ;
    else if (numOutCorrect == maxOutSlice) 
      MpExampleCorrect(net)++ ;
  } /* inSlice */
}

/***********************************************************************/
static void	netIterateRefixate(net, training)
  Net		net ;
  Boolean	training ;
{
  int	slice, numCorrect = 0, outCorrect ;
  int	fixation=0, attention=0, saccade=0, outputOK, done, attOK ; 
  Group	Input, Attention, AttentionOut, Done ;

  if (!MinputGroupName(net))     MinputGroupName(net)     = "Letters" ;
  /* note that this is the INPUT Attention group */
  if (!MattentionGroupName(net)) MattentionGroupName(net) = "cAttention" ;
  if (!MdoneGroupName(net))      MdoneGroupName(net)      = "Done" ;

  if ((Input    =groupFromName(net, MinputGroupName(net)))     == NULL) 
    IErrorAbort("no \"%s\" group", MinputGroupName(net)) ;
  if ((Attention=groupFromName(net, MattentionGroupName(net))) == NULL) 
    IErrorAbort("no \"%s\" group", MattentionGroupName(net)) ;
  AttentionOut = net->group[Attention->extension->copyFromLayer] ;
  if ((Done     =groupFromName(net, MdoneGroupName(net)))      == NULL) 
    IErrorAbort("no \"%s\" group",MdoneGroupName(net)) ;
  
  /* CONTEXT GROUPS MUST BE TYPE BIAS */
  netForAllUnits(net, BIAS, (UnitDataProc)initContextOutput, NULL) ;
  MaveFixations(net)++ ;

  /* FOR EACH TIMESLICE - ONE PER PHONEME */
  for (slice = 0 ; slice < MmaxSlice(net) ; slice++) {
    net->currentTime = slice ;
    /* this is applied to the OUTPUT group because targets are shifted */
    Fixate(Input, AttentionOut, saccade, attention, slice) ;
    copyContext(Attention, NULL) ;
    netForAllUnits(net, INPUT,  (UnitDataProc)clampInput, (void *)slice) ;
    netForAllUnits(net, OUTPUT, (UnitDataProc)setTarget,  (void *)slice) ;

    /* FORWARD AND BACKWARD PASSES */
    netForward(net) ;
    if (!Mrunning(net)) displayStates(net, slice) ;
    if (outCorrect=(McorrectWTA(net) ?
		    outputCorrectWTA(net, McorrectTolerance(net)) :
		    outputCorrect(net, McorrectTolerance(net))))
      numCorrect++ ;
    if (training) {
      outputOK = outCorrect ;
      done     = (outputOK && Done->unit[0]->target==1.0) ;
      netBackward(net) ;
    } else {
      outputOK = outputClean(net, McorrectTolerance(net)) ;
      done     = (Done->unit[0]->output > maxOutput(AttentionOut)) ;
    }
    if (Mdebug(net)) 
      fprintf(dout, "%d\tfixate %d(%d)\t", slice, fixation, attention) ;

    /* DONE - GET OUT */
    if (done) {
      if (Mdebug(net)) fprintf(dout, "done    %s\n", 
			       (attention ? "in periphery" : "at fixation ")) ;
      break ;

    /* CORRECT - TRY NEXT PHONEME WITHOUT REFIXATING */
    } else if (outputOK) {
      if (Mdebug(net)) fprintf(dout, "correct %s - no saccade\n", 
			       (attention ? "in periphery" : "at fixation ")) ;
      saccade   = fixation ;	/* no change */
      attention = (training ? maxTargetPosition(AttentionOut) : maxOutputPosition(AttentionOut)) ;
      netForAllGroups(net, BIAS, (GroupDataProc)copyContext, NULL) ;

    /* FAILED ON FIXATED GRAPHEME - REFIXATE NEXT GRAPHEME */
    } else if (attention == 0) { 
      attention = (training ? maxTargetPosition(AttentionOut) : maxOutputPosition(AttentionOut)) ;
      if (Mdebug(net)) fprintf(dout, "fail    at fixation  - saccade %d\n", attention) ;
      fixation += attention ;
      saccade   = fixation ;
      attention = 0 ;
      MaveFixations(net)++ ;
      if (slice < MmaxSlice(net)-1)
	netForAllGroups(net, BIAS, (GroupDataProc)copyContext, NULL) ;

    /* FAILED ON NONFIXATED GRAPHEME - REFIXATE IT AND TRY AGAIN */
    } else {
      if (Mdebug(net)) fprintf(dout, "fail    in periphery - saccade %d and repeat\n", attention) ;
      fixation += attention ;
      saccade   = attention ;		/* applied to OLD states/targets */
      attention = 0 ;
      MaveFixations(net)++ ;
      slice-- ;
      netForAllUnits     (net, BIAS,       (UnitDataProc)initContextOutput,      NULL) ;
      netForAllOtherUnits(net, INPUT|BIAS, (UnitDataProc)restoreOutputFromSlice, (void *)slice) ;
    }
  } /* slice */
  if (numCorrect == MmaxSlice(net)) MpExampleCorrect(net)++ ;
  MpSliceCorrect(net)  += numCorrect/((Real) MmaxSlice(net)) ;
}

/**********************************************************************/
static void	netIterate(net, training)
  Net		net ;
  Boolean	training ;
{
  int slice, numCorrect = 0 ;
  Real	pCorr, tolerance = McorrectTolerance(net) ;

  MmaxSlice(net) = lastTargetSlice(net) ;

  if (MbridgeLayer(net))
    netIterateBridge(net, training) ;
  else if (Mrefixate(net))
    netIterateRefixate(net, training) ;
  else {
    /* 
     * assumes context groups are type BIAS
     */
    if (MinitializeContext(net))
      netForAllUnits(net, BIAS, (UnitDataProc)initContextOutput, NULL) ;
    for (slice = 0 ; slice < MmaxSlice(net) ; slice++) {
      net->currentTime = slice ;
      if (Mmarcus(net)) {
	if (!MinitializeContext(net)) {		/* pretraining */
	  copyTargetsToOutputs(net->group[MARCUS_NEXT],net->group[MARCUS_INPUT]) ;
	  marcusPickNextSyllable(net) ;	/* sets targets over MARCUS_NEXT */
	  if (MinputNoise(net)>0.0)	  
	    netForAllUnits(net, OUTPUT, (UnitDataProc)addTargetNoise, (void *)&MinputNoise(net)) ;
	  copyOutputsToTargets(net->group[MARCUS_INPUT],net->group[MARCUS_CURRENT]) ;
	  netForAllUnits(net, OUTPUT, (UnitDataProc)saveTarget, (void *)slice) ;
	} else if (MinputNoise(net)>0.0) {
	  if (slice == 0) 
	    netForAllUnits(net, INPUT, (UnitDataProc)clampInput, (void *)slice) ;
	  else
	    copyTargetsToOutputs(net->group[MARCUS_NEXT],net->group[MARCUS_INPUT]) ;
	  netForAllUnits(net, OUTPUT, (UnitDataProc)setTarget, (void *)slice) ;
	  netForAllUnits(net, OUTPUT, (UnitDataProc)addTargetNoise, (void *)&MinputNoise(net)) ;
	  copyOutputsToTargets(net->group[MARCUS_INPUT],net->group[MARCUS_CURRENT]) ;
	  netForAllUnits(net, OUTPUT, (UnitDataProc)saveTarget, (void *)slice) ;
	}
      } else {
	netForAllUnits(net, INPUT,  (UnitDataProc)clampInput, (void *)slice) ;
	netForAllUnits(net, OUTPUT, (UnitDataProc)setTarget,  (void *)slice) ;
      }
      netForward(net) ;
      if (!Mrunning(net)) displayStates(net, slice) ;
      if (training) netBackward(net) ;
      if (tolerance > 0.0 && 
	  (McorrectWTA(net) ?
	   outputCorrectWTA(net, McorrectTolerance(net)) :
	   outputCorrect(net, McorrectTolerance(net))))
	numCorrect++ ;
      if (MpredictOutput(net)==2 && numCorrect <= slice) break ;
      if (slice < MmaxSlice(net)-1)
	netForAllGroups(net, BIAS, (GroupDataProc)copyContext, NULL) ;
      if (!MinitializeContext(net) && slice == MmaxSlice(net)-1)
	netForAllGroups(net, BIAS, (GroupDataProc)copyContext, NULL) ;
    }
    MpSliceCorrect(net) += (pCorr=numCorrect/((Real) MmaxSlice(net))) ;
    if (pCorr == 1.0) MpExampleCorrect(net)++ ;
  }
  if (!Mrunning(net)) net->currentTime = MmaxSlice(net) ;
}
/***********************************************************************
 *			netCalculate
 ***********************************************************************/
static Real	_sumFreq ;

static void	netCalculate(net, exampleSet, training)
  Net		net ;
  ExampleSet	exampleSet ;
  Boolean	training ;
{
  int ex ;
  SMlist *list = (SMlist *)(exampleSet->extension->data) ;

  if (MincTraining(net) && net->currentEpoch == 0) _sumFreq = list->entries[0].frequency ;

  if (!Mcompiled(net)) netCompile(net) ;
  netForAllUnits     (net, ALL,       (UnitDataProc)unitConstrainWeights, NULL) ;
  netForAllOtherUnits(net, INPUT|BIAS, (UnitDataProc)getCompiledWeights, NULL) ;

  Mrunning(net) = TRUE ;
  initStats(net) ;
  if (training) {
    initEpoch(net, MalphaOnset(net)) ;
    netForAllUnits(net, ALL, (UnitDataProc)zeroLinkDerivs, net) ;
  }
  if (net->type&RECURRENT && !MinitializeContext(net) && net->mz->nIterations==0) {
    fprintf(dout,"Initializing context\n") ;
    netForAllUnits(net, BIAS, (UnitDataProc)initContextOutput, NULL) ;
  }

  for (ex = 0 ; ex < MbatchSize(net) ; ++ex) {
    if (!getExample(net, exampleSet, ex)) continue ;

    if (net->type & RECURRENT)				/* elman network */
      netIterate(net, training) ;
    else {
      netForward(net) ;
      if (MincTraining(net)) {
	if (MincTrainingSMnsyl(net)) {
	  if (SMnsylPronounceCorrect(net,list->entries[ex].phonemes)) 
	    MpExampleCorrect(net)++ ;
	} else 
	  if (outputCorrect(net, McorrectTolerance(net)))
	    MpExampleCorrect(net)++ ;
      }
      if (training) netBackward(net) ;
    }
  }
  if (ex <= 0)
    IErrorAbort("netCalculate: no examples processed") ;

  if (training) {
    /* compiled derivatives are put back first because cost functions 
       currently use standard data structures */
    netForAllOtherUnits(net, INPUT|BIAS, (UnitDataProc)putCompiledDerivs, NULL) ;
    netForAllUnits(net, ALL, (UnitDataProc)maybeFixBias, NULL) ;
    if (MlengthScaleDerivs(net))
      netForAllUnits(net, ALL, (UnitDataProc)lengthScaleDerivs, net) ;

    /* update the cost after everything else is done */
    if (MweightCostIncrement(net) > 0.0) 
      netAdaptWeightCost(net) ;
    netUpdateWeightCost(net) ;
  }
  normStats(net) ;
  if (MincTraining(net)) {
    while (MpExampleCorrect(net) >= MincTrainingThreshold(net) &&
	   MbatchSize(net) < exampleSet->numExamples) {
      Real nCorr = ((Real)MbatchSize(net))*MpExampleCorrect(net) ;
      _sumFreq += list->entries[ex].frequency ;
      MgetNumber(exampleSet,MbatchSize(net)) ;
      MupdateNetActivities(currentNet) ;
      if (MincTrainingSMnsyl(net)) {
	if (SMnsylPronounceCorrect(net,list->entries[ex].phonemes)) nCorr++ ;
      } else 
	if (outputCorrect(net, McorrectTolerance(net))) nCorr++ ;
      MbatchSize(net)++ ;
      MpExampleCorrect(net) = nCorr/((Real)MbatchSize(net)) ;
      fprintf(dout,"INC: %d adding \"%s\" batchSize %d  pCorrect %f  sumFreq %f %s\n",
	      net->currentEpoch, exampleSet->example[MbatchSize(net)-1].name, MbatchSize(net), 
	      MpExampleCorrect(net), _sumFreq, 
	      (MbatchSize(net) == exampleSet->numExamples ? "ALL" : "")) ;
    }
  }
  Mrunning(net) = FALSE ;
}

/***********************************************************************/
static void	calculateNetErrorDeriv(net, exampleSet)
  Net		net ;
  ExampleSet	exampleSet ;
{
  netCalculate(net, exampleSet, TRUE) ;
}

/*********************************************************************/
static void	calculateNetError(net, exampleSet)
  Net		net ;
  ExampleSet	exampleSet ;
{
  netCalculate(net, exampleSet, FALSE) ;
}

/**********************************************************************/
Proc	updateNetActivity (net)
  Net	net ;
{
  if (!Mcompiled(net)) netCompile(net) ;
  netForAllUnits     (net, ALL,       (UnitDataProc)unitConstrainWeights, NULL) ;
  netForAllOtherUnits(net, INPUT|BIAS, (UnitDataProc)getCompiledWeights, NULL) ;

  Mrunning(net) = FALSE ;
  net->error = MlastError(net) = MaveError(net) = 0.0 ;
  MaveTime(net) = 0.0 ;
  MoutputsIncorrect(net) = 0 ;
  MpSliceCorrect(net) = MpExampleCorrect(net) =  MaveFixations(net) = 0.0 ;

  if (McomputeExample(net)) {
    computeExample(net, net->trainingExampleSet, net->trainingExampleSet->currentIdx) ;
    markToRebuildDisplay(ACTIVITY_DISPLAY) ;
  }
  if (MmodifyExample(net)) {
    modifyExample(net, net->trainingExampleSet, net->trainingExampleSet->currentIdx) ;
    markToRebuildDisplay(ACTIVITY_DISPLAY) ;
  }

  if (MtargetExternalInput(net)>0.0 && MtargExtInConstant(net)>0 && net->cost > 0.0) 
    /* stored by pronounce in SM.c --- NET->ERROR WILL NOT BE ACCURATE */
    MexampleFrequency(net) = net->cost ;
  else
    MexampleFrequency(net) = 1.0 ;

  if (net->type & RECURRENT)  	/* elman network */
    netIterate(net, FALSE) ;
  else
    netForward(net) ;
    
  IDoTrace(&afterExample) ;
}
/***********************************************************************/
/**********************************************************************/
static void	linkReadLensWeight(link, data)
  Link		link ;
  void		*data ;
{
  Real	wt ;
  if (fscanf((FILE *)data, "%f", &wt))
    link->weight = wt ;
  else 
    IErrorAbort("Unable to read weight for link %s: got %f instead\n", link->name, wt) ;
}

static void	unitReadLensWeights(unit, data)
  Unit		unit ;
  void		*data ;
{
  unitForAllLinks(unit, INCOMING, (LinkDataProc)linkReadLensWeight, data) ;
}  

int	command_loadLensWeights(tokc, tokv)
  int	tokc ;
  char	**tokv ;
{
  int	numWeights ;
  FILE	*fp ;

  IUsage("filename");
  if (GiveHelp(tokc)) {
    ISynopsis("Loads weights written by Lens");
    IHelp
      (IHelpArgs,
       "Loads weights written by Lens.",
       NULL) ;
    return 1 ;
  }

  if (currentNet == NULL) IErrorAbort("There is no current net.") ;
  fp = IOpenFileOrAbort(tokv[1], "r", NULL) ;
  fscanf(fp, "%d\n\n", &numWeights) ;
  fprintf(dout, "Reading %d weights....\n", numWeights) ;
  netForAllUnits(currentNet, ALL, (UnitDataProc)unitReadLensWeights, (void *)fp) ;
  ICloseFile(fp, NULL) ;
  return 1 ;
}
/**********************************************************************/
/**********************************************************************/
static int	morph_syllable, morph_nSyllables, morph_n, morph_exN ;
static Real	morph_inputStrength, morph_tolerance ;
static String	morph_exName ;
#define Mboundary0(unit) ((unit)->inputDeriv)
#define Mboundary1(unit) ((unit)->outputDeriv)

static void	countMorphInputs(unit, data)
  Unit		unit ;
  void		*data ;
{
  Mn(unit) = morph_n++ ;
}

static void	initializeMorphBoundaries(unit, data)
  Unit		unit ;
  void		*data ;
{
  Mboundary0(unit) = 0.0 ;  Mboundary1(unit) = 0.0 ;
}

static void	reduceInputStrength(unit, data)
  Unit		unit ;
  void		*data ;
{
  if ((morph_syllable == 0 && Mn(unit) < 15) || 
      (morph_syllable == 1 && Mn(unit) >= 15)) {
    My(unit) *= morph_inputStrength ;
  }
}

static void	updateMorphBoundaries(unit, data)
  Unit		unit ;
  void		*data ;
{
  if (morph_syllable == 0) {
    if (Mboundary0(unit) == 0.0 && fabs(Md(unit)-My(unit)) > morph_tolerance) {
      Mboundary0(unit) = morph_inputStrength ;
    }
  } else {
    if (Mboundary1(unit) == 0.0 && fabs(Md(unit)-My(unit)) > morph_tolerance) {
      Mboundary1(unit) = morph_inputStrength ;
    }
  }
}

static void	printMorphBoundaries(unit, data)
  Unit		unit ;
  void		*data ;
{
  fprintf(dout, "%s %d %4.2f %4.2f\n", morph_exName, morph_exN, Mboundary0(unit), Mboundary1(unit)) ;
}

int command_distortMorph (tokc, tokv)
  int	tokc ;
  String	tokv[] ;
{
  int		ex, example_n = -1, exampleType = TRAINING ;
  Real		decrement = 0.01, tolerance = 0.5 ;
  String	type=NULL, name=NULL ;
  ExampleSet	exampleSet ;
  struct ARG_DESC *argd ;

  if (currentNet == NULL) IErrorAbort("There is no current net.") ;
  argd = StartArgs(tokv[0]);
  Args(argd, "[-Ptype <type>%s%?]", 	 &type, 	"type of example set (one of TRAINING TESTING VALIDATION)");
  Args(argd, "[-Pname <str>%s%?]", 	 &name, 	"only run examples with name matching <string>") ;
  Args(argd, "[-Ptolerance <real>%r%?]", &tolerance, 	"correct tolerance (diff from target; default 0.5)") ;
  Args(argd, "[-Pdecrement <real>%r%?]", &decrement, 	"decrement to input activations (default 0.01)") ;
  Args(argd, "[-Pexample <n>%i%?]", 	 &example_n, 	"example to run");
  EndArgs(argd);
  if (GiveHelp(tokc)) {
    ISynopsis("componential analysis of morphology network") ;
    IHelp
      (tokc, tokv[0], NULL, synopsis,
       NULL);
    return 0;
  }
  ParseArgs(argd, tokc, tokv, 0);
  if (type) exampleType = findMask(type) ;
  exampleSet = netGetExampleSet(currentNet, exampleType) ;
  if (exampleSet == NULL) IErrorAbort("No example set specified for net \"%s\".", currentNet->name) ;

  Mreset(exampleSet) ;
  morph_tolerance = tolerance ;
  morph_n = 0 ;
  morph_nSyllables = 2 ;
  netForAllUnits(currentNet, INPUT, (UnitDataProc)countMorphInputs, NULL) ;
  for (ex = 0; ex < exampleSet->numExamples; ex++) {
    if ((example_n >= 0 && ex != example_n) ||
	(name           && !strstr(exampleSet->example[ex].name, name)))
      continue ;
    morph_exName = exampleSet->example[ex].name ;
    morph_exN    = ex ;
    netForAllUnits(currentNet, OUTPUT, (UnitDataProc)initializeMorphBoundaries, NULL) ;
    for (morph_syllable = 0 ; morph_syllable < morph_nSyllables ; morph_syllable++) {
      for (morph_inputStrength = 1.0 ; morph_inputStrength > 0.0 ; morph_inputStrength -= decrement) {
	MgetNumber(exampleSet, ex) ;
	netForAllUnits(currentNet, INPUT, (UnitDataProc)reduceInputStrength, NULL) ;
	updateNetActivity(currentNet) ;
	netForAllUnits(currentNet, OUTPUT, (UnitDataProc)updateMorphBoundaries, NULL) ;
      }
    }
    netForAllUnits(currentNet, OUTPUT, (UnitDataProc)printMorphBoundaries, NULL) ;
    fflush(dout) ;
  }
  return 1 ;
}
    
/*****************************************************************************/
#define MORPH_NEXAMPLES 320
#define MORPH_NSTATES 300
static Real **morph_states ;

static void	morphSaveUnitState(unit, data)
  Unit		unit ;
  void		*data ;
{
  morph_states[(int)data][morph_n++] = My(unit) ;
}

int command_corrMorph (tokc, tokv)
  int	tokc ;
  String	tokv[] ;
{
  int		ex, ex2, exampleType = TRAINING ;
  String	type=NULL ;
  ExampleSet	exampleSet ;
  struct ARG_DESC *argd ;

  if (currentNet == NULL) IErrorAbort("There is no current net.") ;
  argd = StartArgs(tokv[0]);
  Args(argd, "[-Ptype <type>%s%?]", 	 &type, 	"type of example set (one of TRAINING TESTING VALIDATION)");
  EndArgs(argd);
  if (GiveHelp(tokc)) {
    ISynopsis("correlations of morphology network hidden representations") ;
    IHelp
      (tokc, tokv[0], NULL, synopsis,
       NULL);
    return 0;
  }
  ParseArgs(argd, tokc, tokv, 0);
  if (type) exampleType = findMask(type) ;
  exampleSet = netGetExampleSet(currentNet, exampleType) ;
  if (exampleSet == NULL) IErrorAbort("No example set specified for net \"%s\".", currentNet->name) ;

  morph_states = ((Real **)calloc(MORPH_NEXAMPLES, sizeof(Real*))) ;
  for (ex = 0 ; ex < MORPH_NEXAMPLES ; ex++)
    morph_states[ex] = ((Real *)calloc(MORPH_NSTATES, sizeof(Real))) ;

  Mreset(exampleSet) ;
  for (ex = 0; ex < exampleSet->numExamples; ex++) {
    int i ;
    MgetNumber(exampleSet, ex) ;
    updateNetActivity(currentNet) ;
    morph_n = 0 ;
    netForAllOtherUnits(currentNet, INPUT|OUTPUT|BIAS, (UnitDataProc)morphSaveUnitState, (void *)ex) ;
  }
  for (ex = 0; ex < exampleSet->numExamples; ex++)
    for (ex2 = ex+1; ex2 < exampleSet->numExamples; ex2++)
      fprintf(dout, "%d %s %d %s %f\n", 
	      ex, exampleSet->example[ex].name, ex2, exampleSet->example[ex2].name,
	      correlation(morph_states[ex],morph_states[ex2],MORPH_NSTATES)) ;

  for (ex = 0 ; ex < MORPH_NEXAMPLES ; ex++) free(morph_states[ex]) ;
  free(morph_states) ;
  return 1 ;

}

/***********************************************************************/
static int _postIdx ;

static void	saveSomeWeightsFromGroup(unit, data)
  Unit		unit ;
  void		*data ;
{
  Group	preGroup = (Group)data ;
  int	idx, unitIdx ;
  int	numIncoming = unit->numIncoming ;
  Link	*incoming   = unit->incomingLink ;

  for (idx = 0 ; idx < unit->group->numUnits ; idx++)
    if (unit == unit->group->unit[idx]) { unitIdx = idx ; continue ; }

  for (idx = 0 ; idx < numIncoming ; ++idx) {
    Link        link    = incoming[idx] ;
    if (link->preUnit->group == preGroup) {
      fprintf(dout, "set currentNet.group[%d].unit[%d].incomingLink[%d].weight = %f\n",
	      _postIdx, unitIdx, idx, link->weight) ;
    }
  }
}

/***********************************************************************/
int	command_saveSomeWeights (tokc, tokv)
  int	tokc ;
  char	*tokv[] ;
{
  int idx ;
  String preGroupRE = NULL, postGroupRE = NULL ;
  Group	*preGroupArray,  *preGroup ;
  Group	*postGroupArray, *postGroup ;
  struct ARG_DESC *argd ;

  if (currentNet == NULL)
    IErrorAbort("There is no current net.") ;

  argd = StartArgs(tokv[0]);
  Args(argd, "[-PpreGroup <regexp>%s%?]",	&preGroupRE,	"regular expression specifying source groups.");
  Args(argd, "[-PpostGroup <regexp>%s%?]",	&postGroupRE,   "regular expression specifying destination groups.");
  EndArgs(argd);
  if (GiveHelp(tokc)) {
    ISynopsis("save some weights") ;
    IHelp
      (tokc, tokv[0], NULL, synopsis, "", 
      NULL);
    return 0;
  }
  ParseArgs(argd, tokc, tokv, 0);

  postGroupArray = groupsFromRegex(currentNet, postGroupRE) ;
  if (postGroupArray == NULL)
    IErrorAbort("Bad string: \"%s\"", postGroupRE) ;
  if (postGroupArray[0] == NULL)
    IErrorAbort("No match: \"%s\"", postGroupRE) ;

  preGroupArray = groupsFromRegex(currentNet, preGroupRE) ;
  if (preGroupArray == NULL)
    IErrorAbort("Bad string: \"%s\"", preGroupRE) ;
  if (preGroupArray[0] == NULL)
    IErrorAbort("No match: \"%s\"", preGroupRE) ;

  for (postGroup = postGroupArray ; *postGroup != NULL ; ++postGroup) {
    for (idx = 0 ; idx < currentNet->numGroups ; idx++)
      if (*postGroup == currentNet->group[idx]) { _postIdx = idx ; continue ; }
    for (preGroup = preGroupArray ; *preGroup != NULL ; ++preGroup)
      groupForAllUnits(*postGroup, (UnitDataProc)saveSomeWeightsFromGroup, *preGroup) ;
  }
  free(preGroupArray) ;
  free(postGroupArray) ;
  return 1 ;
}
